!> @file pmc_interface_mod.f90
!------------------------------------------------------------------------------!
! This file is part of the PALM model system.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2019 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: pmc_interface_mod.f90 3697 2019-01-24 17:16:13Z hellstea $
! Bugfix: upper k-bound in the child initialization interpolation 
! pmci_interp_1sto_all corrected.
! Copying of the nest boundary values into the redundant 2nd and 3rd ghost-node 
! layers is added to the pmci_interp_1sto_*-routines.
! 
! 3681 2019-01-18 15:06:05Z hellstea
! Linear interpolations are replaced by first order interpolations. The linear 
! interpolation routines are still included but not called. In the child 
! inititialization the interpolation is also changed to 1st order and the linear
! interpolation is not kept.
! Subroutine pmci_map_fine_to_coarse_grid is rewritten.
! Several changes in pmci_init_anterp_tophat.
! Child's parent-grid arrays (uc, vc,...) are made non-overlapping on the PE-
! subdomain boundaries in order to allow grid-spacing ratios higher than nbgp. 
! Subroutine pmci_init_tkefactor is removed as unnecessary.
! 
! 3655 2019-01-07 16:51:22Z knoop
! Remove unused variable simulated_time
! 
! 3636 2018-12-19 13:48:34Z raasch
! nopointer option removed
! 
! 3592 2018-12-03 12:38:40Z suehring
! Number of coupled arrays is determined dynamically (instead of a fixed value
! of 32)
! 
! 3524 2018-11-14 13:36:44Z raasch
! declaration statements rearranged to avoid compile time errors
! 
! 3484 2018-11-02 14:41:25Z hellstea
! Introduction of reversibility correction to the interpolation routines in order to 
! guarantee mass and scalar conservation through the nest boundaries. Several errors 
! are corrected and pmci_ensure_nest_mass_conservation is permanently removed. 
! 
! 3274 2018-09-24 15:42:55Z knoop
! Modularization of all bulk cloud physics code components
! 
! 3241 2018-09-12 15:02:00Z raasch
! unused variables removed
! 
! 3217 2018-08-29 12:53:59Z suehring
! Revise calculation of index bounds for array index_list, prevent compiler 
! (cray) to delete the loop at high optimization level.  
! 
! 3215 2018-08-29 09:58:59Z suehring
! Apply an additional switch controlling the nesting of chemical species
! 
! 3209 2018-08-27 16:58:37Z suehring
! Variable names for nest_bound_x replaced by bc_dirichlet_x.
! Remove commented prints into debug files.
! 
! 3182 2018-07-27 13:36:03Z suehring
! dz was replaced by dz(1)
! 
! 3049 2018-05-29 13:52:36Z Giersch
! Error messages revised
! 
! 3045 2018-05-28 07:55:41Z Giersch
! Error messages revised
! 
! 3022 2018-05-18 11:12:35Z suehring
! Minor fix - working precision added to real number
! 
! 3021 2018-05-16 08:14:20Z maronga
! Bugfix: variable lcr was defined as INTENT(OUT) instead of INTENT(INOUT)
! 
! 3020 2018-05-14 10:45:48Z hellstea
! Bugfix in pmci_define_loglaw_correction_parameters
! 
! 3001 2018-04-20 12:27:13Z suehring
! Bugfix, replace MERGE function by an if-condition in the anterpolation (in 
! order to avoid floating-point exceptions).
! 
! 2997 2018-04-19 13:35:17Z suehring
! Mask topography grid points in anterpolation
! 
! 2984 2018-04-18 11:51:30Z hellstea
! Bugfix in the log-law correction initialization. Pivot node cannot any more be 
! selected from outside the subdomain in order to prevent array under/overflows.
! 
! 2967 2018-04-13 11:22:08Z raasch
! bugfix: missing parallel cpp-directives added
! 
! 2951 2018-04-06 09:05:08Z kanani
! Add log_point_s for pmci_model_configuration
! 
! 2938 2018-03-27 15:52:42Z suehring
! - Nesting for RANS mode implemented
!    - Interpolation of TKE onto child domain only if both parent and child are 
!      either in LES mode or in RANS mode
!    - Interpolation of dissipation if both parent and child are in RANS mode
!      and TKE-epsilon closure is applied
!    - Enable anterpolation of TKE and dissipation rate in case parent and 
!      child operate in RANS mode
!
! - Some unused variables removed from ONLY list
! - Some formatting adjustments for particle nesting
! 
! 2936 2018-03-27 14:49:27Z suehring
! Control logics improved to allow nesting also in cases with
! constant_flux_layer = .F. or constant_diffusion = .T.
! 
! 2895 2018-03-15 10:26:12Z hellstea
! Change in the nest initialization (pmci_interp_tril_all). Bottom wall BC is no
! longer overwritten.
! 
! 2868 2018-03-09 13:25:09Z hellstea
! Local conditional Neumann conditions for one-way coupling removed.  
! 
! 2853 2018-03-05 14:44:20Z suehring
! Bugfix in init log-law correction.
! 
! 2841 2018-02-27 15:02:57Z knoop
! Bugfix: wrong placement of include 'mpif.h' corrected
! 
! 2812 2018-02-16 13:40:25Z hellstea
! Bugfixes in computation of the interpolation loglaw-correction parameters
! 
! 2809 2018-02-15 09:55:58Z schwenkel
! Bugfix for gfortran: Replace the function C_SIZEOF with STORAGE_SIZE
! 
! 2806 2018-02-14 17:10:37Z thiele
! Bugfixing Write statements
! 
! 2804 2018-02-14 16:57:03Z thiele
! preprocessor directive for c_sizeof in case of __gfortran added
! 
! 2803 2018-02-14 16:56:32Z thiele
! Introduce particle transfer in nested models.
! 
! 2795 2018-02-07 14:48:48Z hellstea
! Bugfix in computation of the anterpolation under-relaxation functions. 
! 
! 2773 2018-01-30 14:12:54Z suehring
! - Nesting for chemical species
! - Bugfix in setting boundary condition at downward-facing walls for passive
!   scalar
! - Some formatting adjustments
! 
! 2718 2018-01-02 08:49:38Z maronga
! Corrected "Former revisions" section
! 
! 2701 2017-12-15 15:40:50Z suehring
! Changes from last commit documented
! 
! 2698 2017-12-14 18:46:24Z suehring
! Bugfix in get_topography_top_index
!
! 2696 2017-12-14 17:12:51Z kanani
! Change in file header (GPL part)
! - Bugfix in init_tke_factor (MS)
! 
! 2669 2017-12-06 16:03:27Z raasch
! file extension for nested domains changed to "_N##",
! created flag file in order to enable combine_plot_fields to process nest data
! 
! 2663 2017-12-04 17:40:50Z suehring
! Bugfix, wrong coarse-grid index in init_tkefactor used.
! 
! 2602 2017-11-03 11:06:41Z hellstea
! Index-limit related bug (occurred with nesting_mode='vertical') fixed in
! pmci_interp_tril_t. Check for too high nest domain added in pmci_setup_parent.   
! Some cleaning up made. 
! 
! 2582 2017-10-26 13:19:46Z hellstea
! Resetting of e within buildings / topography in pmci_parent_datatrans removed
! as unnecessary since e is not anterpolated, and as incorrect since it overran
! the default Neumann condition (bc_e_b). 
! 
! 2359 2017-08-21 07:50:30Z hellstea
! A minor indexing error in pmci_init_loglaw_correction is corrected. 
! 
! 2351 2017-08-15 12:03:06Z kanani
! Removed check (PA0420) for nopointer version
! 
! 2350 2017-08-15 11:48:26Z kanani
! Bugfix and error message for nopointer version.
! 
! 2318 2017-07-20 17:27:44Z suehring
! Get topography top index via Function call 
! 
! 2317 2017-07-20 17:27:19Z suehring
! Set bottom boundary condition after anterpolation.
! Some variable description added.
! 
! 2293 2017-06-22 12:59:12Z suehring
! In anterpolation, exclude grid points which are used for interpolation. 
! This avoids the accumulation of numerical errors leading to increased 
! variances for shallow child domains.  
! 
! 2292 2017-06-20 09:51:42Z schwenkel
! Implementation of new microphysic scheme: cloud_scheme = 'morrison' 
! includes two more prognostic equations for cloud drop concentration (nc)  
! and cloud water content (qc). 
! 
! 2285 2017-06-15 13:15:41Z suehring
! Consider multiple pure-vertical nest domains in overlap check
! 
! 2283 2017-06-14 10:17:34Z suehring
! Bugfixes in inititalization of log-law correction concerning 
! new topography concept
! 
! 2281 2017-06-13 11:34:50Z suehring
! Bugfix, add pre-preprocessor directives to enable non-parrallel mode
! 
! 2241 2017-06-01 13:46:13Z hellstea
! A minor indexing error in pmci_init_loglaw_correction is corrected.
! 
! 2240 2017-06-01 13:45:34Z hellstea
!
! 2232 2017-05-30 17:47:52Z suehring
! Adjustments to new topography concept
! 
! 2229 2017-05-30 14:52:52Z hellstea
! A minor indexing error in pmci_init_anterp_tophat is corrected.
! 
! 2174 2017-03-13 08:18:57Z maronga
! Added support for cloud physics quantities, syntax layout improvements. Data
! transfer of qc and nc is prepared but currently deactivated until both
! quantities become prognostic variables.
! Some bugfixes.
! 
! 2019 2016-09-30 13:40:09Z hellstea
! Bugfixes mainly in determining the anterpolation index bounds. These errors 
! were detected when first time tested using 3:1 grid-spacing.
! 
! 2003 2016-08-24 10:22:32Z suehring
! Humidity and passive scalar also separated in nesting mode 
! 
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1938 2016-06-13 15:26:05Z hellstea
! Minor clean-up.
! 
! 1901 2016-05-04 15:39:38Z raasch
! Initial version of purely vertical nesting introduced.
! Code clean up. The words server/client changed to parent/child.
!
! 1900 2016-05-04 15:27:53Z raasch
! unused variables removed
!
! 1894 2016-04-27 09:01:48Z raasch
! bugfix: pt interpolations are omitted in case that the temperature equation is
! switched off
!
! 1892 2016-04-26 13:49:47Z raasch
! bugfix: pt is not set as a data array in case that the temperature equation is
! switched off with neutral = .TRUE.
!
! 1882 2016-04-20 15:24:46Z hellstea
! The factor ijfc for nfc used in anterpolation is redefined as 2-D array 
! and precomputed in pmci_init_anterp_tophat. 
! 
! 1878 2016-04-19 12:30:36Z hellstea
! Synchronization rewritten, logc-array index order changed for cache
! optimization
!
! 1850 2016-04-08 13:29:27Z maronga
! Module renamed
!
!
! 1808 2016-04-05 19:44:00Z raasch
! MPI module used by default on all machines
!
! 1801 2016-04-05 13:12:47Z raasch
! bugfix for r1797: zero setting of temperature within topography does not work
! and has been disabled
!
! 1797 2016-03-21 16:50:28Z raasch
! introduction of different datatransfer modes,
! further formatting cleanup, parameter checks added (including mismatches
! between root and nest model settings),
! +routine pmci_check_setting_mismatches
! comm_world_nesting introduced
!
! 1791 2016-03-11 10:41:25Z raasch
! routine pmci_update_new removed,
! pmc_get_local_model_info renamed pmc_get_model_info, some keywords also
! renamed,
! filling up redundant ghost points introduced,
! some index bound variables renamed,
! further formatting cleanup
!
! 1783 2016-03-06 18:36:17Z raasch
! calculation of nest top area simplified,
! interpolation and anterpolation moved to seperate wrapper subroutines
!
! 1781 2016-03-03 15:12:23Z raasch
! _p arrays are set zero within buildings too, t.._m arrays and respective
! settings within buildings completely removed
!
! 1779 2016-03-03 08:01:28Z raasch
! only the total number of PEs is given for the domains, npe_x and npe_y
! replaced by npe_total, two unused elements removed from array
! parent_grid_info_real,
! array management changed from linked list to sequential loop
!
! 1766 2016-02-29 08:37:15Z raasch
! modifications to allow for using PALM's pointer version,
! +new routine pmci_set_swaplevel
!
! 1764 2016-02-28 12:45:19Z raasch
! +cpl_parent_id,
! cpp-statements for nesting replaced by __parallel statements,
! errors output with message-subroutine,
! index bugfixes in pmci_interp_tril_all,
! some adjustments to PALM style
!
! 1762 2016-02-25 12:31:13Z hellstea
! Initial revision by A. Hellsten
!
! Description:
! ------------
! Domain nesting interface routines. The low-level inter-domain communication   
! is conducted by the PMC-library routines.
!
! @todo Remove array_3d variables from USE statements thate not used in the 
!       routine
! @todo Data transfer of qc and nc is prepared but not activated
!------------------------------------------------------------------------------!
 MODULE pmc_interface

    USE ISO_C_BINDING


    USE arrays_3d,                                                             &
        ONLY:  diss, diss_2, dzu, dzw, e, e_p, e_2, nc, nc_2, nc_p, nr, nr_2,  &
               pt, pt_2, q, q_2, qc, qc_2, qr, qr_2, s, s_2,                   &
               u, u_p, u_2, v, v_p, v_2, w, w_p, w_2, zu, zw

    USE control_parameters,                                                    &
        ONLY:  air_chemistry, bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,  &
               bc_dirichlet_s, child_domain,                                   &
               constant_diffusion, constant_flux_layer,                        &
               coupling_char, dt_3d, dz, humidity, message_string,             &
               neutral, passive_scalar, rans_mode, rans_tke_e,                 &
               roughness_length, topography, volume_flow

    USE chem_modules,                                                          &
        ONLY:  nspec

    USE chemistry_model_mod,                                                   &
        ONLY:  chem_species, nest_chemistry, spec_conc_2

    USE cpulog,                                                                &
        ONLY:  cpu_log, log_point_s

    USE grid_variables,                                                        &
        ONLY:  dx, dy

    USE indices,                                                               &
        ONLY:  nbgp, nx, nxl, nxlg, nxlu, nxr, nxrg, ny, nyn, nyng, nys, nysg, &
               nysv, nz, nzb, nzt, wall_flags_0

    USE bulk_cloud_model_mod,                                                  &
        ONLY: bulk_cloud_model, microphysics_morrison, microphysics_seifert

    USE particle_attributes,                                                   &
        ONLY:  particle_advection

    USE kinds

#if defined( __parallel )
#if !defined( __mpifh )
    USE MPI
#endif

    USE pegrid,                                                                &
        ONLY:  collective_wait, comm1dx, comm1dy, comm2d, myid, myidx, myidy,  &
               numprocs, pdims, pleft, pnorth, pright, psouth, status

    USE pmc_child,                                                             &
        ONLY:  pmc_childinit, pmc_c_clear_next_array_list,                     &
               pmc_c_getnextarray, pmc_c_get_2d_index_list, pmc_c_getbuffer,   &
               pmc_c_putbuffer, pmc_c_setind_and_allocmem,                     &
               pmc_c_set_dataarray, pmc_set_dataarray_name

    USE pmc_general,                                                           &
        ONLY:  da_namelen

    USE pmc_handle_communicator,                                               &
        ONLY:  pmc_get_model_info, pmc_init_model, pmc_is_rootmodel,           &
               pmc_no_namelist_found, pmc_parent_for_child, m_couplers

    USE pmc_mpi_wrapper,                                                       &
        ONLY:  pmc_bcast, pmc_recv_from_child, pmc_recv_from_parent,           &
               pmc_send_to_child, pmc_send_to_parent

    USE pmc_parent,                                                            &
        ONLY:  pmc_parentinit, pmc_s_clear_next_array_list, pmc_s_fillbuffer,  &
               pmc_s_getdata_from_buffer, pmc_s_getnextarray,                  &
               pmc_s_setind_and_allocmem, pmc_s_set_active_data_array,         &
               pmc_s_set_dataarray, pmc_s_set_2d_index_list

#endif

    USE surface_mod,                                                           &
        ONLY:  get_topography_top_index_ji, surf_def_h, surf_lsm_h, surf_usm_h

    IMPLICIT NONE

#if defined( __parallel )
#if defined( __mpifh )
    INCLUDE "mpif.h"
#endif
#endif

    PRIVATE
!
!-- Constants
    INTEGER(iwp), PARAMETER ::  child_to_parent = 2   !<
    INTEGER(iwp), PARAMETER ::  parent_to_child = 1   !<
!
!-- Coupler setup
    INTEGER(iwp), SAVE      ::  comm_world_nesting    !<
    INTEGER(iwp), SAVE      ::  cpl_id  = 1           !<
    CHARACTER(LEN=32), SAVE ::  cpl_name              !<
    INTEGER(iwp), SAVE      ::  cpl_npe_total         !<
    INTEGER(iwp), SAVE      ::  cpl_parent_id         !<
!
!-- Control parameters
    CHARACTER(LEN=7), SAVE ::  nesting_datatransfer_mode = 'mixed'  !< steering
                                                         !< parameter for data-
                                                         !< transfer mode
    CHARACTER(LEN=8), SAVE ::  nesting_mode = 'two-way'  !< steering parameter
                                                         !< for 1- or 2-way nesting

    LOGICAL, SAVE ::  nested_run = .FALSE.  !< general switch
    LOGICAL       ::  rans_mode_parent = .FALSE. !< mode of parent model (.F. - LES mode, .T. - RANS mode)

    REAL(wp), SAVE ::  anterp_relax_length_l = -1.0_wp   !<
    REAL(wp), SAVE ::  anterp_relax_length_r = -1.0_wp   !<
    REAL(wp), SAVE ::  anterp_relax_length_s = -1.0_wp   !<
    REAL(wp), SAVE ::  anterp_relax_length_n = -1.0_wp   !<
    REAL(wp), SAVE ::  anterp_relax_length_t = -1.0_wp   !<
!
!-- Geometry
    REAL(wp), SAVE, DIMENSION(:), ALLOCATABLE, PUBLIC ::  coord_x            !<
    REAL(wp), SAVE, DIMENSION(:), ALLOCATABLE, PUBLIC ::  coord_y            !<
    REAL(wp), SAVE, PUBLIC                            ::  lower_left_coord_x !<
    REAL(wp), SAVE, PUBLIC                            ::  lower_left_coord_y !<

!
!-- Children's parent-grid arrays
    INTEGER(iwp), SAVE, DIMENSION(5), PUBLIC    ::  coarse_bound        !< subdomain index bounds for children's parent-grid arrays
    INTEGER(iwp), SAVE, DIMENSION(4), PUBLIC    ::  coarse_bound_aux    !< subdomain index bounds for allocation of index-mapping and other auxiliary arrays
    INTEGER(iwp), SAVE, DIMENSION(4), PUBLIC    ::  coarse_bound_w      !< subdomain index bounds for children's parent-grid work arrays

    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  dissc !< coarse grid array on child domain - dissipation rate
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ec   !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ptc  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  uc   !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  vc   !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  wc   !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_c  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qcc  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qrc  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nrc  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ncc  !<
    REAL(wp), SAVE, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sc   !<
    INTEGER(idp), SAVE, DIMENSION(:,:), ALLOCATABLE, TARGET, PUBLIC ::  nr_partc    !<
    INTEGER(idp), SAVE, DIMENSION(:,:), ALLOCATABLE, TARGET, PUBLIC ::  part_adrc   !<

    REAL(wp), SAVE, DIMENSION(:,:,:,:), ALLOCATABLE, TARGET ::  chem_spec_c !< coarse grid array on child domain - chemical species

!
!-- Child interpolation coefficients and child-array indices to be 
!-- precomputed and stored.
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  ico    !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  icu    !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jco    !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jcv    !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kco    !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kcw    !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  celltmpd !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1xo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2xo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1xu   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2xu   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1yo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2yo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1yv   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2yv   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1zo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2zo   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r1zw   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:)     ::  r2zw   !<

!
!-- Child index arrays and log-ratio arrays for the log-law near-wall
!-- corrections. These are not truly 3-D arrays but multiple 2-D arrays.
    INTEGER(iwp), SAVE :: ncorr  !< 4th dimension of the log_ratio-arrays
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_u_l   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_u_n   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_u_r   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_u_s   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_v_l   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_v_n   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_v_r   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_v_s   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_w_l   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_w_n   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_w_r   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  logc_w_s   !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_u_l !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_u_n !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_u_r !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_u_s !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_v_l !<    
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_v_n !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_v_r !<    
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_v_s !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_w_l !<    
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_w_n !<
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_w_r !<    
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:)   ::  logc_kbounds_w_s !<        
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_u_l   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_u_n   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_u_r   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_u_s   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_v_l   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_v_n   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_v_r   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_v_s   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_w_l   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_w_n   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_w_r   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:,:,:,:)   ::  logc_ratio_w_s   !<
!
    INTEGER(iwp), SAVE ::  igsr     !< Integer grid-spacing ratio in i-direction 
    INTEGER(iwp), SAVE ::  jgsr     !< Integer grid-spacing ratio in j-direction 
    INTEGER(iwp), SAVE ::  kgsr     !< Integer grid-spacing ratio in k-direction 
    INTEGER(iwp), SAVE ::  kcto     !< Upper bound for k in anterpolation of variables other than w.
    INTEGER(iwp), SAVE ::  kctw     !< Upper bound for k in anterpolation of w.
    INTEGER(iwp), SAVE ::  nxlfc    !< Lower index limit in x-direction for fine-to-coarse index mapping and interpolaton coefficient arrays
    INTEGER(iwp), SAVE ::  nxrfc    !< Upper index limit in x-direction for fine-to-coarse index mapping and interpolaton coefficient arrays
    INTEGER(iwp), SAVE ::  nynfc    !< Upper index limit in y-direction for fine-to-coarse index mapping and interpolaton coefficient arrays
    INTEGER(iwp), SAVE ::  nysfc    !< Lower index limit in y-direction for fine-to-coarse index mapping and interpolaton coefficient arrays
!
!-- Upper bound for k in log-law correction in interpolation.
    INTEGER(iwp), SAVE ::  nzt_topo_nestbc_l   !<
    INTEGER(iwp), SAVE ::  nzt_topo_nestbc_n   !<
    INTEGER(iwp), SAVE ::  nzt_topo_nestbc_r   !<
    INTEGER(iwp), SAVE ::  nzt_topo_nestbc_s   !<
!
!-- Spatial under-relaxation coefficients for anterpolation.
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:) ::  frax   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:) ::  fray   !<
    REAL(wp), SAVE, ALLOCATABLE, DIMENSION(:) ::  fraz   !<
!
!-- Child-array indices to be precomputed and stored for anterpolation.
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  iflu   !< child index indicating left bound of parent grid box on u-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  ifuu   !< child index indicating right bound of parent grid box on u-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  iflo   !< child index indicating left bound of parent grid box on scalar-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  ifuo   !< child index indicating right bound of parent grid box on scalar-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jflv   !< child index indicating south bound of parent grid box on v-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jfuv   !< child index indicating north bound of parent grid box on v-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jflo   !< child index indicating south bound of parent grid box on scalar-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  jfuo   !< child index indicating north bound of parent grid box on scalar-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kflw   !< child index indicating lower bound of parent grid box on w-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kfuw   !< child index indicating upper bound of parent grid box on w-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kflo   !< child index indicating lower bound of parent grid box on scalar-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:) ::  kfuo   !< child index indicating upper bound of parent grid box on scalar-grid
!
!-- Number of fine-grid nodes inside coarse-grid ij-faces
!-- to be precomputed for anterpolation.
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  ijkfc_u  !< number of child grid boxes contribution to a parent grid box, u-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  ijkfc_v  !< number of child grid boxes contribution to a parent grid box, v-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  ijkfc_w  !< number of child grid boxes contribution to a parent grid box, w-grid
    INTEGER(iwp), SAVE, ALLOCATABLE, DIMENSION(:,:,:) ::  ijkfc_s  !< number of child grid boxes contribution to a parent grid box, scalar-grid
    
    INTEGER(iwp), DIMENSION(3)          ::  parent_grid_info_int    !<
    REAL(wp), DIMENSION(7)              ::  parent_grid_info_real   !<
    REAL(wp), DIMENSION(2)              ::  zmax_coarse             !<

    TYPE coarsegrid_def
       INTEGER(iwp)                        ::  nx                 !<
       INTEGER(iwp)                        ::  ny                 !<
       INTEGER(iwp)                        ::  nz                 !<
       REAL(wp)                            ::  dx                 !<
       REAL(wp)                            ::  dy                 !<
       REAL(wp)                            ::  dz                 !<
       REAL(wp)                            ::  lower_left_coord_x !<
       REAL(wp)                            ::  lower_left_coord_y !<
       REAL(wp)                            ::  xend               !<
       REAL(wp)                            ::  yend               !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  coord_x            !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  coord_y            !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  dzu                !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  dzw                !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  zu                 !<
       REAL(wp), DIMENSION(:), ALLOCATABLE ::  zw                 !<
    END TYPE coarsegrid_def

    TYPE(coarsegrid_def), SAVE, PUBLIC     ::  cg   !<
!
!-- Variables for particle coupling
    TYPE, PUBLIC :: childgrid_def
       INTEGER(iwp)                        ::  nx                   !<
       INTEGER(iwp)                        ::  ny                   !<
       INTEGER(iwp)                        ::  nz                   !<
       REAL(wp)                            ::  dx                   !<
       REAL(wp)                            ::  dy                   !<
       REAL(wp)                            ::  dz                   !<
       REAL(wp)                            ::  lx_coord, lx_coord_b !<
       REAL(wp)                            ::  rx_coord, rx_coord_b !<
       REAL(wp)                            ::  sy_coord, sy_coord_b !<
       REAL(wp)                            ::  ny_coord, ny_coord_b !<
       REAL(wp)                            ::  uz_coord, uz_coord_b !<
    END TYPE childgrid_def

    TYPE(childgrid_def), SAVE, ALLOCATABLE, DIMENSION(:), PUBLIC :: childgrid !<

    INTEGER(idp),ALLOCATABLE,DIMENSION(:,:),PUBLIC,TARGET    :: nr_part  !<
    INTEGER(idp),ALLOCATABLE,DIMENSION(:,:),PUBLIC,TARGET    :: part_adr !<

!AH 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarr_lr
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarr_sn
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarr_t
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarrc_lr
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarrc_sn
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: workarrc_t
    INTEGER(iwp) :: workarrc_lr_exchange_type
    INTEGER(iwp) :: workarrc_sn_exchange_type
    INTEGER(iwp) :: workarrc_t_exchange_type_x
    INTEGER(iwp) :: workarrc_t_exchange_type_y
!AH
    
    INTERFACE pmci_boundary_conds
       MODULE PROCEDURE pmci_boundary_conds
    END INTERFACE pmci_boundary_conds
    
    INTERFACE pmci_check_setting_mismatches
       MODULE PROCEDURE pmci_check_setting_mismatches
    END INTERFACE

    INTERFACE pmci_child_initialize
       MODULE PROCEDURE pmci_child_initialize
    END INTERFACE

    INTERFACE pmci_synchronize
       MODULE PROCEDURE pmci_synchronize
    END INTERFACE

    INTERFACE pmci_datatrans
       MODULE PROCEDURE pmci_datatrans
    END INTERFACE pmci_datatrans

    INTERFACE pmci_init
       MODULE PROCEDURE pmci_init
    END INTERFACE

    INTERFACE pmci_modelconfiguration
       MODULE PROCEDURE pmci_modelconfiguration
    END INTERFACE

    INTERFACE pmci_parent_initialize
       MODULE PROCEDURE pmci_parent_initialize
    END INTERFACE

    INTERFACE get_number_of_childs
       MODULE PROCEDURE get_number_of_childs
    END  INTERFACE get_number_of_childs

    INTERFACE get_childid
       MODULE PROCEDURE get_childid
    END  INTERFACE get_childid

    INTERFACE get_child_edges
       MODULE PROCEDURE get_child_edges
    END  INTERFACE get_child_edges

    INTERFACE get_child_gridspacing
       MODULE PROCEDURE get_child_gridspacing
    END  INTERFACE get_child_gridspacing


    INTERFACE pmci_set_swaplevel
       MODULE PROCEDURE pmci_set_swaplevel
    END INTERFACE pmci_set_swaplevel

    PUBLIC anterp_relax_length_l, anterp_relax_length_r,                       &
           anterp_relax_length_s, anterp_relax_length_n,                       &
           anterp_relax_length_t, child_to_parent, comm_world_nesting,         &
           cpl_id, nested_run, nesting_datatransfer_mode, nesting_mode,        &
           parent_to_child, rans_mode_parent

    PUBLIC pmci_boundary_conds
    PUBLIC pmci_child_initialize
    PUBLIC pmci_datatrans
    PUBLIC pmci_init
    PUBLIC pmci_modelconfiguration
    PUBLIC pmci_parent_initialize
    PUBLIC pmci_synchronize
    PUBLIC pmci_set_swaplevel
    PUBLIC get_number_of_childs, get_childid, get_child_edges, get_child_gridspacing



 CONTAINS


 SUBROUTINE pmci_init( world_comm )

    USE control_parameters,                                                    &
        ONLY:  message_string

    IMPLICIT NONE

    INTEGER(iwp), INTENT(OUT) ::  world_comm   !<

#if defined( __parallel )

    INTEGER(iwp) ::  pmc_status   !<


    CALL pmc_init_model( world_comm, nesting_datatransfer_mode, nesting_mode,  &
                         pmc_status )

    IF ( pmc_status == pmc_no_namelist_found )  THEN
!
!--    This is not a nested run
       world_comm = MPI_COMM_WORLD
       cpl_id     = 1
       cpl_name   = ""

       RETURN

    ENDIF
!
!-- Check steering parameter values
    IF ( TRIM( nesting_mode ) /= 'one-way'  .AND.                              &
         TRIM( nesting_mode ) /= 'two-way'  .AND.                              &
         TRIM( nesting_mode ) /= 'vertical' )                                  &
    THEN
       message_string = 'illegal nesting mode: ' // TRIM( nesting_mode )
       CALL message( 'pmci_init', 'PA0417', 3, 2, 0, 6, 0 )
    ENDIF

    IF ( TRIM( nesting_datatransfer_mode ) /= 'cascade'  .AND.                 &
         TRIM( nesting_datatransfer_mode ) /= 'mixed'    .AND.                 &
         TRIM( nesting_datatransfer_mode ) /= 'overlap' )                      &
    THEN
       message_string = 'illegal nesting datatransfer mode: '                  &
                        // TRIM( nesting_datatransfer_mode )
       CALL message( 'pmci_init', 'PA0418', 3, 2, 0, 6, 0 )
    ENDIF
!
!-- Set the general steering switch which tells PALM that its a nested run
    nested_run = .TRUE.
!
!-- Get some variables required by the pmc-interface (and in some cases in the
!-- PALM code out of the pmci) out of the pmc-core
    CALL pmc_get_model_info( comm_world_nesting = comm_world_nesting,          &
                             cpl_id = cpl_id, cpl_parent_id = cpl_parent_id,   &
                             cpl_name = cpl_name, npe_total = cpl_npe_total,   &
                             lower_left_x = lower_left_coord_x,                &
                             lower_left_y = lower_left_coord_y )
!
!-- Set the steering switch which tells the models that they are nested (of
!-- course the root domain (cpl_id = 1) is not nested)
    IF ( cpl_id >= 2 )  THEN
       child_domain = .TRUE.
       WRITE( coupling_char, '(A2,I2.2)') '_N', cpl_id
    ENDIF

!
!-- Message that communicators for nesting are initialized.
!-- Attention: myid has been set at the end of pmc_init_model in order to
!-- guarantee that only PE0 of the root domain does the output.
    CALL location_message( 'finished', .TRUE. )
!
!-- Reset myid to its default value
    myid = 0
#else
!
!-- Nesting cannot be used in serial mode. cpl_id is set to root domain (1)
!-- because no location messages would be generated otherwise.
!-- world_comm is given a dummy value to avoid compiler warnings (INTENT(OUT)
!-- should get an explicit value)
    cpl_id     = 1
    nested_run = .FALSE.
    world_comm = 1
#endif

 END SUBROUTINE pmci_init



 SUBROUTINE pmci_modelconfiguration

    IMPLICIT NONE

    INTEGER(iwp) ::  ncpl   !<  number of nest domains

#if defined( __parallel )
    CALL location_message( 'setup the nested model configuration', .FALSE. )
    CALL cpu_log( log_point_s(79), 'pmci_model_config', 'start' )
!
!-- Compute absolute coordinates for all models
    CALL pmci_setup_coordinates
!
!-- Determine the number of coupled arrays
    CALL pmci_num_arrays
!
!-- Initialize the child (must be called before pmc_setup_parent)
    CALL pmci_setup_child
!
!-- Initialize PMC parent
    CALL pmci_setup_parent
!
!-- Check for mismatches between settings of master and child variables
!-- (e.g., all children have to follow the end_time settings of the root master)
    CALL pmci_check_setting_mismatches
!
!-- Set flag file for combine_plot_fields for precessing the nest output data
    OPEN( 90, FILE='3DNESTING', FORM='FORMATTED' )
    CALL pmc_get_model_info( ncpl = ncpl )
    WRITE( 90, '(I2)' )  ncpl
    CLOSE( 90 )

    CALL cpu_log( log_point_s(79), 'pmci_model_config', 'stop' )
    CALL location_message( 'finished', .TRUE. )
#endif

 END SUBROUTINE pmci_modelconfiguration



 SUBROUTINE pmci_setup_parent

#if defined( __parallel )
    IMPLICIT NONE

    CHARACTER(LEN=32) ::  myname
    
    INTEGER(iwp) ::  child_id         !<
    INTEGER(iwp) ::  ierr             !<
    INTEGER(iwp) ::  k                !<
    INTEGER(iwp) ::  m                !<
    INTEGER(iwp) ::  mid              !<
    INTEGER(iwp) ::  mm               !<
    INTEGER(iwp) ::  n = 1            !< running index for chemical species
    INTEGER(iwp) ::  nest_overlap     !<
    INTEGER(iwp) ::  nomatch          !<
    INTEGER(iwp) ::  nx_cl            !<
    INTEGER(iwp) ::  ny_cl            !<
    INTEGER(iwp) ::  nz_cl            !<

    INTEGER(iwp), DIMENSION(5) ::  val    !<


    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ch_xl   !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ch_xr   !<    
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ch_ys   !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ch_yn   !<
    REAL(wp) ::  cl_height        !< 
    REAL(wp) ::  dx_cl            !<
    REAL(wp) ::  dy_cl            !<
    REAL(wp) ::  dz_cl            !<
    REAL(wp) ::  left_limit       !<
    REAL(wp) ::  north_limit      !<
    REAL(wp) ::  right_limit      !<
    REAL(wp) ::  south_limit      !<
    REAL(wp) ::  xez              !<
    REAL(wp) ::  yez              !<

    REAL(wp), DIMENSION(5) ::  fval             !<

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  cl_coord_x   !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  cl_coord_y   !<

!
!   Initialize the pmc parent
    CALL pmc_parentinit

!
!-- Corners of all children of the present parent
    IF ( ( SIZE( pmc_parent_for_child ) - 1 > 0 ) .AND. myid == 0 )  THEN 
       ALLOCATE( ch_xl(1:SIZE( pmc_parent_for_child ) - 1) )
       ALLOCATE( ch_xr(1:SIZE( pmc_parent_for_child ) - 1) )
       ALLOCATE( ch_ys(1:SIZE( pmc_parent_for_child ) - 1) )
       ALLOCATE( ch_yn(1:SIZE( pmc_parent_for_child ) - 1) )
    ENDIF
    IF ( ( SIZE( pmc_parent_for_child ) - 1 > 0 ) )  THEN
       ALLOCATE( childgrid(1:SIZE( pmc_parent_for_child ) - 1) )
    ENDIF

!
!-- Get coordinates from all children
    DO  m = 1, SIZE( pmc_parent_for_child ) - 1

       child_id = pmc_parent_for_child(m)

       IF ( myid == 0 )  THEN

          CALL pmc_recv_from_child( child_id, val,  SIZE(val),  0, 123, ierr )
          CALL pmc_recv_from_child( child_id, fval, SIZE(fval), 0, 124, ierr )
         
          nx_cl     = val(1)
          ny_cl     = val(2)
          dx_cl     = fval(3)
          dy_cl     = fval(4)
          dz_cl     = fval(5)
          cl_height = fval(1)

          nz_cl = nz
!
!--       Find the highest nest level in the parent grid for the reduced z
!--       transfer
          DO  k = 1, nz                 
             IF ( zw(k) > fval(1) )  THEN
                nz_cl = k
                EXIT
             ENDIF
          ENDDO

          zmax_coarse = fval(1:2)
          cl_height   = fval(1)

!    
!--       Get absolute coordinates from the child
          ALLOCATE( cl_coord_x(-nbgp:nx_cl+nbgp) )
          ALLOCATE( cl_coord_y(-nbgp:ny_cl+nbgp) )
         
          CALL pmc_recv_from_child( child_id, cl_coord_x, SIZE( cl_coord_x ),  &
               0, 11, ierr )
          CALL pmc_recv_from_child( child_id, cl_coord_y, SIZE( cl_coord_y ),  &
               0, 12, ierr )
         
          parent_grid_info_real(1) = lower_left_coord_x
          parent_grid_info_real(2) = lower_left_coord_y
          parent_grid_info_real(3) = dx
          parent_grid_info_real(4) = dy
          parent_grid_info_real(5) = lower_left_coord_x + ( nx + 1 ) * dx
          parent_grid_info_real(6) = lower_left_coord_y + ( ny + 1 ) * dy
          parent_grid_info_real(7) = dz(1)

          parent_grid_info_int(1)  = nx
          parent_grid_info_int(2)  = ny
          parent_grid_info_int(3)  = nz_cl
!
!--       Check that the child domain matches parent domain. 
          nomatch = 0
          IF ( nesting_mode == 'vertical' )  THEN
             right_limit = parent_grid_info_real(5)
             north_limit = parent_grid_info_real(6)
             IF ( ( cl_coord_x(nx_cl+1) /= right_limit ) .OR.                  &
                  ( cl_coord_y(ny_cl+1) /= north_limit ) )  THEN
                nomatch = 1
             ENDIF
          ELSE       
!
!--       Check that the child domain is completely inside the parent domain. 
             xez = ( nbgp + 1 ) * dx 
             yez = ( nbgp + 1 ) * dy 
             left_limit  = lower_left_coord_x + xez
             right_limit = parent_grid_info_real(5) - xez
             south_limit = lower_left_coord_y + yez
             north_limit = parent_grid_info_real(6) - yez
             IF ( ( cl_coord_x(0) < left_limit )        .OR.                   &
                  ( cl_coord_x(nx_cl+1) > right_limit ) .OR.                   &
                  ( cl_coord_y(0) < south_limit )       .OR.                   &
                  ( cl_coord_y(ny_cl+1) > north_limit ) )  THEN
                nomatch = 1
             ENDIF
          ENDIF
!             
!--       Child domain must be lower than the parent domain such
!--       that the top ghost layer of the child grid does not exceed 
!--       the parent domain top boundary.

          IF ( cl_height > zw(nz) ) THEN
             nomatch = 1
          ENDIF
!
!--       Check that parallel nest domains, if any, do not overlap.
          nest_overlap = 0
          IF ( SIZE( pmc_parent_for_child ) - 1 > 0 )  THEN
             ch_xl(m) = cl_coord_x(-nbgp)
             ch_xr(m) = cl_coord_x(nx_cl+nbgp)
             ch_ys(m) = cl_coord_y(-nbgp)
             ch_yn(m) = cl_coord_y(ny_cl+nbgp)

             IF ( m > 1 )  THEN
                DO mm = 1, m - 1
                   mid = pmc_parent_for_child(mm)
!
!--                Check only different nest levels
                   IF (m_couplers(child_id)%parent_id /= m_couplers(mid)%parent_id)  THEN
                      IF ( ( ch_xl(m) < ch_xr(mm) .OR.                         &
                             ch_xr(m) > ch_xl(mm) )  .AND.                     &
                           ( ch_ys(m) < ch_yn(mm) .OR.                         &
                             ch_yn(m) > ch_ys(mm) ) )  THEN
                         nest_overlap = 1
                      ENDIF
                   ENDIF
                ENDDO
             ENDIF
          ENDIF

          CALL set_child_edge_coords

          DEALLOCATE( cl_coord_x )
          DEALLOCATE( cl_coord_y )
!
!--       Send information about operating mode (LES or RANS) to child. This will be 
!--       used to control TKE nesting and setting boundary conditions properly.
          CALL pmc_send_to_child( child_id, rans_mode, 1, 0, 19, ierr )  
!
!--       Send coarse grid information to child
          CALL pmc_send_to_child( child_id, parent_grid_info_real,             &
                                   SIZE( parent_grid_info_real ), 0, 21,       &
                                   ierr )
          CALL pmc_send_to_child( child_id, parent_grid_info_int,  3, 0,       &
                                   22, ierr )
!
!--       Send local grid to child
          CALL pmc_send_to_child( child_id, coord_x, nx+1+2*nbgp, 0, 24,       &
                                   ierr )
          CALL pmc_send_to_child( child_id, coord_y, ny+1+2*nbgp, 0, 25,       &
                                   ierr )
!
!--       Also send the dzu-, dzw-, zu- and zw-arrays here
          CALL pmc_send_to_child( child_id, dzu, nz_cl+1, 0, 26, ierr )
          CALL pmc_send_to_child( child_id, dzw, nz_cl+1, 0, 27, ierr )
          CALL pmc_send_to_child( child_id, zu,  nz_cl+2, 0, 28, ierr )
          CALL pmc_send_to_child( child_id, zw,  nz_cl+2, 0, 29, ierr )

       ENDIF

       CALL MPI_BCAST( nomatch, 1, MPI_INTEGER, 0, comm2d, ierr )
       IF ( nomatch /= 0 )  THEN
          WRITE ( message_string, * )  'nested child domain does ',            &
                                       'not fit into its parent domain'
          CALL message( 'pmci_setup_parent', 'PA0425', 3, 2, 0, 6, 0 )
       ENDIF
  
       CALL MPI_BCAST( nest_overlap, 1, MPI_INTEGER, 0, comm2d, ierr )
       IF ( nest_overlap /= 0  .AND.  nesting_mode /= 'vertical' )  THEN
          WRITE ( message_string, * )  'nested parallel child domains overlap'
          CALL message( 'pmci_setup_parent', 'PA0426', 3, 2, 0, 6, 0 )
       ENDIF
      
       CALL MPI_BCAST( nz_cl, 1, MPI_INTEGER, 0, comm2d, ierr )

       CALL MPI_BCAST( childgrid(m), STORAGE_SIZE(childgrid(1))/8, MPI_BYTE, 0, comm2d, ierr )
!
!--    TO_DO: Klaus: please give a comment what is done here
       CALL pmci_create_index_list
!
!--    Include couple arrays into parent content
!--    The adresses of the PALM 2D or 3D array (here server coarse grid) which are candidates
!--    for coupling are stored once into the pmc context. While data transfer, the array do not
!--    have to be specified again

       CALL pmc_s_clear_next_array_list
       DO  WHILE ( pmc_s_getnextarray( child_id, myname ) )
          IF ( INDEX( TRIM( myname ), 'chem_' ) /= 0 )  THEN             
             CALL pmci_set_array_pointer( myname, child_id = child_id,         &
                                          nz_cl = nz_cl, n = n )
             n = n + 1
          ELSE
             CALL pmci_set_array_pointer( myname, child_id = child_id,         &
                                          nz_cl = nz_cl )
          ENDIF
       ENDDO

       CALL pmc_s_setind_and_allocmem( child_id )
    ENDDO

    IF ( ( SIZE( pmc_parent_for_child ) - 1 > 0 ) .AND. myid == 0 )  THEN 
       DEALLOCATE( ch_xl )
       DEALLOCATE( ch_xr )
       DEALLOCATE( ch_ys )
       DEALLOCATE( ch_yn )
    ENDIF

 CONTAINS


   SUBROUTINE pmci_create_index_list

       IMPLICIT NONE

       INTEGER(iwp) ::  i                  !<
       INTEGER(iwp) ::  ic                 !<
       INTEGER(iwp) ::  ierr               !<
       INTEGER(iwp) ::  j                  !<
       INTEGER(iwp) ::  k                  !<
       INTEGER(iwp) ::  npx                !<
       INTEGER(iwp) ::  npy                !<
       INTEGER(iwp) ::  nrx                !<
       INTEGER(iwp) ::  nry                !<
       INTEGER(iwp) ::  px                 !<
       INTEGER(iwp) ::  py                 !<
       INTEGER(iwp) ::  parent_pe          !<

       INTEGER(iwp), DIMENSION(2) ::  scoord             !<
       INTEGER(iwp), DIMENSION(2) ::  size_of_array      !<

       INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE  ::  coarse_bound_all   !<
       INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE  ::  index_list         !<

       IF ( myid == 0 )  THEN
!          
!--       TO_DO: Klaus: give more specific comment what size_of_array stands for
          CALL pmc_recv_from_child( child_id, size_of_array, 2, 0, 40, ierr )
          ALLOCATE( coarse_bound_all(size_of_array(1),size_of_array(2)) )
          CALL pmc_recv_from_child( child_id, coarse_bound_all,                &
                                    SIZE( coarse_bound_all ), 0, 41, ierr )
!
!--       Compute size of index_list.
          ic = 0         
          DO  k = 1, size_of_array(2)
             ic = ic + ( coarse_bound_all(4,k) - coarse_bound_all(3,k) + 1 ) * &
                       ( coarse_bound_all(2,k) - coarse_bound_all(1,k) + 1 )
          ENDDO

          ALLOCATE( index_list(6,ic) )

          CALL MPI_COMM_SIZE( comm1dx, npx, ierr )
          CALL MPI_COMM_SIZE( comm1dy, npy, ierr )
!
!--       Nrx is the same for all PEs and so is nry, thus there is no need to compute
!--       them separately for each PE.
          nrx = nxr - nxl + 1
          nry = nyn - nys + 1
          ic  = 0
!
!--       Loop over all children PEs
          DO  k = 1, size_of_array(2)
!
!--          Area along y required by actual child PE
             DO  j = coarse_bound_all(3,k), coarse_bound_all(4,k)  !: j = jcs, jcn of PE# k 
!
!--             Area along x required by actual child PE
                DO  i = coarse_bound_all(1,k), coarse_bound_all(2,k)  !: i = icl, icr of PE# k 

                   px = i / nrx
                   py = j / nry
                   scoord(1) = px
                   scoord(2) = py
                   CALL MPI_CART_RANK( comm2d, scoord, parent_pe, ierr )
                  
                   ic = ic + 1
!
!--                First index in parent array
                   index_list(1,ic) = i - ( px * nrx ) + 1 + nbgp
!
!--                Second index in parent array
                   index_list(2,ic) = j - ( py * nry ) + 1 + nbgp
!
!--                x index of child coarse grid
                   index_list(3,ic) = i - coarse_bound_all(1,k) + 1
!
!--                y index of child coarse grid
                   index_list(4,ic) = j - coarse_bound_all(3,k) + 1
!
!--                PE number of child
                   index_list(5,ic) = k - 1
!
!--                PE number of parent
                   index_list(6,ic) = parent_pe

                ENDDO
             ENDDO
          ENDDO
!
!--       TO_DO: Klaus: comment what is done here
          CALL pmc_s_set_2d_index_list( child_id, index_list(:,1:ic) )

       ELSE
!
!--       TO_DO: Klaus: comment why this dummy allocation is required
          ALLOCATE( index_list(6,1) )
          CALL pmc_s_set_2d_index_list( child_id, index_list )
       ENDIF

       DEALLOCATE(index_list)

     END SUBROUTINE pmci_create_index_list

     SUBROUTINE set_child_edge_coords
        IMPLICIT  NONE

        INTEGER(iwp) :: nbgp_lpm = 1

        nbgp_lpm = min(nbgp_lpm, nbgp)

        childgrid(m)%nx = nx_cl
        childgrid(m)%ny = ny_cl
        childgrid(m)%nz = nz_cl
        childgrid(m)%dx = dx_cl
        childgrid(m)%dy = dy_cl
        childgrid(m)%dz = dz_cl

        childgrid(m)%lx_coord   = cl_coord_x(0)
        childgrid(m)%lx_coord_b = cl_coord_x(-nbgp_lpm)
        childgrid(m)%rx_coord   = cl_coord_x(nx_cl)+dx_cl
        childgrid(m)%rx_coord_b = cl_coord_x(nx_cl+nbgp_lpm)+dx_cl
        childgrid(m)%sy_coord   = cl_coord_y(0)
        childgrid(m)%sy_coord_b = cl_coord_y(-nbgp_lpm)
        childgrid(m)%ny_coord   = cl_coord_y(ny_cl)+dy_cl
        childgrid(m)%ny_coord_b = cl_coord_y(ny_cl+nbgp_lpm)+dy_cl
        childgrid(m)%uz_coord   = zmax_coarse(2)
        childgrid(m)%uz_coord_b = zmax_coarse(1)

     END SUBROUTINE set_child_edge_coords

#endif
 END SUBROUTINE pmci_setup_parent



 SUBROUTINE pmci_setup_child


#if defined( __parallel )
    IMPLICIT NONE

    CHARACTER(LEN=da_namelen) ::  myname     !<

    INTEGER(iwp) ::  i          !<
    INTEGER(iwp) ::  ierr       !< 
    INTEGER(iwp) ::  icl        !< Left index limit for children's parent-grid arrays
    INTEGER(iwp) ::  icla       !< Left index limit for allocation of index-mapping and other auxiliary arrays
    INTEGER(iwp) ::  iclw       !< Left index limit for children's parent-grid work arrays
    INTEGER(iwp) ::  icr        !< Left index limit for children's parent-grid arrays
    INTEGER(iwp) ::  icra       !< Right index limit for allocation of index-mapping and other auxiliary arrays
    INTEGER(iwp) ::  icrw       !< Right index limit for children's parent-grid work arrays
    INTEGER(iwp) ::  j          !< 
    INTEGER(iwp) ::  jcn        !< North index limit for children's parent-grid arrays
    INTEGER(iwp) ::  jcna       !< North index limit for allocation of index-mapping and other auxiliary arrays
    INTEGER(iwp) ::  jcnw       !< North index limit for children's parent-grid work arrays
    INTEGER(iwp) ::  jcs        !< South index limit for children's parent-grid arrays
    INTEGER(iwp) ::  jcsa       !< South index limit for allocation of index-mapping and other auxiliary arrays
    INTEGER(iwp) ::  jcsw       !< South index limit for children's parent-grid work arrays
    INTEGER(iwp) ::  n          !< Running index for number of chemical species

    INTEGER(iwp), DIMENSION(5) ::  val        !<
    
    REAL(wp) ::  xcs        !<
    REAL(wp) ::  xce        !<
    REAL(wp) ::  ycs        !<
    REAL(wp) ::  yce        !<

    REAL(wp), DIMENSION(5) ::  fval       !<
                                             
!
!-- Child setup
!-- Root model does not have a parent and is not a child, therefore no child setup on root model

    IF ( .NOT. pmc_is_rootmodel() )  THEN

       CALL pmc_childinit
!
!--    Here AND ONLY HERE the arrays are defined, which actualy will be
!--    exchanged between child and parent.
!--    If a variable is removed, it only has to be removed from here.
!--    Please check, if the arrays are in the list of POSSIBLE exchange arrays
!--    in subroutines:
!--    pmci_set_array_pointer (for parent arrays)
!--    pmci_create_child_arrays (for child arrays)
       CALL pmc_set_dataarray_name( 'coarse', 'u'  ,'fine', 'u',  ierr )
       CALL pmc_set_dataarray_name( 'coarse', 'v'  ,'fine', 'v',  ierr )
       CALL pmc_set_dataarray_name( 'coarse', 'w'  ,'fine', 'w',  ierr )
!
!--    Set data array name for TKE. Please note, nesting of TKE is actually
!--    only done if both parent and child are in LES or in RANS mode. Due to 
!--    design of model coupler, however, data array names must be already 
!--    available at this point.
       CALL pmc_set_dataarray_name( 'coarse', 'e'  ,'fine', 'e',  ierr )
!
!--    Nesting of dissipation rate only if both parent and child are in RANS
!--    mode and TKE-epsilo closure is applied. Please so also comment for TKE
!--    above.
       CALL pmc_set_dataarray_name( 'coarse', 'diss'  ,'fine', 'diss',  ierr )

       IF ( .NOT. neutral )  THEN
          CALL pmc_set_dataarray_name( 'coarse', 'pt' ,'fine', 'pt', ierr )
       ENDIF

       IF ( humidity )  THEN

          CALL pmc_set_dataarray_name( 'coarse', 'q'  ,'fine', 'q',  ierr )

          IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN
            CALL pmc_set_dataarray_name( 'coarse', 'qc'  ,'fine', 'qc',  ierr )  
            CALL pmc_set_dataarray_name( 'coarse', 'nc'  ,'fine', 'nc',  ierr ) 
          ENDIF

          IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN
             CALL pmc_set_dataarray_name( 'coarse', 'qr'  ,'fine', 'qr',  ierr )
             CALL pmc_set_dataarray_name( 'coarse', 'nr'  ,'fine', 'nr',  ierr ) 
          ENDIF
     
       ENDIF

       IF ( passive_scalar )  THEN
          CALL pmc_set_dataarray_name( 'coarse', 's'  ,'fine', 's',  ierr )
       ENDIF

       IF( particle_advection )  THEN
          CALL pmc_set_dataarray_name( 'coarse', 'nr_part'  ,'fine',           &
               'nr_part',  ierr )
          CALL pmc_set_dataarray_name( 'coarse', 'part_adr'  ,'fine',          &
               'part_adr',  ierr )
       ENDIF
       
       IF ( air_chemistry  .AND.  nest_chemistry )  THEN
          DO  n = 1, nspec
             CALL pmc_set_dataarray_name( 'coarse',                            &
                                          'chem_' //                           &
                                          TRIM( chem_species(n)%name ),        &
                                         'fine',                               &
                                          'chem_' //                           &
                                          TRIM( chem_species(n)%name ),        &
                                          ierr )
          ENDDO 
       ENDIF

       CALL pmc_set_dataarray_name( lastentry = .TRUE. )
!
!--    Send grid to parent
       val(1)  = nx
       val(2)  = ny
       val(3)  = nz
       val(4)  = dx
       val(5)  = dy
       fval(1) = zw(nzt+1)
       fval(2) = zw(nzt)
       fval(3) = dx
       fval(4) = dy
       fval(5) = dz(1)

       IF ( myid == 0 )  THEN

          CALL pmc_send_to_parent( val, SIZE( val ), 0, 123, ierr )
          CALL pmc_send_to_parent( fval, SIZE( fval ), 0, 124, ierr )
          CALL pmc_send_to_parent( coord_x, nx + 1 + 2 * nbgp, 0, 11, ierr )
          CALL pmc_send_to_parent( coord_y, ny + 1 + 2 * nbgp, 0, 12, ierr )

          CALL pmc_recv_from_parent( rans_mode_parent, 1, 0, 19, ierr )
!
!
!--       Receive Coarse grid information.
          CALL pmc_recv_from_parent( parent_grid_info_real,                    &
                                     SIZE(parent_grid_info_real), 0, 21, ierr )
          CALL pmc_recv_from_parent( parent_grid_info_int,  3, 0, 22, ierr )
!
!--        Debug-printouts - keep them
!           WRITE(0,*) 'Coarse grid from parent '
!           WRITE(0,*) 'startx_tot    = ',parent_grid_info_real(1)
!           WRITE(0,*) 'starty_tot    = ',parent_grid_info_real(2)
!           WRITE(0,*) 'endx_tot      = ',parent_grid_info_real(5)
!           WRITE(0,*) 'endy_tot      = ',parent_grid_info_real(6)
!           WRITE(0,*) 'dx            = ',parent_grid_info_real(3)
!           WRITE(0,*) 'dy            = ',parent_grid_info_real(4)
!           WRITE(0,*) 'dz            = ',parent_grid_info_real(7)
!           WRITE(0,*) 'nx_coarse     = ',parent_grid_info_int(1)
!           WRITE(0,*) 'ny_coarse     = ',parent_grid_info_int(2)
!           WRITE(0,*) 'nz_coarse     = ',parent_grid_info_int(3)
       ENDIF

       CALL MPI_BCAST( parent_grid_info_real, SIZE(parent_grid_info_real),     &
                       MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( parent_grid_info_int, 3, MPI_INTEGER, 0, comm2d, ierr )

       cg%dx = parent_grid_info_real(3)
       cg%dy = parent_grid_info_real(4)
       cg%dz = parent_grid_info_real(7)
       cg%nx = parent_grid_info_int(1)
       cg%ny = parent_grid_info_int(2)
       cg%nz = parent_grid_info_int(3)
!
!--    Get parent coordinates on coarse grid
       ALLOCATE( cg%coord_x(-nbgp:cg%nx+nbgp) )
       ALLOCATE( cg%coord_y(-nbgp:cg%ny+nbgp) )
      
       ALLOCATE( cg%dzu(1:cg%nz+1) )
       ALLOCATE( cg%dzw(1:cg%nz+1) )
       ALLOCATE( cg%zu(0:cg%nz+1) )
       ALLOCATE( cg%zw(0:cg%nz+1) )
!
!--    Get coarse grid coordinates and values of the z-direction from the parent
       IF ( myid == 0)  THEN
          CALL pmc_recv_from_parent( cg%coord_x, cg%nx+1+2*nbgp, 0, 24, ierr )
          CALL pmc_recv_from_parent( cg%coord_y, cg%ny+1+2*nbgp, 0, 25, ierr )
          CALL pmc_recv_from_parent( cg%dzu, cg%nz+1, 0, 26, ierr )
          CALL pmc_recv_from_parent( cg%dzw, cg%nz+1, 0, 27, ierr )
          CALL pmc_recv_from_parent( cg%zu, cg%nz+2, 0, 28, ierr )
          CALL pmc_recv_from_parent( cg%zw, cg%nz+2, 0, 29, ierr )
       ENDIF
!
!--    Broadcast this information
       CALL MPI_BCAST( cg%coord_x, cg%nx+1+2*nbgp, MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( cg%coord_y, cg%ny+1+2*nbgp, MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( cg%dzu, cg%nz+1, MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( cg%dzw, cg%nz+1, MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( cg%zu, cg%nz+2,  MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( cg%zw, cg%nz+2,  MPI_REAL, 0, comm2d, ierr )
       CALL MPI_BCAST( rans_mode_parent, 1, MPI_LOGICAL, 0, comm2d, ierr )
!
!--    Find the index bounds for the nest domain in the coarse-grid index space
       CALL pmci_map_fine_to_coarse_grid
!
!--    TO_DO: Klaus give a comment what is happening here
       CALL pmc_c_get_2d_index_list
!
!--    Include couple arrays into child content
!--    TO_DO: Klaus: better explain the above comment (what is child content?)
       CALL  pmc_c_clear_next_array_list

       n = 1
       DO  WHILE ( pmc_c_getnextarray( myname ) )
!--       Note that cg%nz is not the original nz of parent, but the highest
!--       parent-grid level needed for nesting.
!--       Please note, in case of chemical species an additional parameter
!--       need to be passed, which is required to set the pointer correctly
!--       to the chemical-species data structure. Hence, first check if current
!--       variable is a chemical species. If so, pass index id of respective
!--       species and increment this subsequently.
          IF ( INDEX( TRIM( myname ), 'chem_' ) /= 0 )  THEN             
             CALL pmci_create_child_arrays ( myname, icl, icr, jcs, jcn, cg%nz, n )
             n = n + 1
          ELSE
             CALL pmci_create_child_arrays ( myname, icl, icr, jcs, jcn, cg%nz )
          ENDIF
       ENDDO
       CALL pmc_c_setind_and_allocmem
!
!--    Precompute interpolation coefficients and child-array indices
       CALL pmci_init_interp_tril
!
!--    Precompute the log-law correction index- and ratio-arrays 
       IF ( constant_flux_layer )  THEN
          CALL pmci_init_loglaw_correction
       ENDIF
!
!--    Two-way coupling for general and vertical nesting.
!--    Precompute the index arrays and relaxation functions for the
!--    anterpolation
!
!--    Note that the anterpolation index bounds are needed also in case
!--    of one-way coupling because of the reversibility correction
!--    included in the interpolation algorithms.
       CALL pmci_init_anterp_tophat

    ENDIF

 CONTAINS


    SUBROUTINE pmci_map_fine_to_coarse_grid
!
!--    Determine index bounds of interpolation/anterpolation area in the coarse
!--    grid index space
       IMPLICIT NONE

       INTEGER(iwp), DIMENSION(5,numprocs) ::  coarse_bound_all   !<
       INTEGER(iwp), DIMENSION(2)          ::  size_of_array      !<
                                   
       INTEGER(iwp) :: i        !< 
       INTEGER(iwp) :: iauxl    !<
       INTEGER(iwp) :: iauxr    !<
       INTEGER(iwp) :: ijaux    !< 
       INTEGER(iwp) :: j        !<
       INTEGER(iwp) :: jauxs    !<
       INTEGER(iwp) :: jauxn    !< 
       REAL(wp) ::  loffset     !<
       REAL(wp) ::  noffset     !<
       REAL(wp) ::  roffset     !<
       REAL(wp) ::  soffset     !<
       REAL(wp) ::  xexl        !< Parent-grid array exceedance behind the left edge of the child PE subdomain
       REAL(wp) ::  xexr        !< Parent-grid array exceedance behind the right edge of the child PE subdomain 
       REAL(wp) ::  yexs        !< Parent-grid array exceedance behind the south edge of the child PE subdomain
       REAL(wp) ::  yexn        !< Parent-grid array exceedance behind the north edge of the child PE subdomain

!AH!
!AH!--    If the fine- and coarse grid nodes do not match:
!AH       loffset = MOD( coord_x(nxl), cg%dx )
!AH       xexl    = cg%dx + loffset
!AH       xcs  = coord_x(nxl) - xexl
!AH       DO  i = 0, cg%nx
!AH          IF ( cg%coord_x(i) > xcs )  THEN
!AH             icl = MAX( -1, i-1 )
!AH             EXIT
!AH          ENDIF
!AH       ENDDO
!AH!
!AH!--    If the fine- and coarse grid nodes do not match
!AH       roffset = MOD( coord_x(nxr+1), cg%dx )
!AH       xexr    = cg%dx + roffset
!AH       xce  = coord_x(nxr+1)
!AH       IF ( nxr == nx )  THEN
!AH          xce = xce + xexr
!AH       ENDIF
!AH       DO  i = cg%nx, 0 , -1
!AH          IF ( cg%coord_x(i) < xce )  THEN
!AH             icr = MIN( cg%nx+1, i+1 )
!AH             EXIT
!AH          ENDIF
!AH       ENDDO
!AH!
!AH!--    If the fine- and coarse grid nodes do not match
!AH       soffset = MOD( coord_y(nys), cg%dy )
!AH       yexs    = cg%dy + soffset
!AH       ycs  = coord_y(nys) - yexs
!AH       DO  j = 0, cg%ny
!AH          IF ( cg%coord_y(j) > ycs )  THEN
!AH             jcs = MAX( -nbgp, j-1 )
!AH             EXIT
!AH          ENDIF
!AH       ENDDO
!AH!
!AH!--    If the fine- and coarse grid nodes do not match
!AH       noffset = MOD( coord_y(nyn+1), cg%dy )
!AH       yexn    = cg%dy + noffset
!AH       yce  = coord_y(nyn+1)
!AH       IF ( nyn == ny )  THEN
!AH          yce = yce + yexn
!AH       ENDIF
!AH       DO  j = cg%ny, 0, -1
!AH          IF ( cg%coord_y(j) < yce )  THEN
!AH             jcn = MIN( cg%ny + nbgp, j+1 )
!AH             EXIT
!AH          ENDIF
!AH       ENDDO
!AH
!AH       coarse_bound(1) = icl
!AH       coarse_bound(2) = icr
!AH       coarse_bound(3) = jcs
!AH       coarse_bound(4) = jcn
!AH       coarse_bound(5) = myid
!
!--    Determine the anterpolation index limits. If at least half of the 
!--    parent-grid cell is within the current child sub-domain, then it 
!--    is included in the current sub-domain's anterpolation domain. 
!--    Else the parent-grid cell is included in the neighbouring subdomain's 
!--    anterpolation domain, or not included at all if we are at the outer 
!--    edge of the child domain.

!
!--    Left
       IF  ( bc_dirichlet_l )  THEN
          loffset = MOD( coord_x(nxl), cg%dx )
          xexl  = 2 * cg%dx + loffset
          iauxl = 0
       ELSE
          xexl  = 0.0_wp
          iauxl = 1
       ENDIF
       xcs     = coord_x(nxl) - xexl
       DO  i = 0, cg%nx
          IF ( cg%coord_x(i) + 0.5_wp * cg%dx >= xcs )  THEN
             icl = MAX( 0, i )
             EXIT
          ENDIF
       ENDDO
!
!--    Right
       IF  ( bc_dirichlet_r )  THEN
          roffset = MOD( coord_x(nxr+1), cg%dx )
          xexr  = 2 * cg%dx + roffset
          iauxr = 0  
       ELSE
          xexr  = 0.0_wp
          iauxr = 1  
       ENDIF
       xce  = coord_x(nxr+1) + xexr
       DO  i = cg%nx, 0 , -1
          IF ( cg%coord_x(i) + 0.5_wp * cg%dx <= xce )  THEN
             icr = MIN( cg%nx, i )
             EXIT
          ENDIF
       ENDDO
!
!--    South
       IF  ( bc_dirichlet_s )  THEN
          soffset = MOD( coord_y(nys), cg%dy )
          yexs  = 2 * cg%dy + soffset
          jauxs = 0  
       ELSE
          yexs  = 0.0_wp
          jauxs = 1  
       ENDIF
       ycs  = coord_y(nys) - yexs
       DO  j = 0, cg%ny
          IF ( cg%coord_y(j) + 0.5_wp * cg%dy >= ycs )  THEN
             jcs = MAX( 0, j )
             EXIT
          ENDIF
       ENDDO
!
!--    North
       IF  ( bc_dirichlet_n )  THEN
          noffset = MOD( coord_y(nyn+1), cg%dy )
          yexn  = 2 * cg%dy + noffset
          jauxn = 0
       ELSE
          yexn  = 0.0_wp
          jauxn = 1
       ENDIF
       yce  = coord_y(nyn+1) + yexn
       DO  j = cg%ny, 0 , -1
          IF ( cg%coord_y(j) + 0.5_wp * cg%dy <= yce )  THEN
             jcn = MIN( cg%ny, j )
             EXIT
          ENDIF
       ENDDO
!
!--    Make sure that the indexing is contiguous (no gaps, no overlaps) 
#if defined( __parallel )
       IF ( nxl == 0 )  THEN
          CALL MPI_SEND( icr, 1, MPI_INTEGER, pright, 717, comm2d, ierr )
       ELSE IF ( nxr == nx )  THEN
          CALL MPI_RECV( ijaux, 1, MPI_INTEGER, pleft, 717, comm2d, status, ierr )
          icl = ijaux + 1
       ELSE
          CALL MPI_SEND( icr, 1, MPI_INTEGER, pright, 717, comm2d, ierr )
          CALL MPI_RECV( ijaux, 1, MPI_INTEGER, pleft, 717, comm2d, status, ierr ) 
          icl = ijaux + 1
       ENDIF
       IF ( nys == 0 )  THEN
          CALL MPI_SEND( jcn, 1, MPI_INTEGER, pnorth, 719, comm2d, ierr )
       ELSE IF ( nyn == ny )  THEN
          CALL MPI_RECV( ijaux, 1, MPI_INTEGER, psouth, 719, comm2d, status, ierr )
          jcs = ijaux + 1
       ELSE
          CALL MPI_SEND( jcn, 1, MPI_INTEGER, pnorth, 719, comm2d, ierr )
          CALL MPI_RECV( ijaux, 1, MPI_INTEGER, psouth, 719, comm2d, status, ierr ) 
          jcs = ijaux + 1
       ENDIF
#endif       

       WRITE(9,"('Pmci_map_fine_to_coarse_grid. parent-grid array bounds: ',4(i3,2x))") icl, icr, jcs, jcn
       FLUSH(9)

       coarse_bound(1) = icl
       coarse_bound(2) = icr
       coarse_bound(3) = jcs
       coarse_bound(4) = jcn
       coarse_bound(5) = myid
!
!--    The following index bounds are used for allocating index mapping and some other auxiliary arrays
       coarse_bound_aux(1) = icl - iauxl
       coarse_bound_aux(2) = icr + iauxr
       coarse_bound_aux(3) = jcs - jauxs
       coarse_bound_aux(4) = jcn + jauxn
!
!--    Note that MPI_Gather receives data from all processes in the rank order
!--    TO_DO: refer to the line where this fact becomes important
       CALL MPI_GATHER( coarse_bound, 5, MPI_INTEGER, coarse_bound_all, 5,     &
                        MPI_INTEGER, 0, comm2d, ierr )

       IF ( myid == 0 )  THEN
          size_of_array(1) = SIZE( coarse_bound_all, 1 )
          size_of_array(2) = SIZE( coarse_bound_all, 2 )
          CALL pmc_send_to_parent( size_of_array, 2, 0, 40, ierr )
          CALL pmc_send_to_parent( coarse_bound_all, SIZE( coarse_bound_all ), &
                                   0, 41, ierr )
       ENDIF

    END SUBROUTINE pmci_map_fine_to_coarse_grid



    SUBROUTINE pmci_init_interp_tril
!
!--    Precomputation of the interpolation coefficients and child-array indices
!--    to be used by the interpolation routines interp_tril_lr, interp_tril_ns
!--    and interp_tril_t.

       IMPLICIT NONE

       INTEGER(iwp) ::  acsize   !< Maximum dimension of anterpolation cell.
       INTEGER(iwp) ::  i        !< Child-grid i-index
       INTEGER(iwp) ::  ierr     !< MPI error code
       INTEGER(iwp) ::  j        !< Child-grid j-index
       INTEGER(iwp) ::  k        !< Child-grid k-index
       INTEGER(iwp) ::  kc       !< 
       INTEGER(iwp) ::  kdzo     !<
       INTEGER(iwp) ::  kdzw     !<
       INTEGER(iwp) ::  moff     !< Parent-grid bound offset in j-direction
       INTEGER(iwp) ::  loff     !< Parent-grid bound offset in i-direction

       REAL(wp) ::  dzmin        !<
       REAL(wp) ::  parentdzmax  !<
       REAL(wp) ::  xb           !<
       REAL(wp) ::  xcsu         !<
       REAL(wp) ::  xfso         !<
       REAL(wp) ::  xcso         !<
       REAL(wp) ::  xfsu         !<
       REAL(wp) ::  yb           !<
       REAL(wp) ::  ycso         !<
       REAL(wp) ::  ycsv         !<
       REAL(wp) ::  yfso         !<
       REAL(wp) ::  yfsv         !<
       REAL(wp) ::  zcso         !<
       REAL(wp) ::  zcsw         !<
       REAL(wp) ::  zfso         !<
       REAL(wp) ::  zfsw         !<
      
       
!AH
!
!--    Allocate child-grid work arrays for interpolation.
       CALL pmci_allocate_finegrid_workarrays
!       
!--    Determine index bounds for the parent-grid work arrays for
!--    interpolation and allocate them.
       CALL pmci_allocate_coarsegrid_workarrays
!       
!--    Define the MPI-datatypes for parent-grid work array 
!--    exchange between the PE-subdomains.
       CALL pmci_create_coarsegrid_workarray_exchange_datatypes
!
!--    Determine index bounds for the fine-to-coarse grid index mapping arrays 
!--    and interpolation-coefficient arrays and allocate them.       
       CALL pmci_allocate_fine_to_coarse_mapping_arrays
!AH       
!
       xb   = nxl * dx
       IF  ( bc_dirichlet_l )  THEN
          loff = 2
       ELSE
          loff = 0
       ENDIF
!AH       DO  i = nxlg, nxrg
       DO  i = nxl-1, nxr+1
          xfsu    = coord_x(i) - ( lower_left_coord_x + xb )
          xfso    = coord_x(i) + 0.5_wp * dx - ( lower_left_coord_x + xb )
!
!--       icl points to 2 parent-grid cells left form the left nest boundary,
!--       thence icl + loff points to the left nest boundary.
          icu(i)  = icl + loff + FLOOR( xfsu / cg%dx )
          ico(i)  = icl + loff + FLOOR( ( xfso - 0.5_wp * cg%dx ) / cg%dx )
          xcsu    = ( icu(i) - ( icl + loff ) ) * cg%dx
          xcso    = ( ico(i) - ( icl + loff ) + 0.5_wp ) * cg%dx
          r2xu(i) = ( xfsu - xcsu ) / cg%dx
          r2xo(i) = ( xfso - xcso ) / cg%dx
          r1xu(i) = 1.0_wp - r2xu(i)
          r1xo(i) = 1.0_wp - r2xo(i)
       ENDDO
!       
!--    Fill up the values behind nest boundaries by copying from inside 
!--    the domain.
       IF  ( bc_dirichlet_l )  THEN
          icu(nxlfc:nxl-1)  = icu(nxlfc+igsr:nxl-1+igsr) - 1
          r1xu(nxlfc:nxl-1) = r1xu(nxlfc+igsr:nxl-1+igsr)
          r2xu(nxlfc:nxl-1) = r2xu(nxlfc+igsr:nxl-1+igsr)
          ico(nxlfc:nxl-1)  = ico(nxlfc+igsr:nxl-1+igsr) - 1
          r1xo(nxlfc:nxl-1) = r1xo(nxlfc+igsr:nxl-1+igsr)
          r2xo(nxlfc:nxl-1) = r2xo(nxlfc+igsr:nxl-1+igsr)
       ENDIF
       
       IF  ( bc_dirichlet_r )  THEN
          icu(nxr+1:nxrfc)  = icu(nxr+1-igsr:nxrfc-igsr) + 1
          r1xu(nxr+1:nxrfc) = r1xu(nxr+1-igsr:nxrfc-igsr)
          r2xu(nxr+1:nxrfc) = r2xu(nxr+1-igsr:nxrfc-igsr)
          ico(nxr+1:nxrfc)  = ico(nxr+1-igsr:nxrfc-igsr) + 1
          r1xo(nxr+1:nxrfc) = r1xo(nxr+1-igsr:nxrfc-igsr)
          r2xo(nxr+1:nxrfc) = r2xo(nxr+1-igsr:nxrfc-igsr)
       ENDIF
!
!--    Print out the indices and coefficients for checking and debugging purposes
       DO  i = nxlfc, nxrfc
          WRITE(9,"('pmci_init_interp_tril: i, icu, r1xu r2xu ', 2(i4,2x),2(e12.5,2x))") &
               i, icu(i), r1xu(i), r2xu(i)
          FLUSH(9)
       ENDDO
       WRITE(9,*)
       DO  i = nxlfc, nxrfc
          WRITE(9,"('pmci_init_interp_tril: i, ico, r1xo r2xo ', 2(i4,2x),2(e12.5,2x))") &
               i, ico(i), r1xo(i), r2xo(i)
          FLUSH(9)
       ENDDO
       WRITE(9,*)

       yb   = nys * dy
       IF  ( bc_dirichlet_s )  THEN
          moff = 2
       ELSE
          moff = 0
       ENDIF
!AH       DO  j = nysg, nyng
       DO  j = nys-1, nyn+1
          yfsv    = coord_y(j) - ( lower_left_coord_y + yb )
          yfso    = coord_y(j) + 0.5_wp * dy - ( lower_left_coord_y + yb )
!
!--       jcs points to 2 parent-grid cells south form the south nest boundary,
!--       thence jcs + moff points to the south nest boundary.
          jcv(j)  = jcs + moff + FLOOR( yfsv / cg%dy )
          jco(j)  = jcs + moff + FLOOR( ( yfso - 0.5_wp * cg%dy ) / cg%dy )
          ycsv    = ( jcv(j) - ( jcs + moff ) ) * cg%dy
          ycso    = ( jco(j) - ( jcs + moff ) + 0.5_wp ) * cg%dy
          r2yv(j) = ( yfsv - ycsv ) / cg%dy
          r2yo(j) = ( yfso - ycso ) / cg%dy
          r1yv(j) = 1.0_wp - r2yv(j)
          r1yo(j) = 1.0_wp - r2yo(j)
       ENDDO
!       
!--    Fill up the values behind nest boundaries by copying from inside 
!--    the domain.
       IF  ( bc_dirichlet_s )  THEN
          jcv(nysfc:nys-1)  = jcv(nysfc+jgsr:nys-1+jgsr) - 1
          r1yv(nysfc:nys-1) = r1yv(nysfc+jgsr:nys-1+jgsr)
          r2yv(nysfc:nys-1) = r2yv(nysfc+jgsr:nys-1+jgsr)
          jco(nysfc:nys-1)  = jco(nysfc+jgsr:nys-1+jgsr) - 1
          r1yo(nysfc:nys-1) = r1yo(nysfc+jgsr:nys-1+jgsr)
          r2yo(nysfc:nys-1) = r2yo(nysfc+jgsr:nys-1+jgsr)
       ENDIF
       
       IF  ( bc_dirichlet_n )  THEN
          jcv(nyn+1:nynfc)  = jcv(nyn+1-jgsr:nynfc-jgsr) + 1
          r1yv(nyn+1:nynfc) = r1yv(nyn+1-jgsr:nynfc-jgsr)
          r2yv(nyn+1:nynfc) = r2yv(nyn+1-jgsr:nynfc-jgsr)
          jco(nyn+1:nynfc)  = jco(nyn+1-jgsr:nynfc-jgsr) + 1
          r1yo(nyn+1:nynfc) = r1yo(nyn+1-jgsr:nynfc-jgsr)
          r2yo(nyn+1:nynfc) = r2yo(nyn+1-jgsr:nynfc-jgsr)
       ENDIF
!
!--    Print out the indices and coefficients for checking and debugging purposes
       DO  j = nysfc, nynfc
          WRITE(9,"('pmci_init_interp_tril: j, jcv, r1yv r2yv ', 2(i4,2x),2(e12.5,2x))") &
               j, jcv(j), r1yv(j), r2yv(j)
          FLUSH(9)
       ENDDO
       WRITE(9,*)
       DO  j = nysfc, nynfc
          WRITE(9,"('pmci_init_interp_tril: j, jco, r1yo r2yo ', 2(i4,2x),2(e12.5,2x))") &
               j, jco(j), r1yo(j), r2yo(j)
          FLUSH(9)
       ENDDO
       WRITE(9,*)

       DO  k = nzb, nzt + 1
          zfsw = zw(k)
          zfso = zu(k)

          DO kc = 0, cg%nz+1
             IF ( cg%zw(kc) > zfsw )  EXIT
          ENDDO
          kcw(k) = kc - 1
          
          DO kc = 0, cg%nz+1
             IF ( cg%zu(kc) > zfso )  EXIT
          ENDDO
          kco(k) = kc - 1

          zcsw    = cg%zw(kcw(k))
          zcso    = cg%zu(kco(k))
          kdzw    = MIN( kcw(k)+1, cg%nz+1 )
          kdzo    = MIN( kco(k)+1, cg%nz+1 )
          r2zw(k) = ( zfsw - zcsw ) / cg%dzw(kdzw)
          r2zo(k) = ( zfso - zcso ) / cg%dzu(kdzo)
          r1zw(k) = 1.0_wp - r2zw(k)
          r1zo(k) = 1.0_wp - r2zo(k)
       ENDDO
!
!--    Set the interpolation index- and coefficient-information to the 
!--    child-grid cells within the uppermost parent-grid cell. This 
!--    information is only needed for the reversibility correction.   
       kco(nzt+2:nzt+kgsr)  = kco(nzt+1) + 1 
       r1zo(nzt+2:nzt+kgsr) = r1zo(nzt+2-kgsr:nzt)
       r2zo(nzt+2:nzt+kgsr) = r2zo(nzt+2-kgsr:nzt)
! 
!--    kcw, r1zw and r2zw are not needed when k > nzt+1
       kcw(nzt+2:nzt+kgsr)  = 0
       r1zw(nzt+2:nzt+kgsr) = 0.0_wp
       r2zw(nzt+2:nzt+kgsr) = 0.0_wp
!
!--    Print out the indices and coefficients for checking and debugging purposes
       DO  k = nzb, nzt+1
          WRITE(9,"('pmci_init_interp_tril: k, kcw, r1zw r2zw ', 2(i4,2x),2(e12.5,2x))") &
               k, kcw(k), r1zw(k), r2zw(k)
          FLUSH(9)
       ENDDO
       WRITE(9,*)
       DO  k = nzb, nzt + kgsr
          WRITE(9,"('pmci_init_interp_tril: k, kco, r1zo r2zo ', 2(i4,2x),2(e12.5,2x))") &
               k, kco(k), r1zo(k), r2zo(k)
          FLUSH(9)
       ENDDO
       WRITE(9,*)
!
!--    Determine the maximum dimension of anterpolation cells and allocate the
!--    work array celltmpd needed in the reversibility correction in the 
!--    interpolation
       dzmin = 999999.9_wp
       DO k = 1, nzt+1
          dzmin = MIN( dzmin, dzu(k), dzw(k) )
       ENDDO
       parentdzmax = 0.0_wp
       DO k = 1, cg%nz+1
          parentdzmax = MAX(parentdzmax , cg%dzu(k), cg%dzw(k) )
       ENDDO
       acsize = CEILING( cg%dx / dx ) * CEILING( cg%dy / dy ) *                 &
            CEILING( parentdzmax / dzmin )
       ALLOCATE( celltmpd(1:acsize) )
      
    END SUBROUTINE pmci_init_interp_tril
    


    SUBROUTINE pmci_allocate_finegrid_workarrays
!
!--    Allocate child-grid work-arrays for interpolation
       IMPLICIT NONE


       igsr = NINT( cg%dx / dx, iwp )
       jgsr = NINT( cg%dy / dy, iwp )
       kgsr = NINT( cg%dzu(cg%nz+1) / dzu(nzt+1), iwp )
       write(9,"('igsr, jgsr, kgsr: ',3(i3,2x))") igsr, jgsr, kgsr
       flush(9)
!       
!--    Note that i-indexing for workarr_lr runs from 0 to igsr-1.
!--    For u, only 0-element is used and all elements 0:igsr-1
!--    are used for all other variables.
       ALLOCATE( workarr_lr(nzb:nzt+1,nysg:nyng,0:igsr-1) )
!       
!--    Note that j-indexing for workarr_sn runs from 0 to jgsr-1.
!--    For v, only 0-element is used and all elements 0:jgsr-1
!--    are used for all other variables.
       ALLOCATE( workarr_sn(nzb:nzt+1,0:jgsr-1,nxlg:nxrg) )
!
!--    Note that genuine k-indexing is used for workarr_t.
!--    Only nzt-element is used for w and elements nzt+1:nzt+kgsr
!--    are used for all other variables. 
       ALLOCATE( workarr_t(nzt:nzt+kgsr,nysg:nyng,nxlg:nxrg) )

    END SUBROUTINE pmci_allocate_finegrid_workarrays



    SUBROUTINE pmci_allocate_coarsegrid_workarrays
!
!--    Allocate parent-grid work-arrays for interpolation
       IMPLICIT NONE

!
!--    Determine and store the PE-subdomain dependent index bounds
       IF  ( bc_dirichlet_l )  THEN
          iclw = icl + 1
       ELSE
          iclw = icl - 1
       ENDIF

       IF  ( bc_dirichlet_r )  THEN
          icrw = icr - 1
       ELSE
          icrw = icr + 1
       ENDIF

       IF  ( bc_dirichlet_s )  THEN
          jcsw = jcs + 1
       ELSE
          jcsw = jcs - 1
       ENDIF

       IF  ( bc_dirichlet_n )  THEN
          jcnw = jcn - 1
       ELSE
          jcnw = jcn + 1
       ENDIF
    
       coarse_bound_w(1) = iclw
       coarse_bound_w(2) = icrw
       coarse_bound_w(3) = jcsw
       coarse_bound_w(4) = jcnw
!
!--    Left and right boundaries.
       ALLOCATE( workarrc_lr(0:cg%nz+1,jcsw:jcnw,0:2) )
!
!--    South and north boundaries.
       ALLOCATE( workarrc_sn(0:cg%nz+1,0:2,iclw:icrw) )
!
!--    Top boundary.
       ALLOCATE( workarrc_t(0:2,jcsw:jcnw,iclw:icrw) )

    END SUBROUTINE pmci_allocate_coarsegrid_workarrays



    SUBROUTINE pmci_create_coarsegrid_workarray_exchange_datatypes
!
!--    Define specific MPI types for workarrc-exhchange.
       IMPLICIT NONE

#if defined( __parallel )       
!
!--    For the left and right boundaries
       CALL MPI_TYPE_VECTOR( 3, cg%nz+2, (jcnw-jcsw+1)*(cg%nz+2), MPI_REAL,     &
            workarrc_lr_exchange_type, ierr )
       CALL MPI_TYPE_COMMIT( workarrc_lr_exchange_type, ierr )
!
!--    For the south and north boundaries
       CALL MPI_TYPE_VECTOR( 1, 3*(cg%nz+2), 3*(cg%nz+2), MPI_REAL,             &
            workarrc_sn_exchange_type, ierr )
       CALL MPI_TYPE_COMMIT( workarrc_sn_exchange_type, ierr )
!
!--    For the top-boundary x-slices
       CALL MPI_TYPE_VECTOR( icrw-iclw+1, 3, 3*(jcnw-jcsw+1), MPI_REAL,         &
            workarrc_t_exchange_type_x, ierr )
       CALL MPI_TYPE_COMMIT( workarrc_t_exchange_type_x, ierr )
!
!--    For the top-boundary y-slices
       CALL MPI_TYPE_VECTOR( 1, 3*(jcnw-jcsw+1), 3*(jcnw-jcsw+1), MPI_REAL,         &
            workarrc_t_exchange_type_y, ierr )
       CALL MPI_TYPE_COMMIT( workarrc_t_exchange_type_y, ierr )
#endif
       
    END SUBROUTINE pmci_create_coarsegrid_workarray_exchange_datatypes



    SUBROUTINE pmci_allocate_fine_to_coarse_mapping_arrays
!
!--    Define index limits and allocate the fine-to-coarse grid index mapping 
!--    arrays and interpolation coefficient arrays.      
       IMPLICIT NONE

       
       IF  ( bc_dirichlet_l )  THEN
!AH          nxlfc = MIN( nxl-igsr, nxlg )          
          nxlfc = nxl - igsr
       ELSE
!AH          nxlfc = nxlg
          nxlfc = nxl - 1
       ENDIF
       IF  ( bc_dirichlet_r )  THEN
!AH          nxrfc = MAX( nxr+igsr, nxrg )
          nxrfc = nxr + igsr
       ELSE
!AH          nxrfc = nxrg
          nxrfc = nxr + 1          
       ENDIF

       IF  ( bc_dirichlet_s )  THEN
!AH          nysfc = MIN( nys-jgsr, nysg )
          nysfc = nys - jgsr
       ELSE
!AH          nysfc = nysg
          nysfc = nys - 1
       ENDIF
       IF  ( bc_dirichlet_n )  THEN
!AH          nynfc = MAX( nyn+jgsr, nyng )
          nynfc = nyn + jgsr
       ELSE
!AH          nynfc = nyng          
          nynfc = nyn + 1          
       ENDIF

       ALLOCATE( icu(nxlfc:nxrfc) )       
       ALLOCATE( ico(nxlfc:nxrfc) )
       ALLOCATE( jcv(nysfc:nynfc) )
       ALLOCATE( jco(nysfc:nynfc) )
       ALLOCATE( kcw(nzb:nzt+kgsr) )
       ALLOCATE( kco(nzb:nzt+kgsr) )

       ALLOCATE( r1xu(nxlfc:nxrfc) )
       ALLOCATE( r2xu(nxlfc:nxrfc) )
       ALLOCATE( r1xo(nxlfc:nxrfc) )
       ALLOCATE( r2xo(nxlfc:nxrfc) )
       ALLOCATE( r1yv(nysfc:nynfc) )
       ALLOCATE( r2yv(nysfc:nynfc) )
       ALLOCATE( r1yo(nysfc:nynfc) )
       ALLOCATE( r2yo(nysfc:nynfc) )
       ALLOCATE( r1zw(nzb:nzt+kgsr) )
       ALLOCATE( r2zw(nzb:nzt+kgsr) )
       ALLOCATE( r1zo(nzb:nzt+kgsr) )
       ALLOCATE( r2zo(nzb:nzt+kgsr) )

    END SUBROUTINE pmci_allocate_fine_to_coarse_mapping_arrays

     
       
    SUBROUTINE pmci_init_loglaw_correction
!
!--    Precomputation of the index and log-ratio arrays for the log-law
!--    corrections for near-wall nodes after the nest-BC interpolation.
!--    These are used by the interpolation routines interp_tril_lr and
!--    interp_tril_ns.

       IMPLICIT NONE

       INTEGER(iwp) ::  direction      !< Wall normal index: 1=k, 2=j, 3=i.
       INTEGER(iwp) ::  dum            !< dummy value for reduce operation
       INTEGER(iwp) ::  i              !<
       INTEGER(iwp) ::  ierr           !< MPI status
       INTEGER(iwp) ::  inc            !< Wall outward-normal index increment -1
                                       !< or 1, for direction=1, inc=1 always
       INTEGER(iwp) ::  j              !<
       INTEGER(iwp) ::  k              !<
       INTEGER(iwp) ::  k_wall_u_ji    !< topography top index on u-grid
       INTEGER(iwp) ::  k_wall_u_ji_p  !< topography top index on u-grid
       INTEGER(iwp) ::  k_wall_u_ji_m  !< topography top index on u-grid
       INTEGER(iwp) ::  k_wall_v_ji    !< topography top index on v-grid
       INTEGER(iwp) ::  k_wall_v_ji_p  !< topography top index on v-grid
       INTEGER(iwp) ::  k_wall_v_ji_m  !< topography top index on v-grid
       INTEGER(iwp) ::  k_wall_w_ji    !< topography top index on w-grid
       INTEGER(iwp) ::  k_wall_w_ji_p  !< topography top index on w-grid
       INTEGER(iwp) ::  k_wall_w_ji_m  !< topography top index on w-grid
       INTEGER(iwp) ::  kb             !<
       INTEGER(iwp) ::  lc             !<
       INTEGER(iwp) ::  ni             !<
       INTEGER(iwp) ::  nj             !<
       INTEGER(iwp) ::  nk             !<
       INTEGER(iwp) ::  nzt_topo_max   !<
       INTEGER(iwp) ::  wall_index     !<  Index of the wall-node coordinate

       REAL(wp)     ::  z0_topo      !<  roughness at vertical walls
       REAL(wp), ALLOCATABLE, DIMENSION(:) ::  lcr   !<

!
!--    First determine the maximum k-index needed for the near-wall corrections.
!--    This maximum is individual for each boundary to minimize the storage
!--    requirements and to minimize the corresponding loop k-range in the
!--    interpolation routines.
       nzt_topo_nestbc_l = nzb
       IF ( bc_dirichlet_l )  THEN
          DO  i = nxl-1, nxl
             DO  j = nys, nyn
!
!--             Concept need to be reconsidered for 3D-topography
!--             Determine largest topography index on scalar grid
                nzt_topo_nestbc_l = MAX( nzt_topo_nestbc_l,                    &
                                    get_topography_top_index_ji( j, i, 's' ) )
!
!--             Determine largest topography index on u grid
                nzt_topo_nestbc_l = MAX( nzt_topo_nestbc_l,                    &
                                    get_topography_top_index_ji( j, i, 'u' ) )
!
!--             Determine largest topography index on v grid
                nzt_topo_nestbc_l = MAX( nzt_topo_nestbc_l,                    &
                                    get_topography_top_index_ji( j, i, 'v' ) )
!
!--             Determine largest topography index on w grid
                nzt_topo_nestbc_l = MAX( nzt_topo_nestbc_l,                    &
                                    get_topography_top_index_ji( j, i, 'w' ) )
             ENDDO
          ENDDO
          nzt_topo_nestbc_l = nzt_topo_nestbc_l + 1
       ENDIF
      
       nzt_topo_nestbc_r = nzb
       IF ( bc_dirichlet_r )  THEN
          i = nxr + 1
          DO  j = nys, nyn
!
!--             Concept need to be reconsidered for 3D-topography
!--             Determine largest topography index on scalar grid
                nzt_topo_nestbc_r = MAX( nzt_topo_nestbc_r,                    &
                                    get_topography_top_index_ji( j, i, 's' ) )
!
!--             Determine largest topography index on u grid
                nzt_topo_nestbc_r = MAX( nzt_topo_nestbc_r,                    &
                                    get_topography_top_index_ji( j, i, 'u' ) )
!
!--             Determine largest topography index on v grid
                nzt_topo_nestbc_r = MAX( nzt_topo_nestbc_r,                    &
                                    get_topography_top_index_ji( j, i, 'v' ) )
!
!--             Determine largest topography index on w grid
                nzt_topo_nestbc_r = MAX( nzt_topo_nestbc_r,                    &
                                    get_topography_top_index_ji( j, i, 'w' ) )
          ENDDO
          nzt_topo_nestbc_r = nzt_topo_nestbc_r + 1
       ENDIF

       nzt_topo_nestbc_s = nzb
       IF ( bc_dirichlet_s )  THEN
          DO  j = nys-1, nys
             DO  i = nxl, nxr
!
!--             Concept need to be reconsidered for 3D-topography
!--             Determine largest topography index on scalar grid
                nzt_topo_nestbc_s = MAX( nzt_topo_nestbc_s,                    &
                                    get_topography_top_index_ji( j, i, 's' ) )
!
!--             Determine largest topography index on u grid
                nzt_topo_nestbc_s = MAX( nzt_topo_nestbc_s,                    &
                                    get_topography_top_index_ji( j, i, 'u' ) )
!
!--             Determine largest topography index on v grid
                nzt_topo_nestbc_s = MAX( nzt_topo_nestbc_s,                    &
                                    get_topography_top_index_ji( j, i, 'v' ) )
!
!--             Determine largest topography index on w grid
                nzt_topo_nestbc_s = MAX( nzt_topo_nestbc_s,                    &
                                    get_topography_top_index_ji( j, i, 'w' ) )
             ENDDO
          ENDDO
          nzt_topo_nestbc_s = nzt_topo_nestbc_s + 1
       ENDIF

       nzt_topo_nestbc_n = nzb
       IF ( bc_dirichlet_n )  THEN
          j = nyn + 1
          DO  i = nxl, nxr
!
!--             Concept need to be reconsidered for 3D-topography
!--             Determine largest topography index on scalar grid
                nzt_topo_nestbc_n = MAX( nzt_topo_nestbc_n,                    &
                                    get_topography_top_index_ji( j, i, 's' ) )
!
!--             Determine largest topography index on u grid
                nzt_topo_nestbc_n = MAX( nzt_topo_nestbc_n,                    &
                                    get_topography_top_index_ji( j, i, 'u' ) )
!
!--             Determine largest topography index on v grid
                nzt_topo_nestbc_n = MAX( nzt_topo_nestbc_n,                    &
                                    get_topography_top_index_ji( j, i, 'v' ) )
!
!--             Determine largest topography index on w grid
                nzt_topo_nestbc_n = MAX( nzt_topo_nestbc_n,                    &
                                    get_topography_top_index_ji( j, i, 'w' ) )
          ENDDO
          nzt_topo_nestbc_n = nzt_topo_nestbc_n + 1
       ENDIF

#if defined( __parallel )
!
!--       Determine global topography-top index along child boundary. 
          dum = nzb
          CALL MPI_ALLREDUCE( nzt_topo_nestbc_l, dum, 1, MPI_INTEGER,          &
                              MPI_MAX, comm1dy, ierr )
          nzt_topo_nestbc_l = dum

          dum = nzb
          CALL MPI_ALLREDUCE( nzt_topo_nestbc_r, dum, 1, MPI_INTEGER,          &
                              MPI_MAX, comm1dy, ierr )
          nzt_topo_nestbc_r = dum

          dum = nzb
          CALL MPI_ALLREDUCE( nzt_topo_nestbc_n, dum, 1, MPI_INTEGER,          &
                              MPI_MAX, comm1dx, ierr )
          nzt_topo_nestbc_n = dum

          dum = nzb
          CALL MPI_ALLREDUCE( nzt_topo_nestbc_s, dum, 1, MPI_INTEGER,          &
                              MPI_MAX, comm1dx, ierr )
          nzt_topo_nestbc_s = dum
#endif
!
!--    Then determine the maximum number of near-wall nodes per wall point based
!--    on the grid-spacing ratios.
       nzt_topo_max = MAX( nzt_topo_nestbc_l, nzt_topo_nestbc_r,               &
                           nzt_topo_nestbc_s, nzt_topo_nestbc_n )
!
!--    Note that the outer division must be integer division.
       ni = CEILING( cg%dx / dx ) / 2
       nj = CEILING( cg%dy / dy ) / 2
       nk = 1
       DO  k = 1, nzt_topo_max
          nk = MAX( nk, CEILING( cg%dzu(kco(k)+1) / dzu(k) ) )
       ENDDO
       nk = nk / 2   !  Note that this must be integer division.
       ncorr =  MAX( ni, nj, nk )

       ALLOCATE( lcr(0:ncorr-1) )
       lcr = 1.0_wp

       z0_topo = roughness_length
!
!--    First horizontal walls. Note that also logc_w_? and logc_ratio_w_? and
!--    logc_kbounds_* need to be allocated and initialized here.
!--    Left boundary
       IF ( bc_dirichlet_l )  THEN

          ALLOCATE( logc_u_l(1:2,nzb:nzt_topo_nestbc_l,nys:nyn) )
          ALLOCATE( logc_v_l(1:2,nzb:nzt_topo_nestbc_l,nys:nyn) )
          ALLOCATE( logc_w_l(1:2,nzb:nzt_topo_nestbc_l,nys:nyn) )
          ALLOCATE( logc_kbounds_u_l(1:2,nys:nyn) )
          ALLOCATE( logc_kbounds_v_l(1:2,nys:nyn) )
          ALLOCATE( logc_kbounds_w_l(1:2,nys:nyn) )
          ALLOCATE( logc_ratio_u_l(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_l,nys:nyn) )
          ALLOCATE( logc_ratio_v_l(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_l,nys:nyn) )
          ALLOCATE( logc_ratio_w_l(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_l,nys:nyn) )
          logc_u_l       = 0
          logc_v_l       = 0
          logc_w_l       = 0
          logc_ratio_u_l = 1.0_wp
          logc_ratio_v_l = 1.0_wp
          logc_ratio_w_l = 1.0_wp
          direction      = 1
          inc            = 1

          DO  j = nys, nyn
!
!--          Left boundary for u
             i   = 0
!
!--          For loglaw correction the roughness z0 is required. z0, however, 
!--          is part of the surfacetypes now. Set default roughness instead. 
!--          Determine topography top index on u-grid
             kb  = get_topography_top_index_ji( j, i, 'u' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                            j, inc, wall_index, z0_topo,                       &
                            kb, direction, ncorr )

             logc_u_l(1,k,j) = lc
             logc_ratio_u_l(1,0:ncorr-1,k,j) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp
!
!--          Left boundary for v
             i   = -1
!
!--          Determine topography top index on v-grid
             kb  = get_topography_top_index_ji( j, i, 'v' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                            j, inc, wall_index, z0_topo,                       &
                            kb, direction, ncorr )

             logc_v_l(1,k,j) = lc
             logc_ratio_v_l(1,0:ncorr-1,k,j) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp

          ENDDO

       ENDIF
!
!--    Right boundary
       IF ( bc_dirichlet_r )  THEN
           
          ALLOCATE( logc_u_r(1:2,nzb:nzt_topo_nestbc_r,nys:nyn) )
          ALLOCATE( logc_v_r(1:2,nzb:nzt_topo_nestbc_r,nys:nyn) )
          ALLOCATE( logc_w_r(1:2,nzb:nzt_topo_nestbc_r,nys:nyn) )          
          ALLOCATE( logc_kbounds_u_r(1:2,nys:nyn) )
          ALLOCATE( logc_kbounds_v_r(1:2,nys:nyn) )
          ALLOCATE( logc_kbounds_w_r(1:2,nys:nyn) )
          ALLOCATE( logc_ratio_u_r(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_r,nys:nyn) )
          ALLOCATE( logc_ratio_v_r(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_r,nys:nyn) )
          ALLOCATE( logc_ratio_w_r(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_r,nys:nyn) )
          logc_u_r       = 0
          logc_v_r       = 0
          logc_w_r       = 0
          logc_ratio_u_r = 1.0_wp
          logc_ratio_v_r = 1.0_wp
          logc_ratio_w_r = 1.0_wp
          direction      = 1
          inc            = 1

          DO  j = nys, nyn
!
!--          Right boundary for u
             i   = nxr + 1
!
!--          For loglaw correction the roughness z0 is required. z0, however, 
!--          is part of the surfacetypes now, so call subroutine according
!--          to the present surface tpye. 
!--          Determine topography top index on u-grid
             kb  = get_topography_top_index_ji( j, i, 'u' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )

             logc_u_r(1,k,j) = lc
             logc_ratio_u_r(1,0:ncorr-1,k,j) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp
!
!--          Right boundary for v
             i   = nxr + 1
!
!--          Determine topography top index on v-grid
             kb  = get_topography_top_index_ji( j, i, 'v' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )

             logc_v_r(1,k,j) = lc
             logc_ratio_v_r(1,0:ncorr-1,k,j) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp

          ENDDO

       ENDIF
!
!--    South boundary
       IF ( bc_dirichlet_s )  THEN

          ALLOCATE( logc_u_s(1:2,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          ALLOCATE( logc_v_s(1:2,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          ALLOCATE( logc_w_s(1:2,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          ALLOCATE( logc_kbounds_u_s(1:2,nxl:nxr) )
          ALLOCATE( logc_kbounds_v_s(1:2,nxl:nxr) )
          ALLOCATE( logc_kbounds_w_s(1:2,nxl:nxr) )
          ALLOCATE( logc_ratio_u_s(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          ALLOCATE( logc_ratio_v_s(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          ALLOCATE( logc_ratio_w_s(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_s,nxl:nxr) )
          logc_u_s       = 0
          logc_v_s       = 0
          logc_w_s       = 0
          logc_ratio_u_s = 1.0_wp
          logc_ratio_v_s = 1.0_wp
          logc_ratio_w_s = 1.0_wp
          direction      = 1
          inc            = 1

          DO  i = nxl, nxr
!
!--          South boundary for u
             j   = -1
!
!--          Determine topography top index on u-grid
             kb  = get_topography_top_index_ji( j, i, 'u' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )

             logc_u_s(1,k,i) = lc
             logc_ratio_u_s(1,0:ncorr-1,k,i) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp
!
!--          South boundary for v
             j   = 0
!
!--          Determine topography top index on v-grid
             kb  = get_topography_top_index_ji( j, i, 'v' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )

             logc_v_s(1,k,i) = lc
             logc_ratio_v_s(1,0:ncorr-1,k,i) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp

          ENDDO

       ENDIF
!
!--    North boundary
       IF ( bc_dirichlet_n )  THEN

          ALLOCATE( logc_u_n(1:2,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          ALLOCATE( logc_v_n(1:2,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          ALLOCATE( logc_w_n(1:2,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          ALLOCATE( logc_kbounds_u_n(1:2,nxl:nxr) )
          ALLOCATE( logc_kbounds_v_n(1:2,nxl:nxr) )
          ALLOCATE( logc_kbounds_w_n(1:2,nxl:nxr) )
          ALLOCATE( logc_ratio_u_n(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          ALLOCATE( logc_ratio_v_n(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          ALLOCATE( logc_ratio_w_n(1:2,0:ncorr-1,nzb:nzt_topo_nestbc_n,nxl:nxr) )
          logc_u_n       = 0
          logc_v_n       = 0
          logc_w_n       = 0
          logc_ratio_u_n = 1.0_wp
          logc_ratio_v_n = 1.0_wp
          logc_ratio_w_n = 1.0_wp
          direction      = 1
          inc            = 1

          DO  i = nxl, nxr
!
!--          North boundary for u
             j   = nyn + 1
!
!--          Determine topography top index on u-grid
             kb  = get_topography_top_index_ji( j, i, 'u' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )

             logc_u_n(1,k,i) = lc
             logc_ratio_u_n(1,0:ncorr-1,k,i) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp
!
!--          North boundary for v
             j   = nyn + 1
!
!--          Determine topography top index on v-grid
             kb  = get_topography_top_index_ji( j, i, 'v' )
             k   = kb + 1
             wall_index = kb

             CALL pmci_define_loglaw_correction_parameters( lc, lcr, k,        &
                                                 j, inc, wall_index, z0_topo,  &
                                                 kb, direction, ncorr )
             logc_v_n(1,k,i) = lc
             logc_ratio_v_n(1,0:ncorr-1,k,i) = lcr(0:ncorr-1)
             lcr(0:ncorr-1) = 1.0_wp

          ENDDO

       ENDIF
!       
!--    Then vertical walls and corners if necessary
       IF ( topography /= 'flat' )  THEN
!
!--       Workaround, set z0 at vertical surfaces simply to the given roughness
!--       lenth, which is required to determine the logarithmic correction 
!--       factors at the child boundaries, which are at the ghost-points. 
!--       The surface data type for vertical surfaces, however, is not defined 
!--       at ghost-points, so that no z0 can be retrieved at this point. 
!--       Maybe, revise this later and define vertical surface datattype also
!--       at ghost-points.
          z0_topo = roughness_length

          kb = 0       ! kb is not used when direction > 1        
!        
!--       Left boundary
          IF ( bc_dirichlet_l )  THEN
             logc_kbounds_u_l(1:2,nys:nyn) = 0
             logc_kbounds_v_l(1:2,nys:nyn) = 0             
             logc_kbounds_w_l(1:2,nys:nyn) = 0
             
             direction  = 2

             DO  j = nys, nyn
!
!--             Determine the lowest k-indices for u at j,i, j+1,i and j-1,i. 
                i             = 0
                k_wall_u_ji   = get_topography_top_index_ji( j,   i, 'u' )
                k_wall_u_ji_p = get_topography_top_index_ji( j+1, i, 'u' )
                k_wall_u_ji_m = get_topography_top_index_ji( j-1, i, 'u' )
!
!--             Wall for u on the south side.
                IF ( ( k_wall_u_ji <  k_wall_u_ji_m ) .AND.                    &
                     ( k_wall_u_ji >= k_wall_u_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  j
!
!--                Store the kbounds for use in pmci_interp_tril_lr.
                   logc_kbounds_u_l(1,j) = k_wall_u_ji + 1
                   logc_kbounds_u_l(2,j) = k_wall_u_ji_m
                   DO  k = logc_kbounds_u_l(1,j), logc_kbounds_u_l(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_u_l(2,k,j) = 0
                         logc_ratio_u_l(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_u_l(2,k,j) = inc * lc
                         logc_ratio_u_l(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for u on the north side.
                IF ( ( k_wall_u_ji <  k_wall_u_ji_p ) .AND.                    &
                     ( k_wall_u_ji >= k_wall_u_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  j + 1
!
!--                Store the kbounds for use in pmci_interp_tril_lr.                   
                   logc_kbounds_u_l(1,j) = k_wall_u_ji + 1
                   logc_kbounds_u_l(2,j) = k_wall_u_ji_p
                   DO  k = logc_kbounds_u_l(1,j), logc_kbounds_u_l(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_u_l(2,k,j) = 0
                         logc_ratio_u_l(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_u_l(2,k,j) = inc * lc
                         logc_ratio_u_l(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Determine the lowest k-indices for w at j,i, j+1,i and j-1,i. 
                i             = -1
                k_wall_w_ji   = get_topography_top_index_ji( j,   i, 'w' )
                k_wall_w_ji_p = get_topography_top_index_ji( j+1, i, 'w' )
                k_wall_w_ji_m = get_topography_top_index_ji( j-1, i, 'w' )
!
!--             Wall for w on the south side.                
                IF ( ( k_wall_w_ji <  k_wall_w_ji_m ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  j
!
!--                Store the kbounds for use in pmci_interp_tril_lr.
                   logc_kbounds_w_l(1,j) = k_wall_w_ji + 1
                   logc_kbounds_w_l(2,j) = k_wall_w_ji_m
                   DO  k = logc_kbounds_w_l(1,j), logc_kbounds_w_l(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_l(2,k,j) = 0
                         logc_ratio_w_l(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_l(2,k,j) = inc * lc
                         logc_ratio_w_l(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for w on the north side.
                IF ( ( k_wall_w_ji <  k_wall_w_ji_p ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  j+1
!
!--                Store the kbounds for use in pmci_interp_tril_lr.
                   logc_kbounds_w_l(1,j) = k_wall_w_ji + 1
                   logc_kbounds_w_l(2,j) = k_wall_w_ji_p
                   DO  k = logc_kbounds_w_l(1,j), logc_kbounds_w_l(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_l(2,k,j) = 0
                         logc_ratio_w_l(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_l(2,k,j) = inc * lc
                         logc_ratio_w_l(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
                   
             ENDDO

          ENDIF   !  IF ( bc_dirichlet_l )
!        
!--       Right boundary
          IF ( bc_dirichlet_r )  THEN
             logc_kbounds_u_r(1:2,nys:nyn) = 0
             logc_kbounds_v_r(1:2,nys:nyn) = 0             
             logc_kbounds_w_r(1:2,nys:nyn) = 0

             direction  = 2
             i  = nx + 1

             DO  j = nys, nyn
!
!--             Determine the lowest k-indices for u at j,i, j+1,i and j-1,i. 
                k_wall_u_ji   = get_topography_top_index_ji( j,   i, 'u' )
                k_wall_u_ji_p = get_topography_top_index_ji( j+1, i, 'u' )
                k_wall_u_ji_m = get_topography_top_index_ji( j-1, i, 'u' )
!
!--             Wall for u on the south side.
                IF ( ( k_wall_u_ji <  k_wall_u_ji_m ) .AND.                    &
                     ( k_wall_u_ji >= k_wall_u_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  j
!
!--                Store the kbounds for use in pmci_interp_tril_lr.                  
                   logc_kbounds_u_r(1,j) = k_wall_u_ji + 1
                   logc_kbounds_u_r(2,j) = k_wall_u_ji_m
                   DO  k = logc_kbounds_u_r(1,j), logc_kbounds_u_r(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction, ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_u_r(2,k,j) = 0
                         logc_ratio_u_r(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_u_r(2,k,j) = inc * lc
                         logc_ratio_u_r(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for u on the south side.
                IF ( ( k_wall_u_ji <  k_wall_u_ji_p ) .AND.                    &
                     ( k_wall_u_ji >= k_wall_u_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  j + 1                 
!
!--                Store the kbounds for use in pmci_interp_tril_lr.                   
                   logc_kbounds_u_r(1,j) = k_wall_u_ji + 1
                   logc_kbounds_u_r(2,j) = k_wall_u_ji_p
                   DO  k = logc_kbounds_u_r(1,j), logc_kbounds_u_r(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_u_r(2,k,j) = 0
                         logc_ratio_u_r(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_u_r(2,k,j) = inc * lc
                         logc_ratio_u_r(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Determine the lowest k-indices for w at j,i, j+1,i and j-1,i. 
                k_wall_w_ji   = get_topography_top_index_ji( j,   i, 'w' )
                k_wall_w_ji_p = get_topography_top_index_ji( j+1, i, 'w' )
                k_wall_w_ji_m = get_topography_top_index_ji( j-1, i, 'w' )
!
!--             Wall for w on the south side.                
                IF ( ( k_wall_w_ji <  k_wall_w_ji_m ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  j
!
!--                Store the kbounds for use in pmci_interp_tril_lr.                    
                   logc_kbounds_w_r(1,j) = k_wall_w_ji + 1
                   logc_kbounds_w_r(2,j) = k_wall_w_ji_m
                   DO  k = logc_kbounds_w_r(1,j), logc_kbounds_w_r(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_r(2,k,j) = 0
                         logc_ratio_w_r(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_r(2,k,j) = inc * lc
                         logc_ratio_w_r(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for w on the north side.
                IF ( ( k_wall_w_ji <  k_wall_w_ji_p ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  j+1
!
!--                Store the kbounds for use in pmci_interp_tril_lr.                   
                   logc_kbounds_w_r(1,j) = k_wall_w_ji + 1
                   logc_kbounds_w_r(2,j) = k_wall_w_ji_p
                   DO  k = logc_kbounds_w_r(1,j), logc_kbounds_w_r(2,j)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, j, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_r(2,k,j) = 0
                         logc_ratio_w_r(2,0:ncorr-1,k,j) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_r(2,k,j) = inc * lc
                         logc_ratio_w_r(2,0:ncorr-1,k,j) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
                   
             ENDDO
             
          ENDIF   !  IF ( bc_dirichlet_r )
!        
!--       South boundary
          IF ( bc_dirichlet_s )  THEN
             logc_kbounds_u_s(1:2,nxl:nxr) = 0
             logc_kbounds_v_s(1:2,nxl:nxr) = 0
             logc_kbounds_w_s(1:2,nxl:nxr) = 0

             direction  = 3

             DO  i = nxl, nxr
!
!--             Determine the lowest k-indices for v at j,i, j,i+1 and j,i-1.
                j             = 0                
                k_wall_v_ji   = get_topography_top_index_ji( j, i,   'v' )
                k_wall_v_ji_p = get_topography_top_index_ji( j, i+1, 'v' )
                k_wall_v_ji_m = get_topography_top_index_ji( j, i-1, 'v' )
!
!--             Wall for v on the left side.
                IF ( ( k_wall_v_ji <  k_wall_v_ji_m ) .AND.                    &
                     ( k_wall_v_ji >= k_wall_v_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  i
!
!--                Store the kbounds for use in pmci_interp_tril_sn.                    
                   logc_kbounds_v_s(1,i) = k_wall_v_ji + 1
                   logc_kbounds_v_s(2,i) = k_wall_v_ji_m
                   DO  k = logc_kbounds_v_s(1,i), logc_kbounds_v_s(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_v_s(2,k,i) = 0
                         logc_ratio_v_s(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_v_s(2,k,i) = inc * lc
                         logc_ratio_v_s(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for v on the right side.
                IF ( ( k_wall_v_ji <  k_wall_v_ji_p ) .AND.                    &
                     ( k_wall_v_ji >= k_wall_v_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  i+1
!
!--                Store the kbounds for use in pmci_interp_tril_sn.                   
                   logc_kbounds_v_s(1,i) = k_wall_v_ji + 1
                   logc_kbounds_v_s(2,i) = k_wall_v_ji_p
                   DO  k = logc_kbounds_v_s(1,i), logc_kbounds_v_s(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_v_s(2,k,i) = 0
                         logc_ratio_v_s(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_v_s(2,k,i) = inc * lc
                         logc_ratio_v_s(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Determine the lowest k-indices for w at j,i, j,i+1 and j,i-1. 
                j             = -1
                k_wall_w_ji   = get_topography_top_index_ji( j, i,   'w' )
                k_wall_w_ji_p = get_topography_top_index_ji( j, i+1, 'w' )
                k_wall_w_ji_m = get_topography_top_index_ji( j, i-1, 'w' )
!
!--             Wall for w on the left side.
                IF ( ( k_wall_w_ji <  k_wall_w_ji_m ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_p ) )  THEN
                   inc        =  1
                   wall_index =  i
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_w_s(1,i) = k_wall_w_ji + 1
                   logc_kbounds_w_s(2,i) = k_wall_w_ji_m
                   DO  k = logc_kbounds_w_s(1,i), logc_kbounds_w_s(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_s(2,k,i) = 0
                         logc_ratio_w_s(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_s(2,k,i) = inc * lc
                         logc_ratio_w_s(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for w on the right side.
                IF ( ( k_wall_w_ji <  k_wall_w_ji_p ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  i+1
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_w_s(1,i) = k_wall_w_ji + 1
                   logc_kbounds_w_s(2,i) = k_wall_w_ji_p
                   DO  k = logc_kbounds_w_s(1,i), logc_kbounds_w_s(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_s(2,k,i) = 0
                         logc_ratio_w_s(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_s(2,k,i) = inc * lc
                         logc_ratio_w_s(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF

             ENDDO

          ENDIF   !  IF (bc_dirichlet_s )
!        
!--       North boundary
          IF ( bc_dirichlet_n )  THEN
             logc_kbounds_u_n(1:2,nxl:nxr) = 0             
             logc_kbounds_v_n(1:2,nxl:nxr) = 0
             logc_kbounds_w_n(1:2,nxl:nxr) = 0

             direction  = 3
             j  = ny + 1

             DO  i = nxl, nxr
!
!--             Determine the lowest k-indices for v at j,i, j,i+1 and j,i-1
                k_wall_v_ji   = get_topography_top_index_ji( j, i,   'v' )
                k_wall_v_ji_p = get_topography_top_index_ji( j, i+1, 'v' )
                k_wall_v_ji_m = get_topography_top_index_ji( j, i-1, 'v' )
!
!--             Wall for v on the left side.
                IF ( ( k_wall_v_ji <  k_wall_v_ji_m ) .AND.                    &
                     ( k_wall_v_ji >= k_wall_v_ji_p ) )  THEN
                   inc        = 1
                   wall_index = i                   
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_v_n(1,i) = k_wall_v_ji + 1
                   logc_kbounds_v_n(2,i) = k_wall_v_ji_m
                   DO  k = logc_kbounds_v_n(1,i), logc_kbounds_v_n(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_v_n(2,k,i) = 0
                         logc_ratio_v_n(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_v_n(2,k,i) = inc * lc
                         logc_ratio_v_n(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for v on the right side.
                IF ( ( k_wall_v_ji <  k_wall_v_ji_p ) .AND.                    &
                     ( k_wall_v_ji >= k_wall_v_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  i + 1
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_v_n(1,i) = k_wall_v_ji + 1
                   logc_kbounds_v_n(2,i) = k_wall_v_ji_p
                   DO  k = logc_kbounds_v_n(1,i), logc_kbounds_v_n(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_v_n(2,k,i) = 0
                         logc_ratio_v_n(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_v_n(2,k,i) = inc * lc
                         logc_ratio_v_n(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Determine the lowest k-indices for w at j,i, j,i+1 and j,i-1.
                k_wall_w_ji   = get_topography_top_index_ji( j, i,   'w' )
                k_wall_w_ji_p = get_topography_top_index_ji( j, i+1, 'w' )
                k_wall_w_ji_m = get_topography_top_index_ji( j, i-1, 'w' )                   
!
!--             Wall for w on the left side.
                IF ( ( k_wall_w_ji <  k_wall_w_ji_m ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_p ) )  THEN
                   inc        = 1
                   wall_index = i
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_w_n(1,i) = k_wall_w_ji + 1
                   logc_kbounds_w_n(2,i) = k_wall_w_ji_m
                   DO  k = logc_kbounds_w_n(1,i), logc_kbounds_w_n(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_n(2,k,i) = 0
                         logc_ratio_w_n(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_n(2,k,i) = inc * lc
                         logc_ratio_w_n(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF
!
!--             Wall for w on the right side, but not on the left side
                IF ( ( k_wall_w_ji <  k_wall_w_ji_p ) .AND.                    &
                     ( k_wall_w_ji >= k_wall_w_ji_m ) )  THEN
                   inc        = -1
                   wall_index =  i+1
!
!--                Store the kbounds for use in pmci_interp_tril_sn.
                   logc_kbounds_w_n(1,i) = k_wall_w_ji + 1
                   logc_kbounds_w_n(2,i) = k_wall_w_ji_p
                   DO  k = logc_kbounds_w_n(1,i), logc_kbounds_w_n(2,i)
                      CALL pmci_define_loglaw_correction_parameters( lc, lcr,  &
                           k, i, inc, wall_index, z0_topo, kb, direction,      &
                           ncorr )
                      IF ( lc == -99 )  THEN
!                         
!--                      The pivot point is out of subdomain, skip the correction. 
                         logc_w_n(2,k,i) = 0
                         logc_ratio_w_n(2,0:ncorr-1,k,i) = 1.0_wp
                      ELSE
!
!--                      The direction of the wall-normal index is stored as the
!--                      sign of the logc-element.
                         logc_w_n(2,k,i) = inc * lc
                         logc_ratio_w_n(2,0:ncorr-1,k,i) = lcr(0:ncorr-1)
                      ENDIF
                      lcr(0:ncorr-1) = 1.0_wp
                   ENDDO
                ENDIF

             ENDDO

          ENDIF   !  IF ( bc_dirichlet_n )

       ENDIF   !  IF ( topography /= 'flat' )

    END SUBROUTINE pmci_init_loglaw_correction



    SUBROUTINE pmci_define_loglaw_correction_parameters( lc, lcr, k, ij, inc,  &
         wall_index, z0_l, kb, direction, ncorr )

       IMPLICIT NONE

       INTEGER(iwp), INTENT(IN)  ::  direction                 !<
       INTEGER(iwp), INTENT(IN)  ::  ij                        !<
       INTEGER(iwp), INTENT(IN)  ::  inc                       !<
       INTEGER(iwp), INTENT(IN)  ::  k                         !<
       INTEGER(iwp), INTENT(IN)  ::  kb                        !<
       INTEGER(iwp), INTENT(OUT) ::  lc                        !<
       INTEGER(iwp), INTENT(IN)  ::  ncorr                     !<
       INTEGER(iwp), INTENT(IN)  ::  wall_index                !<

       INTEGER(iwp) ::  alcorr                                 !<
       INTEGER(iwp) ::  corr_index                             !<
       INTEGER(iwp) ::  lcorr                                  !<

       LOGICAL      ::  more                                   !<             

       REAL(wp), DIMENSION(0:ncorr-1), INTENT(INOUT) ::  lcr   !<
       REAL(wp), INTENT(IN)      ::  z0_l                      !<
      
       REAL(wp)     ::  logvelc1                               !<
      

       SELECT CASE ( direction )

          CASE (1)   !  k
             more = .TRUE.
             lcorr = 0
             DO  WHILE ( more .AND. lcorr <= ncorr-1 )
                corr_index = k + lcorr
                IF ( lcorr == 0 )  THEN
                   CALL pmci_find_logc_pivot_k( lc, logvelc1, z0_l, kb )
                ENDIF
               
                IF ( corr_index < lc )  THEN
                   lcr(lcorr) = LOG( ( zu(k) - zw(kb) ) / z0_l ) / logvelc1
                   more = .TRUE.
                ELSE
                   lcr(lcorr) = 1.0_wp
                   more = .FALSE.
                ENDIF
                lcorr = lcorr + 1
             ENDDO

          CASE (2)   !  j
             more = .TRUE.
             lcorr  = 0
             alcorr = 0
             DO  WHILE ( more  .AND.  alcorr <= ncorr-1 )
                corr_index = ij + lcorr   ! In this case (direction = 2) ij is j
                IF ( lcorr == 0 )  THEN
                   CALL pmci_find_logc_pivot_j( lc, logvelc1, ij, wall_index,  &
                                                z0_l, inc )
                ENDIF
!
!--             The role of inc here is to make the comparison operation "<"
!--             valid in both directions
                IF ( ( inc * corr_index < inc * lc ) .AND. ( lc /= -99 ) )  THEN
                   lcr(alcorr) = LOG( ABS( coord_y(corr_index) + 0.5_wp * dy   &
                                         - coord_y(wall_index) ) / z0_l )      &
                                 / logvelc1
                   more = .TRUE.
                ELSE
                   lcr(alcorr) = 1.0_wp
                   more = .FALSE.
                ENDIF
                lcorr  = lcorr + inc
                alcorr = ABS( lcorr )
             ENDDO

          CASE (3)   !  i
             more = .TRUE.
             lcorr  = 0
             alcorr = 0
             DO  WHILE ( more  .AND.  alcorr <= ncorr-1 )
                corr_index = ij + lcorr   ! In this case (direction = 3) ij is i
                IF ( lcorr == 0 )  THEN
                   CALL pmci_find_logc_pivot_i( lc, logvelc1, ij, wall_index,  &
                                                z0_l, inc )
                ENDIF
!
!--             The role of inc here is to make the comparison operation "<"
!--             valid in both directions
                IF ( ( inc * corr_index < inc * lc ) .AND. ( lc /= -99 ) )  THEN
                   lcr(alcorr) = LOG( ABS( coord_x(corr_index) + 0.5_wp * dx   &
                                         - coord_x(wall_index) ) / z0_l )      &
                                 / logvelc1
                   more = .TRUE.
                ELSE
                   lcr(alcorr) = 1.0_wp
                   more = .FALSE.
                ENDIF
                lcorr  = lcorr + inc
                alcorr = ABS( lcorr )
             ENDDO

       END SELECT

    END SUBROUTINE pmci_define_loglaw_correction_parameters



    SUBROUTINE pmci_find_logc_pivot_k( lc, logzc1, z0_l, kb )
!
!--    Finds the pivot node and the log-law factor for near-wall nodes for
!--    which the wall-parallel velocity components will be log-law corrected
!--    after interpolation. This subroutine is only for horizontal walls.

       IMPLICIT NONE

       INTEGER(iwp), INTENT(IN)  ::  kb   !<
       INTEGER(iwp), INTENT(OUT) ::  lc   !<

       INTEGER(iwp) ::  kbc               !<
       INTEGER(iwp) ::  k1                !<

       REAL(wp), INTENT(OUT) ::  logzc1   !<
       REAL(wp), INTENT(IN)  ::  z0_l     !<

       REAL(wp) ::  zuc1                  !<

!
!--    kbc is the first coarse-grid point above the surface
       kbc = nzb + 1
       DO  WHILE ( cg%zu(kbc) < zu(kb) )
          kbc = kbc + 1
       ENDDO
       zuc1  = cg%zu(kbc)
       k1    = kb + 1
       DO  WHILE ( zu(k1) < zuc1 )  !  Important: must be <, not <=
          k1 = k1 + 1
       ENDDO
       logzc1 = LOG( (zu(k1) - zw(kb) ) / z0_l )
       lc = k1

    END SUBROUTINE pmci_find_logc_pivot_k



    SUBROUTINE pmci_find_logc_pivot_j( lc, logyc1, j, jw, z0_l, inc )
!
!--    Finds the pivot node and the log-law factor for near-wall nodes for
!--    which the wall-parallel velocity components will be log-law corrected
!--    after interpolation. This subroutine is only for vertical walls on
!--    south/north sides of the node. If the pivot node is found to be outside 
!--    the subdomain, a marker value of -99 is set to lc and this tells 
!--    pmci_init_loglaw_correction to not do the correction in this case.
      
       IMPLICIT NONE

       INTEGER(iwp), INTENT(IN)  ::  inc    !<  increment must be 1 or -1.
       INTEGER(iwp), INTENT(IN)  ::  j      !<
       INTEGER(iwp), INTENT(IN)  ::  jw     !<
       INTEGER(iwp), INTENT(OUT) ::  lc     !<

       INTEGER(iwp) ::  jwc                 !<
       INTEGER(iwp) ::  j1                  !<

       REAL(wp), INTENT(IN)  ::  z0_l       !<
       REAL(wp), INTENT(OUT) ::  logyc1     !<

       REAL(wp) ::  ycb                     !<       
       REAL(wp) ::  yc1                     !<
       
!
!--    yc1 is the y-coordinate of the first coarse-grid u- and w-nodes out from
!--    the wall. Here we assume that the wall index in the coarse grid is the
!--    closest one if they don't match.
       jwc  = pmci_find_nearest_coarse_grid_index_j( jw )
       yc1  = cg%coord_y(jwc) - lower_left_coord_y + 0.5_wp * inc * cg%dy
!       
!--    Check if yc1 is out of the subdomain y-range. In such case set the marker 
!--    value -99 for lc in order to skip the loglaw-correction locally.
       IF ( yc1 < ( REAL( nysg, KIND=wp ) + 0.5_wp ) * dy  )  THEN
          lc = -99
          logyc1 = 1.0_wp
       ELSE IF ( yc1 > ( REAL( nyng, KIND=wp ) + 0.5_wp ) * dy )  THEN
          lc = -99
          logyc1 = 1.0_wp
       ELSE
!
!--       j1 is the first fine-grid index further away from the wall than yc1
          j1 = j
!
!--       Important: the binary relation must be <, not <=
          ycb = 0.5_wp * dy - lower_left_coord_y
          DO  WHILE ( inc * ( coord_y(j1) + ycb ) < inc * yc1 )
             j1 = j1 + inc
          ENDDO
          
          logyc1 = LOG( ABS( coord_y(j1) + 0.5_wp * dy - coord_y(jw) ) / z0_l )
          lc = j1
       ENDIF
       
    END SUBROUTINE pmci_find_logc_pivot_j



    SUBROUTINE pmci_find_logc_pivot_i( lc, logxc1, i, iw, z0_l, inc )
!
!--    Finds the pivot node and the log-law factor for near-wall nodes for
!--    which the wall-parallel velocity components will be log-law corrected
!--    after interpolation. This subroutine is only for vertical walls on
!--    left/right sides of the node. If the pivot node is found to be outside 
!--    the subdomain, a marker value of -99 is set to lc and this tells 
!--    pmci_init_loglaw_correction to not do the correction in this case.

       IMPLICIT NONE

       INTEGER(iwp), INTENT(IN)  ::  i      !<
       INTEGER(iwp), INTENT(IN)  ::  inc    !< increment must be 1 or -1.
       INTEGER(iwp), INTENT(IN)  ::  iw     !<
       INTEGER(iwp), INTENT(OUT) ::  lc     !<

       INTEGER(iwp) ::  iwc                 !<
       INTEGER(iwp) ::  i1                  !<

       REAL(wp), INTENT(IN)  ::  z0_l       !<
       REAL(wp), INTENT(OUT) ::  logxc1     !<

       REAL(wp) ::  xcb                     !<
       REAL(wp) ::  xc1                     !<

!
!--    xc1 is the x-coordinate of the first coarse-grid v- and w-nodes out from
!--    the wall. Here we assume that the wall index in the coarse grid is the
!--    closest one if they don't match.
       iwc  = pmci_find_nearest_coarse_grid_index_i( iw )
       xc1  = cg%coord_x(iwc) - lower_left_coord_x + 0.5_wp * inc * cg%dx
!       
!--    Check if xc1 is out of the subdomain x-range. In such case set the marker 
!--    value -99 for lc in order to skip the loglaw-correction locally.        
       IF ( xc1 < ( REAL( nxlg, KIND=wp ) + 0.5_wp ) * dx  )  THEN
          lc = -99
          logxc1 = 1.0_wp
       ELSE IF ( xc1 > ( REAL( nxrg, KIND=wp ) + 0.5_wp ) * dx )  THEN
          lc = -99
          logxc1 = 1.0_wp
       ELSE    
!
!--       i1 is the first fine-grid index futher away from the wall than xc1.
          i1 = i
!
!--       Important: the binary relation must be <, not <= 
          xcb = 0.5_wp * dx - lower_left_coord_x
          DO  WHILE ( inc * ( coord_x(i1) + xcb ) < inc * xc1 )
             i1 = i1 + inc
          ENDDO
      
          logxc1 = LOG( ABS( coord_x(i1) + 0.5_wp*dx - coord_x(iw) ) / z0_l )
          lc = i1
       ENDIF
       
    END SUBROUTINE pmci_find_logc_pivot_i


    
    FUNCTION pmci_find_nearest_coarse_grid_index_j( jw ) 

      IMPLICIT NONE
      INTEGER(iwp) :: jw         !< Fine-grid wall index

      INTEGER(iwp) :: jc
      INTEGER(iwp) :: pmci_find_nearest_coarse_grid_index_j
      REAL(wp) :: dist
      REAL(wp) :: newdist

      
      dist = coord_y(nyn) - coord_y(nys)
      DO jc = jcs, jcn
         newdist = ABS( cg%coord_y(jc) - coord_y(jw) )
         IF ( newdist < dist )  THEN
            pmci_find_nearest_coarse_grid_index_j = jc
            dist = newdist
         ENDIF
      ENDDO
      
    END FUNCTION pmci_find_nearest_coarse_grid_index_j


    
    FUNCTION pmci_find_nearest_coarse_grid_index_i( iw ) 

      IMPLICIT NONE
      INTEGER(iwp) :: iw         !< Fine-grid wall index

      INTEGER(iwp) :: ic
      INTEGER(iwp) :: pmci_find_nearest_coarse_grid_index_i
      REAL(wp) :: dist
      REAL(wp) :: newdist

      
      dist = coord_x(nxr) - coord_x(nxl)
      DO ic = icl, icr
         newdist = ABS( cg%coord_x(ic) - coord_x(iw) )
         IF ( newdist < dist )  THEN
            pmci_find_nearest_coarse_grid_index_i = ic
            dist = newdist
         ENDIF
      ENDDO
      
    END FUNCTION pmci_find_nearest_coarse_grid_index_i

    
      
    SUBROUTINE pmci_init_anterp_tophat
!
!--    Precomputation of the child-array indices for
!--    corresponding coarse-grid array index and the
!--    Under-relaxation coefficients to be used by anterp_tophat.

       IMPLICIT NONE

       INTEGER(iwp) ::  i        !< Fine-grid index
       INTEGER(iwp) ::  ii       !< Coarse-grid index
       INTEGER(iwp) ::  istart   !<
       INTEGER(iwp) ::  ir       !<
       INTEGER(iwp) ::  iw       !< Fine-grid index limited to -1 <= iw <= nx+1
       INTEGER(iwp) ::  j        !< Fine-grid index
       INTEGER(iwp) ::  jj       !< Coarse-grid index
       INTEGER(iwp) ::  jstart   !<
       INTEGER(iwp) ::  jr       !<
       INTEGER(iwp) ::  jw       !< Fine-grid index limited to -1 <= jw <= ny+1
       INTEGER(iwp) ::  k        !< Fine-grid index
       INTEGER(iwp) ::  kk       !< Coarse-grid index
       INTEGER(iwp) ::  kstart   !<
       INTEGER(iwp) ::  kw       !< Fine-grid index limited to kw <= nzt+1
       REAL(wp)     ::  xi       !<
       REAL(wp)     ::  eta      !<
       REAL(wp)     ::  tolerance !<
       REAL(wp)     ::  zeta     !<
     
!
!--    Default values for under-relaxation lengths:
       IF ( anterp_relax_length_l < 0.0_wp )  THEN
          anterp_relax_length_l = 0.1_wp * ( nx + 1 ) * dx
       ENDIF
       IF ( anterp_relax_length_r < 0.0_wp )  THEN
          anterp_relax_length_r = 0.1_wp * ( nx + 1 ) * dx
       ENDIF
       IF ( anterp_relax_length_s < 0.0_wp )  THEN
          anterp_relax_length_s = 0.1_wp * ( ny + 1 ) * dy
       ENDIF
       IF ( anterp_relax_length_n < 0.0_wp )  THEN
          anterp_relax_length_n = 0.1_wp * ( ny + 1 ) * dy
       ENDIF
       IF ( anterp_relax_length_t < 0.0_wp )  THEN
          anterp_relax_length_t = 0.1_wp * zu(nzt)
       ENDIF
!
!--    First determine kcto and kctw which refer to the uppermost 
!--    coarse-grid levels below the child top-boundary level.
       kk = 0
       DO  WHILE ( cg%zu(kk) <= zu(nzt) )
          kk = kk + 1
       ENDDO
       kcto = kk - 1

       kk = 0
       DO  WHILE ( cg%zw(kk) <= zw(nzt-1) )
          kk = kk + 1
       ENDDO
       kctw = kk - 1
!AH
       write(9,"('kcto, kctw = ', 2(i3,2x))") kcto, kctw

!AH
!       ALLOCATE( iflu(icl:icr) )
!       ALLOCATE( iflo(icl:icr) )
!       ALLOCATE( ifuu(icl:icr) )
!       ALLOCATE( ifuo(icl:icr) )
!       ALLOCATE( jflv(jcs:jcn) )
!       ALLOCATE( jflo(jcs:jcn) )
!       ALLOCATE( jfuv(jcs:jcn) )
!       ALLOCATE( jfuo(jcs:jcn) )
!       
!       ALLOCATE( iflu(icl-1:icr+1) )
!       ALLOCATE( iflo(icl-1:icr+1) )
!       ALLOCATE( ifuu(icl-1:icr+1) )
!       ALLOCATE( ifuo(icl-1:icr+1) )
!       ALLOCATE( jflv(jcs-1:jcn+1) )
!       ALLOCATE( jflo(jcs-1:jcn+1) )
!       ALLOCATE( jfuv(jcs-1:jcn+1) )
!       ALLOCATE( jfuo(jcs-1:jcn+1) )
!
       icla = coarse_bound_aux(1)
       icra = coarse_bound_aux(2)
       jcsa = coarse_bound_aux(3)
       jcna = coarse_bound_aux(4)
       ALLOCATE( iflu(icla:icra) )
       ALLOCATE( iflo(icla:icra) )
       ALLOCATE( ifuu(icla:icra) )
       ALLOCATE( ifuo(icla:icra) )
       ALLOCATE( jflv(jcsa:jcna) )
       ALLOCATE( jflo(jcsa:jcna) )
       ALLOCATE( jfuv(jcsa:jcna) )
       ALLOCATE( jfuo(jcsa:jcna) )       
!AH
       ALLOCATE( kflw(0:cg%nz+1) )
       ALLOCATE( kflo(0:cg%nz+1) )
       ALLOCATE( kfuw(0:cg%nz+1) )
       ALLOCATE( kfuo(0:cg%nz+1) )
!AH
!       ALLOCATE( ijkfc_u(0:cg%nz+1,jcs:jcn,icl:icr) )
!       ALLOCATE( ijkfc_v(0:cg%nz+1,jcs:jcn,icl:icr) )
!       ALLOCATE( ijkfc_w(0:cg%nz+1,jcs:jcn,icl:icr) )
!       ALLOCATE( ijkfc_s(0:cg%nz+1,jcs:jcn,icl:icr) )
!
!       ALLOCATE( ijkfc_u(0:cg%nz+1,jcs-1:jcn+1,icl-1:icr+1) )
!       ALLOCATE( ijkfc_v(0:cg%nz+1,jcs-1:jcn+1,icl-1:icr+1) )
!       ALLOCATE( ijkfc_w(0:cg%nz+1,jcs-1:jcn+1,icl-1:icr+1) )
!       ALLOCATE( ijkfc_s(0:cg%nz+1,jcs-1:jcn+1,icl-1:icr+1) )
!
       ALLOCATE( ijkfc_u(0:cg%nz+1,jcsa:jcna,icla:icra) )
       ALLOCATE( ijkfc_v(0:cg%nz+1,jcsa:jcna,icla:icra) )
       ALLOCATE( ijkfc_w(0:cg%nz+1,jcsa:jcna,icla:icra) )
       ALLOCATE( ijkfc_s(0:cg%nz+1,jcsa:jcna,icla:icra) )
!AH

       ijkfc_u = 0
       ijkfc_v = 0
       ijkfc_w = 0
       ijkfc_s = 0
!
!--    i-indices of u for each ii-index value
!--    ii=icr is redundant for anterpolation
       tolerance = 0.000001_wp * dx
       istart = nxlg
!AH       DO  ii = icl, icr
!AH       DO  ii = icl-1, icr+1
          DO  ii = icla, icra

!
!--       In case the child and parent grid lines match in x 
!--       use only the local k,j-child-grid plane for the anterpolation, 
!--       i.e use 2-D anterpolation. Else, use a 3-D anterpolation.
          i = istart
          DO WHILE ( coord_x(i) < cg%coord_x(ii) - tolerance  .AND.  i < nxrg )
             i = i + 1
          ENDDO
          IF ( ABS( coord_x(i) - cg%coord_x(ii) ) < tolerance )  THEN
             i = istart
             DO WHILE ( coord_x(i) < cg%coord_x(ii) .AND. i < nxrg )
                i = i + 1
             ENDDO
             iflu(ii) = MIN( MAX( i, nxlg ), nxrg )
             ifuu(ii) = iflu(ii)
             istart   = iflu(ii)
          ELSE
             i = istart
             DO  WHILE ( ( coord_x(i) < cg%coord_x(ii) - 0.5_wp * cg%dx )       &
                  .AND.  ( i < nxrg ) )
                i  = i + 1
             ENDDO
             iflu(ii) = MIN( MAX( i, nxlg ), nxrg )
             ir = i
             DO  WHILE ( ( coord_x(ir) <= cg%coord_x(ii) + 0.5_wp * cg%dx )     &
                  .AND.  ( i < nxrg+1 ) )
                i  = i + 1
                ir = MIN( i, nxrg )
             ENDDO
             ifuu(ii) = MIN( MAX( i-1, iflu(ii) ), nxrg )
             istart = iflu(ii)
          ENDIF
!
!--       Print out the index bounds for checking and debugging purposes
          write(9,"('pmci_init_anterp_tophat, ii, iflu, ifuu: ', 3(i4,2x))")    &
               ii, iflu(ii), ifuu(ii)
          flush(9)

       ENDDO
       write(9,*)
!
!--    i-indices of others for each ii-index value
!--    ii=icr is redundant for anterpolation
       istart = nxlg
!AH       DO  ii = icl, icr
!AH       DO  ii = icl-1, icr+1
       DO  ii = icla, icra
          i = istart
          DO  WHILE ( ( coord_x(i) + 0.5_wp * dx < cg%coord_x(ii) )  .AND.     &
                      ( i < nxrg ) )
             i  = i + 1
          ENDDO
          iflo(ii) = MIN( MAX( i, nxlg ), nxrg )
          ir = i
          DO  WHILE ( ( coord_x(ir) + 0.5_wp * dx <= cg%coord_x(ii) + cg%dx )  &
                      .AND.  ( i < nxrg+1 ) )
             i  = i + 1
             ir = MIN( i, nxrg )
          ENDDO
          ifuo(ii) = MIN( MAX( i-1, iflo(ii) ), nxrg )
          istart = iflo(ii)
!
!--       Print out the index bounds for checking and debugging purposes
          write(9,"('pmci_init_anterp_tophat, ii, iflo, ifuo: ', 3(i4,2x))")    &
               ii, iflo(ii), ifuo(ii)
          flush(9)
       ENDDO
       write(9,*)
!
!--    j-indices of v for each jj-index value
!--    jj=jcn is redundant for anterpolation
       tolerance = 0.000001_wp * dy
       jstart = nysg
!AH       DO  jj = jcs, jcn
!AH       DO  jj = jcs-1, jcn+1
       DO  jj = jcsa, jcna
!
!--       In case the child and parent grid lines match in y 
!--       use only the local k,i-child-grid plane for the anterpolation, 
!--       i.e use 2-D anterpolation. Else, use a 3-D anterpolation.
          j = jstart
          DO WHILE ( coord_y(j) < cg%coord_y(jj) - tolerance  .AND.  j < nyng )
             j = j + 1
          ENDDO
          IF ( ABS( coord_y(j) - cg%coord_y(jj) ) < tolerance )  THEN
             j = jstart
             DO WHILE ( coord_y(j) < cg%coord_y(jj) .AND. j < nyng )
                j = j + 1
             ENDDO
             jflv(jj) = MIN( MAX( j, nysg ), nyng )
             jfuv(jj) = jflv(jj)
             jstart   = jflv(jj)
          ELSE
             j = jstart
             DO  WHILE ( ( coord_y(j) < cg%coord_y(jj) - 0.5_wp * cg%dy )       &
                  .AND.  ( j < nyng ) )
                j  = j + 1
             ENDDO
             jflv(jj) = MIN( MAX( j, nysg ), nyng )
             jr = j
             DO  WHILE ( ( coord_y(jr) <= cg%coord_y(jj) + 0.5_wp * cg%dy )     &
                  .AND.  ( j < nyng+1 ) )
                j  = j + 1
                jr = MIN( j, nyng )
             ENDDO
             jfuv(jj) = MIN( MAX( j-1, jflv(jj) ), nyng )
             jstart = jflv(jj)
          ENDIF
!
!--       Print out the index bounds for checking and debugging purposes
          write(9,"('pmci_init_anterp_tophat, jj, jflv, jfuv: ', 3(i4,2x))")    &
               jj, jflv(jj), jfuv(jj)
          flush(9)
       ENDDO
       write(9,*)
!
!--    j-indices of others for each jj-index value
!--    jj=jcn is redundant for anterpolation
       jstart = nysg
!AH       DO  jj = jcs, jcn
!AH       DO  jj = jcs-1, jcn+1
       DO  jj = jcsa, jcna
          j = jstart
          DO  WHILE ( ( coord_y(j) + 0.5_wp * dy < cg%coord_y(jj) )  .AND.     &
                      ( j < nyng ) )
             j  = j + 1
          ENDDO
          jflo(jj) = MIN( MAX( j, nysg ), nyng )
          jr = j
          DO  WHILE ( ( coord_y(jr) + 0.5_wp * dy <= cg%coord_y(jj) + cg%dy )  &
                      .AND.  ( j < nyng+1 ) )
             j  = j + 1
             jr = MIN( j, nyng )
          ENDDO
          jfuo(jj) = MIN( MAX( j-1, jflo(jj) ), nyng )
          jstart = jflo(jj)
!
!--       Print out the index bounds for checking and debugging purposes
          write(9,"('pmci_init_anterp_tophat, jj, jflo, jfuo: ', 3(i4,2x))")    &
               jj, jflo(jj), jfuo(jj)
          flush(9)
       ENDDO
       write(9,*)
!
!--    k-indices of w for each kk-index value
!--    Note that anterpolation index limits are needed also for the top boundary 
!--    ghost cell level because of the reversibility correction in the interpolation.
       kstart  = 0
       kflw(0) = 0
       kfuw(0) = 0
       tolerance = 0.000001_wp * dzw(1)
       DO kk = 1, cg%nz+1
!
!--       In case the child and parent grid lines match in z
!--       use only the local j,i-child-grid plane for the anterpolation, 
!--       i.e use 2-D anterpolation. Else, use a 3-D anterpolation.
          k = kstart
          DO WHILE ( zw(k) < cg%zw(kk) - tolerance  .AND.  k < nzt+1 )
             k = k + 1
          ENDDO
          IF ( ABS( zw(k) - cg%zw(kk) ) < tolerance )  THEN
             k = kstart
             DO WHILE ( ( zw(k) < cg%zw(kk) )  .AND.  ( k < nzt+1 ) )
                k = k + 1
             ENDDO
             kflw(kk) = MIN( MAX( k, 1 ), nzt + 1 )
             kfuw(kk) = kflw(kk)
             kstart   = kflw(kk)
          ELSE
             k = kstart
             DO  WHILE ( ( zw(k) < cg%zu(kk) )  .AND.  ( k < nzt ) )
                k = k + 1
             ENDDO
             kflw(kk) = MIN( MAX( k, 1 ), nzt + 1 )
             IF ( kk+1 <= cg%nz+1 )  THEN
                DO  WHILE ( ( zw(k) <= cg%zu(kk+1) )  .AND.  ( k < nzt+1 ) )
                   k  = k + 1
                   IF ( k > nzt + 1 ) EXIT  ! This EXIT is to prevent zu(k) from flowing over.
                ENDDO
                kfuw(kk) = MIN( MAX( k-1, kflw(kk) ), nzt + 1 )
             ELSE
                kfuw(kk) = kflw(kk)
             ENDIF
             kstart = kflw(kk)
          ENDIF
!AH
          write(9,"('pmci_init_anterp_tophat, kk, kflw, kfuw: ', 4(i4,2x), 2(e12.5,2x))") &
               kk, kflw(kk), kfuw(kk), nzt,  cg%zu(kk), cg%zw(kk)
          flush(9)
       ENDDO
       write(9,*)
!
!--    k-indices of others for each kk-index value
       kstart  = 0
       kflo(0) = 0
       kfuo(0) = 0
!
!--    Note that anterpolation index limits are needed also for the top boundary 
!--    ghost cell level because of the reversibility correction in the interpolation.
       DO  kk = 1, cg%nz+1
          k = kstart
          DO  WHILE ( ( zu(k) < cg%zw(kk-1) )  .AND.  ( k <= nzt ) )
             k = k + 1
          ENDDO
          kflo(kk) = MIN( MAX( k, 1 ), nzt + 1 )
          DO  WHILE ( ( zu(k) <= cg%zw(kk) )  .AND.  ( k <= nzt+1 ) )
             k = k + 1
             IF ( k > nzt + 1 ) EXIT  ! This EXIT is to prevent zu(k) from flowing over.
          ENDDO
          kfuo(kk) = MIN( MAX( k-1, kflo(kk) ), nzt + 1 )
          kstart = kflo(kk)
       ENDDO
!
!--    Set the k-index bounds separately for the parent-grid cells cg%nz and cg%nz+1. 
!--    Index bounds for cg%nz are needed for the reversibility correction.
       kflo(cg%nz)   = nzt+1    ! Needed for the reversibility correction
       kfuo(cg%nz)   = nzt+kgsr ! Needed for the reversibility correction
       kflo(cg%nz+1) = nzt+kgsr ! Obsolete
       kfuo(cg%nz+1) = nzt+kgsr ! Obsolete 

       DO  kk = 1, cg%nz+1
          write(9,"('pmci_init_anterp_tophat, kk, kflo, kfuo: ', 4(i4,2x), 2(e12.5,2x))") &
               kk, kflo(kk), kfuo(kk), nzt,  cg%zu(kk), cg%zw(kk)
          flush(9)
       ENDDO
       write(9,*)
!AH

!
!--    Precomputation of number of fine-grid nodes inside parent-grid cells.
!--    Note that ii, jj, and kk are parent-grid indices. 
!--    This information is needed in anterpolation and in reversibility 
!--    correction in interpolation. For the reversibility correction, ijkfc-
!--    information is needed also beyond the indices for wall_flags_0-masking 
!--    must be modified here in the boundary-normal direction (iw, jw, kw). 
!AH       DO  ii = icl, icr
!AH          DO  jj = jcs, jcn
!AH       DO  ii = icl-1, icr+1
!AH          DO  jj = jcs-1, jcn+1
       DO  ii = icla, icra
          DO  jj = jcsa, jcna
             DO kk = 0, cg%nz+1
!
!--             u-component
                DO  i = iflu(ii), ifuu(ii)
                   iw = MAX( MIN( i, nx+1 ), -1 )
                   DO  j = jflo(jj), jfuo(jj)
                      jw = MAX( MIN( j, ny+1 ), -1 )
                      DO  k = kflo(kk), kfuo(kk)
                         kw = MIN( k, nzt+1 )                
                         ijkfc_u(kk,jj,ii) = ijkfc_u(kk,jj,ii)                  &
                              + MERGE( 1, 0, BTEST( wall_flags_0(kw,jw,iw), 1 ) )
                      ENDDO
                   ENDDO
                ENDDO
!
!--             v-component 
                DO  i = iflo(ii), ifuo(ii)
                   iw = MAX( MIN( i, nx+1 ), -1 )
                   DO  j = jflv(jj), jfuv(jj)
                      jw = MAX( MIN( j, ny+1 ), -1 )
                      DO  k = kflo(kk), kfuo(kk)
                         kw = MIN( k, nzt+1 )                                        
                         ijkfc_v(kk,jj,ii) = ijkfc_v(kk,jj,ii)                  &
                              + MERGE( 1, 0, BTEST( wall_flags_0(kw,jw,iw), 2 ) )
                      ENDDO
                   ENDDO
                ENDDO
!
!--             scalars
                DO  i = iflo(ii), ifuo(ii)
                   iw = MAX( MIN( i, nx+1 ), -1 )
                   DO  j = jflo(jj), jfuo(jj)
                      jw = MAX( MIN( j, ny+1 ), -1 )
                      DO  k = kflo(kk), kfuo(kk)
                         kw = MIN( k, nzt+1 )
                         ijkfc_s(kk,jj,ii) = ijkfc_s(kk,jj,ii)                  &
                              + MERGE( 1, 0, BTEST( wall_flags_0(kw,jw,iw), 0 ) )
                      ENDDO
                   ENDDO
                ENDDO
!
!--             w-component 
                DO  i = iflo(ii), ifuo(ii)
                   iw = MAX( MIN( i, nx+1 ), -1 )
                   DO  j = jflo(jj), jfuo(jj)
                      jw = MAX( MIN( j, ny+1 ), -1 )
                      DO  k = kflw(kk), kfuw(kk)
                         kw = MIN( k, nzt+1 )
                         ijkfc_w(kk,jj,ii) = ijkfc_w(kk,jj,ii) + MERGE( 1, 0,   &
                              BTEST( wall_flags_0(kw,jw,iw), 3 ) )
                      ENDDO
                   ENDDO
                ENDDO

             ENDDO  ! kk        
          ENDDO  ! jj
       ENDDO  ! ii
!
!--    Spatial under-relaxation coefficients
       ALLOCATE( frax(icl:icr) )
       ALLOCATE( fray(jcs:jcn) )
       
       frax(icl:icr) = 1.0_wp
       fray(jcs:jcn) = 1.0_wp

!AH       IF ( nesting_mode /= 'vertical' )  THEN
!AH          DO  ii = icl, icr
!AH             IF ( ifuu(ii) < ( nx + 1 ) / 2 )  THEN   
!AH                xi = ( MAX( 0.0_wp, ( cg%coord_x(ii) -                         &
!AH                     lower_left_coord_x ) ) / anterp_relax_length_l )**4
!AH                frax(ii) = xi / ( 1.0_wp + xi )
!AH             ELSE
!AH                xi = ( MAX( 0.0_wp, ( lower_left_coord_x + ( nx + 1 ) * dx -   &
!AH                                      cg%coord_x(ii) ) ) /                     &
!AH                       anterp_relax_length_r )**4
!AH                frax(ii) = xi / ( 1.0_wp + xi )                
!AH             ENDIF
!AH          ENDDO
!AH
!AH          DO  jj = jcs, jcn
!AH             IF ( jfuv(jj) < ( ny + 1 ) / 2 )  THEN 
!AH                eta = ( MAX( 0.0_wp, ( cg%coord_y(jj) -                        &
!AH                     lower_left_coord_y ) ) / anterp_relax_length_s )**4
!AH                fray(jj) = eta / ( 1.0_wp + eta )
!AH             ELSE
!AH                eta = ( MAX( 0.0_wp, ( lower_left_coord_y + ( ny + 1 ) * dy -  &
!AH                                       cg%coord_y(jj)) ) /                     &
!AH                        anterp_relax_length_n )**4
!AH                fray(jj) = eta / ( 1.0_wp + eta )
!AH             ENDIF
!AH          ENDDO
!AH       ENDIF
      
       ALLOCATE( fraz(0:kcto) )
       fraz(0:kcto) = 1.0_wp
!AH       DO  kk = 0, kcto
!AH          zeta = ( ( zu(nzt) - cg%zu(kk) ) / anterp_relax_length_t )**4
!AH          fraz(kk) = zeta / ( 1.0_wp + zeta )
!AH       ENDDO

    END SUBROUTINE pmci_init_anterp_tophat

#endif
 END SUBROUTINE pmci_setup_child



 SUBROUTINE pmci_setup_coordinates

#if defined( __parallel )
    IMPLICIT NONE

    INTEGER(iwp) ::  i   !<
    INTEGER(iwp) ::  j   !<

!
!-- Create coordinate arrays.
    ALLOCATE( coord_x(-nbgp:nx+nbgp) )
    ALLOCATE( coord_y(-nbgp:ny+nbgp) )
     
    DO  i = -nbgp, nx + nbgp
       coord_x(i) = lower_left_coord_x + i * dx
    ENDDO

    DO  j = -nbgp, ny + nbgp
       coord_y(j) = lower_left_coord_y + j * dy
    ENDDO

#endif
 END SUBROUTINE pmci_setup_coordinates

!------------------------------------------------------------------------------!
! Description:
! ------------
!> In this subroutine the number of coupled arrays is determined. 
!------------------------------------------------------------------------------! 
  SUBROUTINE pmci_num_arrays  
               
#if defined( __parallel ) 
    USE pmc_general,                                                           &
        ONLY:  pmc_max_array

    IMPLICIT NONE
!
!-- The number of coupled arrays depends on the model settings. At least
!-- 5 arrays need to be coupled (u, v, w, e, diss).  Please note, actually
!-- e and diss (TKE and dissipation rate) are only required if RANS-RANS
!-- nesting is applied, but memory is allocated nevertheless. This is because
!-- the information whether they are needed or not is retrieved at a later
!-- point in time. In case e and diss are not needed, they are also not 
!-- exchanged between parent and child. 
    pmc_max_array = 5
!
!-- pt
    IF ( .NOT. neutral )  pmc_max_array = pmc_max_array + 1
    
    IF ( humidity )  THEN
!
!--    q
       pmc_max_array = pmc_max_array + 1
!
!--    qc, nc
       IF ( bulk_cloud_model  .AND.  microphysics_morrison )                   &
          pmc_max_array = pmc_max_array + 2
!
!--    qr, nr
       IF ( bulk_cloud_model  .AND.  microphysics_seifert )                    &
          pmc_max_array = pmc_max_array + 2
    ENDIF
!
!-- s
    IF ( passive_scalar )  pmc_max_array = pmc_max_array + 1
!
!-- nr_part, part_adr
    IF ( particle_advection )  pmc_max_array = pmc_max_array + 2
!
!-- Chemistry, depends on number of species
    IF ( air_chemistry  .AND.  nest_chemistry )                                &
       pmc_max_array = pmc_max_array + nspec
#endif
 END SUBROUTINE pmci_num_arrays



 SUBROUTINE pmci_set_array_pointer( name, child_id, nz_cl, n )

    IMPLICIT NONE

    INTEGER(iwp), INTENT(IN)          ::  child_id    !<
    INTEGER(iwp), INTENT(IN)          ::  nz_cl       !<
    INTEGER(iwp), INTENT(IN),OPTIONAL ::  n           !< index of chemical species

    CHARACTER(LEN=*), INTENT(IN) ::  name        !<

#if defined( __parallel )
    INTEGER(iwp) ::  ierr        !<

    REAL(wp), POINTER, DIMENSION(:,:)     ::  p_2d        !<
    REAL(wp), POINTER, DIMENSION(:,:)     ::  p_2d_sec    !<
    REAL(wp), POINTER, DIMENSION(:,:,:)   ::  p_3d        !<
    REAL(wp), POINTER, DIMENSION(:,:,:)   ::  p_3d_sec    !<
    INTEGER(idp), POINTER, DIMENSION(:,:) ::  i_2d        !<


    NULLIFY( p_3d )
    NULLIFY( p_2d )
    NULLIFY( i_2d )

!
!-- List of array names, which can be coupled.
!-- In case of 3D please change also the second array for the pointer version
    IF ( TRIM(name) == "u"          )  p_3d => u
    IF ( TRIM(name) == "v"          )  p_3d => v
    IF ( TRIM(name) == "w"          )  p_3d => w
    IF ( TRIM(name) == "e"          )  p_3d => e
    IF ( TRIM(name) == "pt"         )  p_3d => pt
    IF ( TRIM(name) == "q"          )  p_3d => q
    IF ( TRIM(name) == "qc"         )  p_3d => qc
    IF ( TRIM(name) == "qr"         )  p_3d => qr
    IF ( TRIM(name) == "nr"         )  p_3d => nr
    IF ( TRIM(name) == "nc"         )  p_3d => nc
    IF ( TRIM(name) == "s"          )  p_3d => s
    IF ( TRIM(name) == "diss"       )  p_3d => diss   
    IF ( TRIM(name) == "nr_part"    )  i_2d => nr_part
    IF ( TRIM(name) == "part_adr"   )  i_2d => part_adr
    IF ( INDEX( TRIM(name), "chem_" ) /= 0 )  p_3d => chem_species(n)%conc

!
!-- Next line is just an example for a 2D array (not active for coupling!)
!-- Please note, that z0 has to be declared as TARGET array in modules.f90
!    IF ( TRIM(name) == "z0" )    p_2d => z0

    IF ( TRIM(name) == "u"    )  p_3d_sec => u_2
    IF ( TRIM(name) == "v"    )  p_3d_sec => v_2
    IF ( TRIM(name) == "w"    )  p_3d_sec => w_2
    IF ( TRIM(name) == "e"    )  p_3d_sec => e_2
    IF ( TRIM(name) == "pt"   )  p_3d_sec => pt_2
    IF ( TRIM(name) == "q"    )  p_3d_sec => q_2
    IF ( TRIM(name) == "qc"   )  p_3d_sec => qc_2
    IF ( TRIM(name) == "qr"   )  p_3d_sec => qr_2
    IF ( TRIM(name) == "nr"   )  p_3d_sec => nr_2
    IF ( TRIM(name) == "nc"   )  p_3d_sec => nc_2
    IF ( TRIM(name) == "s"    )  p_3d_sec => s_2
    IF ( TRIM(name) == "diss" )  p_3d_sec => diss_2
    IF ( INDEX( TRIM(name), "chem_" ) /= 0 )  p_3d_sec => spec_conc_2(:,:,:,n)

    IF ( ASSOCIATED( p_3d ) )  THEN
       CALL pmc_s_set_dataarray( child_id, p_3d, nz_cl, nz,                    &
                                 array_2 = p_3d_sec )
    ELSEIF ( ASSOCIATED( p_2d ) )  THEN
       CALL pmc_s_set_dataarray( child_id, p_2d )
    ELSEIF ( ASSOCIATED( i_2d ) )  THEN
       CALL pmc_s_set_dataarray( child_id, i_2d )
    ELSE
!
!--    Give only one message for the root domain
       IF ( myid == 0  .AND.  cpl_id == 1 )  THEN

          message_string = 'pointer for array "' // TRIM( name ) //            &
                           '" can''t be associated'
          CALL message( 'pmci_set_array_pointer', 'PA0117', 3, 2, 0, 6, 0 )
       ELSE
!
!--       Avoid others to continue
          CALL MPI_BARRIER( comm2d, ierr )
       ENDIF

   ENDIF

#endif
END SUBROUTINE pmci_set_array_pointer



INTEGER FUNCTION get_number_of_childs ()

   IMPLICIT NONE

#if defined( __parallel )
   get_number_of_childs = SIZE( pmc_parent_for_child ) - 1
#else
   get_number_of_childs = 0
#endif

   RETURN

END FUNCTION get_number_of_childs


INTEGER FUNCTION get_childid (id_index)

   IMPLICIT NONE

   INTEGER,INTENT(IN)                 :: id_index

#if defined( __parallel )
   get_childid = pmc_parent_for_child(id_index)
#else
   get_childid = 0
#endif

   RETURN

END FUNCTION get_childid



SUBROUTINE  get_child_edges (m, lx_coord, lx_coord_b, rx_coord, rx_coord_b,    &
                               sy_coord, sy_coord_b, ny_coord, ny_coord_b,     &
                               uz_coord, uz_coord_b)
   IMPLICIT NONE
   INTEGER,INTENT(IN)             ::  m
   REAL(wp),INTENT(OUT)           ::  lx_coord, lx_coord_b
   REAL(wp),INTENT(OUT)           ::  rx_coord, rx_coord_b
   REAL(wp),INTENT(OUT)           ::  sy_coord, sy_coord_b
   REAL(wp),INTENT(OUT)           ::  ny_coord, ny_coord_b
   REAL(wp),INTENT(OUT)           ::  uz_coord, uz_coord_b

   lx_coord = childgrid(m)%lx_coord
   rx_coord = childgrid(m)%rx_coord
   sy_coord = childgrid(m)%sy_coord
   ny_coord = childgrid(m)%ny_coord
   uz_coord = childgrid(m)%uz_coord

   lx_coord_b = childgrid(m)%lx_coord_b
   rx_coord_b = childgrid(m)%rx_coord_b
   sy_coord_b = childgrid(m)%sy_coord_b
   ny_coord_b = childgrid(m)%ny_coord_b
   uz_coord_b = childgrid(m)%uz_coord_b

END SUBROUTINE get_child_edges



SUBROUTINE  get_child_gridspacing (m, dx,dy,dz)

   IMPLICIT NONE
   INTEGER,INTENT(IN)             ::  m
   REAL(wp),INTENT(OUT)           ::  dx,dy
   REAL(wp),INTENT(OUT),OPTIONAL  ::  dz

   dx = childgrid(m)%dx
   dy = childgrid(m)%dy
   IF(PRESENT(dz))   THEN
      dz = childgrid(m)%dz
   ENDIF

END SUBROUTINE get_child_gridspacing



SUBROUTINE pmci_create_child_arrays( name, is, ie, js, je, nzc, n  )

    IMPLICIT NONE

    CHARACTER(LEN=*), INTENT(IN) ::  name    !<

    INTEGER(iwp), INTENT(IN) ::  ie      !<
    INTEGER(iwp), INTENT(IN) ::  is      !<
    INTEGER(iwp), INTENT(IN) ::  je      !<
    INTEGER(iwp), INTENT(IN) ::  js      !<
    INTEGER(iwp), INTENT(IN) ::  nzc     !<  nzc is cg%nz, but note that cg%nz is not the original nz of parent, but the highest parent-grid level needed for nesting.

    INTEGER(iwp), INTENT(IN), OPTIONAL ::  n  !< number of chemical species

#if defined( __parallel )
    INTEGER(iwp) ::  ierr    !<

    REAL(wp), POINTER,DIMENSION(:,:)       ::  p_2d    !<
    REAL(wp), POINTER,DIMENSION(:,:,:)     ::  p_3d    !<
    INTEGER(idp), POINTER,DIMENSION(:,:)   ::  i_2d    !<


    NULLIFY( p_3d )
    NULLIFY( p_2d )
    NULLIFY( i_2d )

!
!-- List of array names, which can be coupled
    IF ( TRIM( name ) == "u" )  THEN
       IF ( .NOT. ALLOCATED( uc ) )  ALLOCATE( uc(0:nzc+1,js:je,is:ie) )
       p_3d => uc
    ELSEIF ( TRIM( name ) == "v" )  THEN
       IF ( .NOT. ALLOCATED( vc ) )  ALLOCATE( vc(0:nzc+1,js:je,is:ie) )
       p_3d => vc
    ELSEIF ( TRIM( name ) == "w" )  THEN
       IF ( .NOT. ALLOCATED( wc ) )  ALLOCATE( wc(0:nzc+1,js:je,is:ie) )
       p_3d => wc
    ELSEIF ( TRIM( name ) == "e" )  THEN
       IF ( .NOT. ALLOCATED( ec ) )  ALLOCATE( ec(0:nzc+1,js:je,is:ie) )
       p_3d => ec
    ELSEIF ( TRIM( name ) == "diss" )  THEN
       IF ( .NOT. ALLOCATED( dissc ) )  ALLOCATE( dissc(0:nzc+1,js:je,is:ie) )
       p_3d => dissc
    ELSEIF ( TRIM( name ) == "pt")  THEN
       IF ( .NOT. ALLOCATED( ptc ) )  ALLOCATE( ptc(0:nzc+1,js:je,is:ie) )
       p_3d => ptc
    ELSEIF ( TRIM( name ) == "q")  THEN
       IF ( .NOT. ALLOCATED( q_c ) )  ALLOCATE( q_c(0:nzc+1,js:je,is:ie) )
       p_3d => q_c
    ELSEIF ( TRIM( name ) == "qc")  THEN
       IF ( .NOT. ALLOCATED( qcc ) )  ALLOCATE( qcc(0:nzc+1,js:je,is:ie) )
       p_3d => qcc
    ELSEIF ( TRIM( name ) == "qr")  THEN
       IF ( .NOT. ALLOCATED( qrc ) )  ALLOCATE( qrc(0:nzc+1,js:je,is:ie) )
       p_3d => qrc
    ELSEIF ( TRIM( name ) == "nr")  THEN
       IF ( .NOT. ALLOCATED( nrc ) )  ALLOCATE( nrc(0:nzc+1,js:je,is:ie) )
       p_3d => nrc
    ELSEIF ( TRIM( name ) == "nc")  THEN
       IF ( .NOT. ALLOCATED( ncc ) )  ALLOCATE( ncc(0:nzc+1,js:je,is:ie) )
       p_3d => ncc
    ELSEIF ( TRIM( name ) == "s")  THEN
       IF ( .NOT. ALLOCATED( sc ) )  ALLOCATE( sc(0:nzc+1,js:je,is:ie) )
       p_3d => sc
    ELSEIF ( TRIM( name ) == "nr_part") THEN
       IF ( .NOT. ALLOCATED( nr_partc ) )  ALLOCATE( nr_partc(js:je,is:ie) )
       i_2d => nr_partc
    ELSEIF ( TRIM( name ) == "part_adr") THEN
       IF ( .NOT. ALLOCATED( part_adrc ) )  ALLOCATE( part_adrc(js:je,is:ie) )
       i_2d => part_adrc
    ELSEIF ( TRIM( name(1:5) ) == "chem_" )  THEN
       IF ( .NOT. ALLOCATED( chem_spec_c ) )                                   &
          ALLOCATE( chem_spec_c(0:nzc+1,js:je,is:ie,1:nspec) )
       p_3d => chem_spec_c(:,:,:,n)
    !ELSEIF (trim(name) == "z0") then
       !IF (.not.allocated(z0c))  allocate(z0c(js:je, is:ie))
       !p_2d => z0c
    ENDIF

    IF ( ASSOCIATED( p_3d ) )  THEN
       CALL pmc_c_set_dataarray( p_3d )
    ELSEIF ( ASSOCIATED( p_2d ) )  THEN
       CALL pmc_c_set_dataarray( p_2d )
    ELSEIF ( ASSOCIATED( i_2d ) )  THEN
       CALL pmc_c_set_dataarray( i_2d )
    ELSE
!
!--    Give only one message for the first child domain
       IF ( myid == 0  .AND.  cpl_id == 2 )  THEN

          message_string = 'pointer for array "' // TRIM( name ) //            &
                           '" can''t be associated'
          CALL message( 'pmci_create_child_arrays', 'PA0170', 3, 2, 0, 6, 0 )
       ELSE
!
!--       Prevent others from continuing
          CALL MPI_BARRIER( comm2d, ierr )
       ENDIF
    ENDIF

#endif
 END SUBROUTINE pmci_create_child_arrays



 SUBROUTINE pmci_parent_initialize

!
!-- Send data for the children in order to let them create initial 
!-- conditions by interpolating the parent-domain fields.
#if defined( __parallel )
    IMPLICIT NONE

    INTEGER(iwp) ::  child_id    !<
    INTEGER(iwp) ::  m           !<

    REAL(wp) ::  waittime        !<


    DO  m = 1, SIZE( pmc_parent_for_child ) - 1
       child_id = pmc_parent_for_child(m)
       CALL pmc_s_fillbuffer( child_id, waittime=waittime )
    ENDDO

#endif
 END SUBROUTINE pmci_parent_initialize



 SUBROUTINE pmci_child_initialize

!
!-- Create initial conditions for the current child domain by interpolating 
!-- the parent-domain fields.
#if defined( __parallel )
    IMPLICIT NONE

    INTEGER(iwp) ::  i          !<
    INTEGER(iwp) ::  icl        !<
    INTEGER(iwp) ::  icla       !<
    INTEGER(iwp) ::  iclw       !<
    INTEGER(iwp) ::  icr        !<
    INTEGER(iwp) ::  icra       !<
    INTEGER(iwp) ::  icrw       !<
    INTEGER(iwp) ::  j          !<
    INTEGER(iwp) ::  jcn        !<
    INTEGER(iwp) ::  jcna       !<
    INTEGER(iwp) ::  jcnw       !<
    INTEGER(iwp) ::  jcs        !<
    INTEGER(iwp) ::  jcsa       !<
    INTEGER(iwp) ::  jcsw       !<
    INTEGER(iwp) ::  k          !< 
    INTEGER(iwp) ::  n          !< running index for chemical species

    REAL(wp) ::  waittime       !<

!
!-- Root model is never anyone's child
    IF ( cpl_id > 1 )  THEN
!
!--    Child domain boundaries in the parent index space
       icl  = coarse_bound(1)
       icr  = coarse_bound(2)
       jcs  = coarse_bound(3)
       jcn  = coarse_bound(4)
       icla = coarse_bound_aux(1)
       icra = coarse_bound_aux(2)
       jcsa = coarse_bound_aux(3)
       jcna = coarse_bound_aux(4)
       iclw = coarse_bound_w(1)
       icrw = coarse_bound_w(2)
       jcsw = coarse_bound_w(3)
       jcnw = coarse_bound_w(4)

!
!--    Get data from the parent
       CALL pmc_c_getbuffer( waittime = waittime )
!
!--    The interpolation.
       CALL pmci_interp_1sto_all ( u,  uc,  icu, jco, kco, r1xu, r2xu, r1yo,   &
                                   r2yo, r1zo, r2zo, kcto, iflu, ifuu,         &
                                   jflo, jfuo, kflo, kfuo, ijkfc_u, 'u' )
       CALL pmci_interp_1sto_all ( v,  vc,  ico, jcv, kco, r1xo, r2xo, r1yv,   &
                                   r2yv, r1zo, r2zo, kcto, iflo, ifuo,         &
                                   jflv, jfuv, kflo, kfuo, ijkfc_v, 'v' )
       CALL pmci_interp_1sto_all ( w,  wc,  ico, jco, kcw, r1xo, r2xo, r1yo,   &
                                   r2yo, r1zw, r2zw, kctw, iflo, ifuo,         &
                                   jflo, jfuo, kflw, kfuw, ijkfc_w, 'w' )

       IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.          &
            (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.           &
               .NOT. constant_diffusion ) )  THEN
          CALL pmci_interp_1sto_all ( e,  ec,  ico, jco, kco, r1xo, r2xo, r1yo, &
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,       &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 'e' )
       ENDIF

       IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
          CALL pmci_interp_1sto_all ( diss,  dissc,  ico, jco, kco, r1xo, r2xo,&
                                      r1yo, r2yo, r1zo, r2zo, kcto, iflo, ifuo,&
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( .NOT. neutral )  THEN
          CALL pmci_interp_1sto_all ( pt, ptc, ico, jco, kco, r1xo, r2xo,      &
                                      r1yo, r2yo, r1zo, r2zo, kcto, iflo, ifuo,&
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( humidity )  THEN

          CALL pmci_interp_1sto_all ( q, q_c, ico, jco, kco, r1xo, r2xo, r1yo, &
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

          IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN
             CALL pmci_interp_1sto_all ( qc, qcc, ico, jco, kco, r1xo, r2xo,   &
                                         r1yo, r2yo, r1zo, r2zo, kcto,         &
                                         iflo, ifuo, jflo, jfuo, kflo, kfuo,   &
                                         ijkfc_s, 's' ) 
             CALL pmci_interp_1sto_all ( nc, ncc, ico, jco, kco, r1xo, r2xo,   &
                                         r1yo, r2yo, r1zo, r2zo, kcto,         &
                                         iflo, ifuo, jflo, jfuo, kflo, kfuo,   &
                                         ijkfc_s, 's' )   
          ENDIF

          IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN
             CALL pmci_interp_1sto_all ( qr, qrc, ico, jco, kco, r1xo, r2xo,   &
                                         r1yo, r2yo, r1zo, r2zo, kcto,         &
                                         iflo, ifuo, jflo, jfuo, kflo, kfuo,   &
                                         ijkfc_s, 's' )
             CALL pmci_interp_1sto_all ( nr, nrc, ico, jco, kco, r1xo, r2xo,   &
                                         r1yo, r2yo, r1zo, r2zo, kcto,         &
                                         iflo, ifuo, jflo, jfuo, kflo, kfuo,   &
                                         ijkfc_s, 's' )
          ENDIF

       ENDIF

       IF ( passive_scalar )  THEN
          CALL pmci_interp_1sto_all ( s, sc, ico, jco, kco, r1xo, r2xo, r1yo,  &
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( air_chemistry  .AND.  nest_chemistry )  THEN
          DO  n = 1, nspec
             CALL pmci_interp_1sto_all ( chem_species(n)%conc,                 &
                                         chem_spec_c(:,:,:,n),                 &
                                         ico, jco, kco, r1xo, r2xo, r1yo,      &
                                         r2yo, r1zo, r2zo, kcto, iflo, ifuo,   &
                                         jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
          ENDDO
       ENDIF

       IF ( topography /= 'flat' )  THEN
!
!--       Inside buildings set velocities and TKE back to zero.
!--       Other scalars (pt, q, s, km, kh, p, sa, ...) are ignored at present,
!--       maybe revise later.
          DO   i = nxlg, nxrg
             DO   j = nysg, nyng
                DO  k = nzb, nzt
                   u(k,j,i)   = MERGE( u(k,j,i), 0.0_wp,                       &
                                       BTEST( wall_flags_0(k,j,i), 1 ) )
                   v(k,j,i)   = MERGE( v(k,j,i), 0.0_wp,                       &
                                       BTEST( wall_flags_0(k,j,i), 2 ) )
                   w(k,j,i)   = MERGE( w(k,j,i), 0.0_wp,                       &
                                       BTEST( wall_flags_0(k,j,i), 3 ) )
!                    e(k,j,i)   = MERGE( e(k,j,i), 0.0_wp,                       &
!                                        BTEST( wall_flags_0(k,j,i), 0 ) )
                   u_p(k,j,i) = MERGE( u_p(k,j,i), 0.0_wp,                     &
                                       BTEST( wall_flags_0(k,j,i), 1 ) )
                   v_p(k,j,i) = MERGE( v_p(k,j,i), 0.0_wp,                     &
                                       BTEST( wall_flags_0(k,j,i), 2 ) )
                   w_p(k,j,i) = MERGE( w_p(k,j,i), 0.0_wp,                     &
                                       BTEST( wall_flags_0(k,j,i), 3 ) )
!                    e_p(k,j,i) = MERGE( e_p(k,j,i), 0.0_wp,                     &
!                                        BTEST( wall_flags_0(k,j,i), 0 ) )
                ENDDO
             ENDDO
          ENDDO
       ENDIF
    ENDIF


 CONTAINS


    SUBROUTINE pmci_interp_1sto_all( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y,    &
                                     r1z, r2z, kct, ifl, ifu, jfl, jfu,        &
                                     kfl, kfu, ijkfc, var )
!
!--    Interpolation of the internal values for the child-domain initialization
!--    This subroutine is based on trilinear interpolation.
!--    Coding based on interp_tril_lr/sn/t
       IMPLICIT NONE

       CHARACTER(LEN=1), INTENT(IN) :: var  !<

       INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)         ::  ic    !<
       INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)         ::  jc    !<
!AH       INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)           ::  kc    !<
       INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN)        ::  kc    !<

       REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg), INTENT(INOUT) :: f !<
       REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: fc       !<
       REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  :: r1x   !<
       REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  :: r2x   !<
       REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  :: r1y   !<
       REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  :: r2y   !<
       REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) :: r1z   !<
       REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) :: r2z   !<

       INTEGER(iwp) :: kct
!AH       INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!AH       INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!AH       INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH       INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
       INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
       INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
       INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
       INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
       INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN) ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
       INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN) ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH
!       INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
       INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

       INTEGER(iwp) ::  i        !<
       INTEGER(iwp) ::  j        !<
       INTEGER(iwp) ::  k        !<
       INTEGER(iwp) ::  l        !<
       INTEGER(iwp) ::  lb       !<
       INTEGER(iwp) ::  le       !<
       INTEGER(iwp) ::  m        !<
       INTEGER(iwp) ::  mb       !<
       INTEGER(iwp) ::  me       !<
       INTEGER(iwp) ::  n        !<
       INTEGER(iwp) ::  var_flag !<


       lb = icl
       le = icr
       mb = jcs
       me = jcn

       IF ( nesting_mode /= 'vertical' )  THEN
          IF ( bc_dirichlet_l )  THEN
             lb = icl + 1
!
!--          For u, nxl is a ghost node, but not for the other variables
             IF ( var == 'u' )  THEN
                lb = icl + 2
             ENDIF
          ENDIF
          IF ( bc_dirichlet_s )  THEN
             mb = jcs + 1
!
!--          For v, nys is a ghost node, but not for the other variables
             IF ( var == 'v' )  THEN
                mb = jcs + 2
             ENDIF
          ENDIF
          IF ( bc_dirichlet_r )  THEN
             le = icr - 1
          ENDIF
          IF ( bc_dirichlet_n )  THEN
             me = jcn - 1
          ENDIF
       ENDIF
!
!--    Is masking needed here, think about it.
       IF ( var == 'u' )  THEN 
          var_flag = 1 
       ELSEIF ( var == 'v' )  THEN
          var_flag = 2 
       ELSEIF ( var == 'w' )  THEN
          var_flag = 3
       ELSE
          var_flag = 0
       ENDIF       

       f(:,:,:) = 0.0_wp

       IF  ( var == 'u' )  THEN

          DO  l = lb, le
             DO  m = mb, me
                DO n = 0, kct + 1 

                   DO  i = ifl(l), ifl(l+1)-1
                      DO  j = jfl(m), jfu(m)
                         DO  k = kfl(n), MIN( kfu(n), nzt+1 )
                            f(k,j,i) = fc(n,m,l)
                         ENDDO
                      ENDDO
                   ENDDO
                   
                ENDDO
             ENDDO
          ENDDO

       ELSE IF  ( var == 'v' )  THEN

          DO  l = lb, le
             DO  m = mb, me
                DO n = 0, kct + 1  
               
                   DO i = ifl(l), ifu(l)
                      DO  j = jfl(m), jfl(m+1)-1
                         DO  k = kfl(n), MIN( kfu(n), nzt+1 )
                            f(k,j,i) = fc(n,m,l) 
                         ENDDO
                      ENDDO
                   ENDDO

                ENDDO
             ENDDO
          ENDDO

       ELSE IF  ( var == 'w' )  THEN

          DO  l = lb, le
             DO  m = mb, me
                DO n = 1, kct + 1  

                   DO i = ifl(l), ifu(l)
                      DO  j = jfl(m), jfu(m)
                         f(nzb,j,i) = 0.0_wp   ! Because the n-loop starts from n=1 instead of 0
                         DO  k = kfu(n-1)+1, kfu(n) 
                            f(k,j,i) = fc(n,m,l)
                         ENDDO
                      ENDDO
                   ENDDO
                   
                ENDDO
             ENDDO
          ENDDO

       ELSE   ! scalars

          DO  l = lb, le
             DO  m = mb, me
                DO n = 0, kct + 1
                   
                   DO i = ifl(l), ifu(l)
                      DO  j = jfl(m), jfu(m)
                         DO  k = kfl(n), MIN( kfu(n), nzt+1 )
                            f(k,j,i) = fc(n,m,l)
                         ENDDO
                      ENDDO
                   ENDDO
                   
                ENDDO
             ENDDO
          ENDDO

       ENDIF  ! var

    END SUBROUTINE pmci_interp_1sto_all

#endif
 END SUBROUTINE pmci_child_initialize



 SUBROUTINE pmci_check_setting_mismatches
!
!-- Check for mismatches between settings of master and child variables
!-- (e.g., all children have to follow the end_time settings of the root model).
!-- The root model overwrites variables in the other models, so these variables
!-- only need to be set once in file PARIN.

#if defined( __parallel )

    USE control_parameters,                                                    &
        ONLY:  dt_restart, end_time, message_string, restart_time, time_restart

    IMPLICIT NONE

    INTEGER ::  ierr

    REAL(wp) ::  dt_restart_root
    REAL(wp) ::  end_time_root
    REAL(wp) ::  restart_time_root
    REAL(wp) ::  time_restart_root

!
!-- Check the time to be simulated.
!-- Here, and in the following, the root process communicates the respective
!-- variable to all others, and its value will then be compared with the local
!-- values.
    IF ( pmc_is_rootmodel() )  end_time_root = end_time
    CALL MPI_BCAST( end_time_root, 1, MPI_REAL, 0, comm_world_nesting, ierr )

    IF ( .NOT. pmc_is_rootmodel() )  THEN
       IF ( end_time /= end_time_root )  THEN
          WRITE( message_string, * )  'mismatch between root model and ',      &
               'child settings:& end_time(root) = ', end_time_root,            &
               '& end_time(child) = ', end_time, '& child value is set',       &
               ' to root value'
          CALL message( 'pmci_check_setting_mismatches', 'PA0419', 0, 1, 0, 6, &
                        0 )
          end_time = end_time_root
       ENDIF
    ENDIF
!
!-- Same for restart time
    IF ( pmc_is_rootmodel() )  restart_time_root = restart_time
    CALL MPI_BCAST( restart_time_root, 1, MPI_REAL, 0, comm_world_nesting, ierr )

    IF ( .NOT. pmc_is_rootmodel() )  THEN
       IF ( restart_time /= restart_time_root )  THEN
          WRITE( message_string, * )  'mismatch between root model and ',      &
               'child settings: & restart_time(root) = ', restart_time_root,   &
               '& restart_time(child) = ', restart_time, '& child ',           &
               'value is set to root value'
          CALL message( 'pmci_check_setting_mismatches', 'PA0419', 0, 1, 0, 6, &
                        0 )
          restart_time = restart_time_root
       ENDIF
    ENDIF
!
!-- Same for dt_restart
    IF ( pmc_is_rootmodel() )  dt_restart_root = dt_restart
    CALL MPI_BCAST( dt_restart_root, 1, MPI_REAL, 0, comm_world_nesting, ierr )

    IF ( .NOT. pmc_is_rootmodel() )  THEN
       IF ( dt_restart /= dt_restart_root )  THEN
          WRITE( message_string, * )  'mismatch between root model and ',      &
               'child settings: & dt_restart(root) = ', dt_restart_root,       &
               '& dt_restart(child) = ', dt_restart, '& child ',               &
               'value is set to root value'
          CALL message( 'pmci_check_setting_mismatches', 'PA0419', 0, 1, 0, 6, &
                        0 )
          dt_restart = dt_restart_root
       ENDIF
    ENDIF
!
!-- Same for time_restart
    IF ( pmc_is_rootmodel() )  time_restart_root = time_restart
    CALL MPI_BCAST( time_restart_root, 1, MPI_REAL, 0, comm_world_nesting, ierr )

    IF ( .NOT. pmc_is_rootmodel() )  THEN
       IF ( time_restart /= time_restart_root )  THEN
          WRITE( message_string, * )  'mismatch between root model and ',      &
               'child settings: & time_restart(root) = ', time_restart_root,   &
               '& time_restart(child) = ', time_restart, '& child ',           &
               'value is set to root value'
          CALL message( 'pmci_check_setting_mismatches', 'PA0419', 0, 1, 0, 6, &
                        0 )
          time_restart = time_restart_root
       ENDIF
    ENDIF

#endif

 END SUBROUTINE pmci_check_setting_mismatches



 SUBROUTINE pmci_synchronize

#if defined( __parallel )
!
!-- Unify the time steps for each model and synchronize using 
!-- MPI_ALLREDUCE with the MPI_MIN operator over all processes using 
!-- the global communicator MPI_COMM_WORLD.
   
   IMPLICIT NONE

   INTEGER(iwp)           :: ierr  !<
   REAL(wp), DIMENSION(1) :: dtl   !<
   REAL(wp), DIMENSION(1) :: dtg   !<

   
   dtl(1) = dt_3d 
   CALL MPI_ALLREDUCE( dtl, dtg, 1, MPI_REAL, MPI_MIN, MPI_COMM_WORLD, ierr )
   dt_3d  = dtg(1)

#endif
 END SUBROUTINE pmci_synchronize
                


 SUBROUTINE pmci_set_swaplevel( swaplevel )

!
!-- After each Runge-Kutta sub-timestep, alternately set buffer one or buffer
!-- two active

    IMPLICIT NONE

    INTEGER(iwp), INTENT(IN) ::  swaplevel  !< swaplevel (1 or 2) of PALM's
                                            !< timestep

    INTEGER(iwp)            ::  child_id    !<
    INTEGER(iwp)            ::  m           !<

#if defined( __parallel )
    DO  m = 1, SIZE( pmc_parent_for_child )-1
       child_id = pmc_parent_for_child(m)
       CALL pmc_s_set_active_data_array( child_id, swaplevel )
    ENDDO
#endif
 END SUBROUTINE pmci_set_swaplevel



 SUBROUTINE pmci_datatrans( local_nesting_mode )
!
!-- This subroutine controls the nesting according to the nestpar 
!-- parameter nesting_mode (two-way (default) or one-way) and the 
!-- order of anterpolations according to the nestpar parameter 
!-- nesting_datatransfer_mode (cascade, overlap or mixed (default)).
!-- Although nesting_mode is a variable of this model, pass it as 
!-- an argument to allow for example to force one-way initialization 
!-- phase.

    IMPLICIT NONE

    CHARACTER(LEN=*), INTENT(IN) ::  local_nesting_mode


    IF ( TRIM( local_nesting_mode ) == 'one-way' )  THEN

       CALL pmci_child_datatrans( parent_to_child )
       CALL pmci_parent_datatrans( parent_to_child )

    ELSE

       IF( nesting_datatransfer_mode == 'cascade' )  THEN

          CALL pmci_child_datatrans( parent_to_child )
          CALL pmci_parent_datatrans( parent_to_child )

          CALL pmci_parent_datatrans( child_to_parent )
          CALL pmci_child_datatrans( child_to_parent )

       ELSEIF( nesting_datatransfer_mode == 'overlap')  THEN

          CALL pmci_parent_datatrans( parent_to_child )
          CALL pmci_child_datatrans( parent_to_child )

          CALL pmci_child_datatrans( child_to_parent )
          CALL pmci_parent_datatrans( child_to_parent )

       ELSEIF( TRIM( nesting_datatransfer_mode ) == 'mixed' )  THEN

          CALL pmci_parent_datatrans( parent_to_child )
          CALL pmci_child_datatrans( parent_to_child )

          CALL pmci_parent_datatrans( child_to_parent )
          CALL pmci_child_datatrans( child_to_parent )

       ENDIF

    ENDIF

 END SUBROUTINE pmci_datatrans



 SUBROUTINE pmci_parent_datatrans( direction )

    IMPLICIT NONE

    INTEGER(iwp), INTENT(IN) ::  direction   !<

#if defined( __parallel )
    INTEGER(iwp) ::  child_id    !<
    INTEGER(iwp) ::  i           !<
    INTEGER(iwp) ::  j           !<
    INTEGER(iwp) ::  k           !<
    INTEGER(iwp) ::  m           !<


    DO  m = 1, SIZE( pmc_parent_for_child ) - 1
       child_id = pmc_parent_for_child(m)
       IF ( direction == parent_to_child )  THEN
          CALL cpu_log( log_point_s(71), 'pmc parent send', 'start' )
          CALL pmc_s_fillbuffer( child_id )
          CALL cpu_log( log_point_s(71), 'pmc parent send', 'stop' )
       ELSE
!
!--       Communication from child to parent
          CALL cpu_log( log_point_s(72), 'pmc parent recv', 'start' )
          child_id = pmc_parent_for_child(m)
          CALL pmc_s_getdata_from_buffer( child_id )
          CALL cpu_log( log_point_s(72), 'pmc parent recv', 'stop' )
!
!--       The anterpolated data is now available in u etc
          IF ( topography /= 'flat' )  THEN
!
!--          Inside buildings/topography reset velocities back to zero.
!--          Scalars (pt, q, s, km, kh, p, sa, ...) are ignored at
!--          present, maybe revise later.
!--          Resetting of e is removed as unnecessary since e is not 
!--          anterpolated, and as incorrect since it overran the default 
!--          Neumann condition (bc_e_b). 
             DO   i = nxlg, nxrg
                DO   j = nysg, nyng
                   DO  k = nzb, nzt+1
                      u(k,j,i)  = MERGE( u(k,j,i), 0.0_wp,                     &
                                         BTEST( wall_flags_0(k,j,i), 1 ) )
                      v(k,j,i)  = MERGE( v(k,j,i), 0.0_wp,                     &
                                         BTEST( wall_flags_0(k,j,i), 2 ) )
                      w(k,j,i)  = MERGE( w(k,j,i), 0.0_wp,                     &
                                         BTEST( wall_flags_0(k,j,i), 3 ) )
!
!--                 TO_DO: zero setting of temperature within topography creates
!--                       wrong results
!                   pt(nzb:nzb_s_inner(j,i),j,i) = 0.0_wp
!                   IF ( humidity  .OR.  passive_scalar )  THEN
!                      q(nzb:nzb_s_inner(j,i),j,i) = 0.0_wp
!                   ENDIF
                   ENDDO
                ENDDO
             ENDDO
          ENDIF
       ENDIF
    ENDDO

#endif
 END SUBROUTINE pmci_parent_datatrans



 SUBROUTINE pmci_child_datatrans( direction )

    IMPLICIT NONE

    INTEGER(iwp), INTENT(IN) ::  direction  !< Transfer direction: parent_to_child or child_to_parent

#if defined( __parallel )
    INTEGER(iwp) ::  icl         !< Parent-grid array index bound, left
    INTEGER(iwp) ::  icr         !< Parent-grid array index bound, right
    INTEGER(iwp) ::  jcn         !< Parent-grid array index bound, north
    INTEGER(iwp) ::  jcs         !< Parent-grid array index bound, south
    INTEGER(iwp) ::  icla        !< Auxiliary-array (index-mapping etc) index bound, left
    INTEGER(iwp) ::  icra        !< Auxiliary-array (index-mapping etc) index bound, right
    INTEGER(iwp) ::  jcna        !< Auxiliary-array (index-mapping etc) index bound, north
    INTEGER(iwp) ::  jcsa        !< Auxiliary-array (index-mapping etc) index bound, south
    INTEGER(iwp) ::  iclw        !< Parent-grid work array index bound, left
    INTEGER(iwp) ::  icrw        !< Parent-grid work array index bound, right
    INTEGER(iwp) ::  jcnw        !< Parent-grid work array index bound, north
    INTEGER(iwp) ::  jcsw        !< Parent-grid work array index bound, south

    REAL(wp), DIMENSION(1) ::  dtl         !< Time step size


    dtl = dt_3d
    IF ( cpl_id > 1 )  THEN
!
!--    Child domain boundaries in the parent indice space.
       icl  = coarse_bound(1)
       icr  = coarse_bound(2)
       jcs  = coarse_bound(3)
       jcn  = coarse_bound(4)
       icla = coarse_bound_aux(1)
       icra = coarse_bound_aux(2)
       jcsa = coarse_bound_aux(3)
       jcna = coarse_bound_aux(4)
       iclw = coarse_bound_w(1)
       icrw = coarse_bound_w(2)
       jcsw = coarse_bound_w(3)
       jcnw = coarse_bound_w(4)

       IF ( direction == parent_to_child )  THEN
    
          CALL cpu_log( log_point_s(73), 'pmc child recv', 'start' )
          CALL pmc_c_getbuffer( )
          CALL cpu_log( log_point_s(73), 'pmc child recv', 'stop' )

          CALL cpu_log( log_point_s(75), 'pmc interpolation', 'start' )
          CALL pmci_interpolation
          CALL cpu_log( log_point_s(75), 'pmc interpolation', 'stop' )
     
       ELSE
!
!--       direction == child_to_parent
          CALL cpu_log( log_point_s(76), 'pmc anterpolation', 'start' )
          CALL pmci_anterpolation
          CALL cpu_log( log_point_s(76), 'pmc anterpolation', 'stop' )

          CALL cpu_log( log_point_s(74), 'pmc child send', 'start' )
          CALL pmc_c_putbuffer( )
          CALL cpu_log( log_point_s(74), 'pmc child send', 'stop' )

       ENDIF
    ENDIF

  CONTAINS

   
    SUBROUTINE pmci_interpolation

!
!--    A wrapper routine for all interpolation actions
      
       IMPLICIT NONE

       INTEGER(iwp) ::  ibgp       !< index running over the nbgp boundary ghost points in i-direction
       INTEGER(iwp) ::  jbgp       !< index running over the nbgp boundary ghost points in j-direction
       INTEGER(iwp) ::  n          !< running index for number of chemical species
      
!
!--    In case of vertical nesting no interpolation is needed for the
!--    horizontal boundaries
       IF ( nesting_mode /= 'vertical' )  THEN
!
!--       Left border pe:
          IF ( bc_dirichlet_l )  THEN

             CALL pmci_interp_1sto_lr( u,  uc,  icu, jco, kco, r1xu, r2xu,      &
                                       r1yo, r2yo, r1zo, r2zo,                  &
                                       logc_u_l, logc_ratio_u_l,                &
                                       logc_kbounds_u_l, nzt_topo_nestbc_l,     &
                                       kcto, iflu, ifuu, jflo, jfuo, kflo,      &
                                       kfuo, ijkfc_u, 'l', 'u' )

             CALL pmci_interp_1sto_lr( v,  vc,  ico, jcv, kco, r1xo, r2xo,      &
                                       r1yv, r2yv, r1zo, r2zo,                  &
                                       logc_v_l, logc_ratio_v_l,                &
                                       logc_kbounds_v_l, nzt_topo_nestbc_l,     &
                                       kcto, iflo, ifuo, jflv, jfuv, kflo,      &
                                       kfuo, ijkfc_v, 'l', 'v' )

             CALL pmci_interp_1sto_lr( w,  wc,  ico, jco, kcw, r1xo, r2xo,      &
                                       r1yo, r2yo, r1zw, r2zw,                  &
                                       logc_w_l, logc_ratio_w_l,                &
                                       logc_kbounds_w_l, nzt_topo_nestbc_l,     &
                                       kctw, iflo, ifuo, jflo, jfuo, kflw,      &
                                       kfuw, ijkfc_w, 'l', 'w' )

             IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.   &
                  (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.    &
                     .NOT. constant_diffusion ) )  THEN
!                CALL pmci_interp_1sto_lr( e,  ec,  ico, jco, kco, r1xo, r2xo,  &
!                                          r1yo, r2yo, r1zo, r2zo,              &
!                                          logc_w_l, logc_ratio_w_l,            &
!                                          logc_kbounds_w_l, nzt_topo_nestbc_l, &
!                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
!                                          kfuo, ijkfc_s, 'l', 'e' )
!
!--             Interpolation of e is replaced by the Neumann condition. 
                DO ibgp = -nbgp, -1
                   e(nzb:nzt,nys:nyn,ibgp) = e(nzb:nzt,nys:nyn,0)
                ENDDO

             ENDIF

             IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
                CALL pmci_interp_1sto_lr( diss,  dissc,  ico, jco, kco, r1xo,  &
                                          r2xo, r1yo, r2yo, r1zo, r2zo,        &
                                          logc_w_l, logc_ratio_w_l,            &
                                          logc_kbounds_w_l, nzt_topo_nestbc_l, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'l', 's' )
             ENDIF

             IF ( .NOT. neutral )  THEN
                CALL pmci_interp_1sto_lr( pt, ptc, ico, jco, kco, r1xo, r2xo,  &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_l, logc_ratio_w_l,            &
                                          logc_kbounds_w_l, nzt_topo_nestbc_l, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'l', 's' )
             ENDIF

             IF ( humidity )  THEN

                CALL pmci_interp_1sto_lr( q, q_c, ico, jco, kco, r1xo, r2xo,   &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_l, logc_ratio_w_l,            &
                                          logc_kbounds_w_l, nzt_topo_nestbc_l, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'l', 's' )

                IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN
                   CALL pmci_interp_1sto_lr( qc, qcc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_l, logc_ratio_w_l,         &
                                             logc_kbounds_w_l,                 &
                                             nzt_topo_nestbc_l,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'l', 's' )  

                   CALL pmci_interp_1sto_lr( nc, ncc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_l, logc_ratio_w_l,         &
                                             logc_kbounds_w_l,                 &
                                             nzt_topo_nestbc_l,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'l', 's' )          
                ENDIF

                IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN
                   CALL pmci_interp_1sto_lr( qr, qrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_l, logc_ratio_w_l,         &
                                             logc_kbounds_w_l,                 &
                                             nzt_topo_nestbc_l,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'l', 's' ) 

                   CALL pmci_interp_1sto_lr( nr, nrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_l, logc_ratio_w_l,         &
                                             logc_kbounds_w_l,                 &
                                             nzt_topo_nestbc_l,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'l', 's' )             
                ENDIF

             ENDIF

             IF ( passive_scalar )  THEN
                CALL pmci_interp_1sto_lr( s, sc, ico, jco, kco, r1xo, r2xo,    &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_l, logc_ratio_w_l,            &
                                          logc_kbounds_w_l,                    & 
                                          nzt_topo_nestbc_l,                   &
                                          kcto, iflo, ifuo, jflo, jfuo,        &
                                          kflo, kfuo, ijkfc_s, 'l', 's' )
             ENDIF

             IF ( air_chemistry  .AND.  nest_chemistry )  THEN
                DO  n = 1, nspec
                   CALL pmci_interp_1sto_lr( chem_species(n)%conc,             &
                                             chem_spec_c(:,:,:,n),             &
                                             ico, jco, kco, r1xo, r2xo,        &
                                             r1yo, r2yo, r1zo, r2zo,           &
                                             logc_w_l, logc_ratio_w_l,         &
                                             logc_kbounds_w_l,                 & 
                                             nzt_topo_nestbc_l,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'l', 's' )
                ENDDO
             ENDIF

          ENDIF
!
!--       Right border pe
          IF ( bc_dirichlet_r )  THEN
             
             CALL pmci_interp_1sto_lr( u,  uc,  icu, jco, kco, r1xu, r2xu,     &
                                       r1yo, r2yo, r1zo, r2zo,                 &
                                       logc_u_r, logc_ratio_u_r,               &
                                       logc_kbounds_u_r, nzt_topo_nestbc_r,    &
                                       kcto, iflu, ifuu, jflo, jfuo, kflo,     &
                                       kfuo, ijkfc_u, 'r', 'u' )

             CALL pmci_interp_1sto_lr( v,  vc,  ico, jcv, kco, r1xo, r2xo,     &
                                       r1yv, r2yv, r1zo, r2zo,                 &
                                       logc_v_r, logc_ratio_v_r,               &
                                       logc_kbounds_v_r, nzt_topo_nestbc_r,    &
                                       kcto, iflo, ifuo, jflv, jfuv, kflo,     &
                                       kfuo, ijkfc_v, 'r', 'v' )

             CALL pmci_interp_1sto_lr( w,  wc,  ico, jco, kcw, r1xo, r2xo,     &
                                       r1yo, r2yo, r1zw, r2zw,                 &
                                       logc_w_r, logc_ratio_w_r,               &
                                       logc_kbounds_w_r, nzt_topo_nestbc_r,    &
                                       kctw, iflo, ifuo, jflo, jfuo, kflw,     &
                                       kfuw, ijkfc_w, 'r', 'w' )

             IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.   &
                  (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.    &
                     .NOT. constant_diffusion ) )  THEN
!                CALL pmci_interp_1sto_lr( e,  ec,  ico, jco, kco, r1xo, r2xo,  &
!                                          r1yo,r2yo, r1zo, r2zo,               &
!                                          logc_w_r, logc_ratio_w_r,            &
!                                          logc_kbounds_w_r, nzt_topo_nestbc_r, &
!                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
!                                          kfuo, ijkfc_s, 'r', 'e' )
!
!--             Interpolation of e is replaced by the Neumann condition. 
                DO ibgp = nx+1, nx+nbgp
                   e(nzb:nzt,nys:nyn,ibgp) = e(nzb:nzt,nys:nyn,nx)
                ENDDO
             ENDIF

             IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
                CALL pmci_interp_1sto_lr( diss,  dissc,  ico, jco, kco, r1xo,  &
                                          r2xo, r1yo,r2yo, r1zo, r2zo,         &
                                          logc_w_r, logc_ratio_w_r,            &
                                          logc_kbounds_w_r, nzt_topo_nestbc_r, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'r', 's' )

             ENDIF

             IF ( .NOT. neutral )  THEN
                CALL pmci_interp_1sto_lr( pt, ptc, ico, jco, kco, r1xo, r2xo,  &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_r, logc_ratio_w_r,            &
                                          logc_kbounds_w_r, nzt_topo_nestbc_r, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'r', 's' )
             ENDIF

             IF ( humidity )  THEN
                CALL pmci_interp_1sto_lr( q, q_c, ico, jco, kco, r1xo, r2xo,   &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_r, logc_ratio_w_r,            &
                                          logc_kbounds_w_r, nzt_topo_nestbc_r, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'r', 's' )

                IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN

                   CALL pmci_interp_1sto_lr( qc, qcc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_r, logc_ratio_w_r,         &
                                             logc_kbounds_w_r,                 &
                                             nzt_topo_nestbc_r,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'r', 's' ) 
     
                   CALL pmci_interp_1sto_lr( nc, ncc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_r, logc_ratio_w_r,         &
                                             logc_kbounds_w_r,                 &
                                             nzt_topo_nestbc_r,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'r', 's' )

                ENDIF

                IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN

     
                   CALL pmci_interp_1sto_lr( qr, qrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_r, logc_ratio_w_r,         &
                                             logc_kbounds_w_r,                 &
                                             nzt_topo_nestbc_r,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s,              &
                                             'r', 's' ) 

                   CALL pmci_interp_1sto_lr( nr, nrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_r, logc_ratio_w_r,         &
                                             logc_kbounds_w_r,                 &
                                             nzt_topo_nestbc_r,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'r', 's' ) 

                ENDIF

             ENDIF

             IF ( passive_scalar )  THEN
                CALL pmci_interp_1sto_lr( s, sc, ico, jco, kco, r1xo, r2xo,    &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_r, logc_ratio_w_r,            &
                                          logc_kbounds_w_r,                    &
                                          nzt_topo_nestbc_r,                   &
                                          kcto, iflo, ifuo, jflo, jfuo,        &
                                          kflo, kfuo, ijkfc_s, 'r', 's' )
             ENDIF

             IF ( air_chemistry  .AND.  nest_chemistry )  THEN
                DO  n = 1, nspec
                   CALL pmci_interp_1sto_lr( chem_species(n)%conc,             &
                                             chem_spec_c(:,:,:,n),             &
                                             ico, jco, kco, r1xo, r2xo,        &
                                             r1yo, r2yo, r1zo, r2zo,           &
                                             logc_w_r, logc_ratio_w_r,         &
                                             logc_kbounds_w_r,                 &
                                             nzt_topo_nestbc_r,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'r', 's' )
                ENDDO
             ENDIF
          ENDIF
!
!--       South border pe
          IF ( bc_dirichlet_s )  THEN

             CALL pmci_interp_1sto_sn( v,  vc,  ico, jcv, kco, r1xo, r2xo,     &
                                       r1yv, r2yv, r1zo, r2zo,                 &
                                       logc_v_s, logc_ratio_v_s,               &
                                       logc_kbounds_v_s, nzt_topo_nestbc_s,    &
                                       kcto, iflo, ifuo, jflv, jfuv, kflo,     &
                                       kfuo, ijkfc_v, 's', 'v' )

             CALL pmci_interp_1sto_sn( w,  wc,  ico, jco, kcw, r1xo, r2xo,     &
                                       r1yo, r2yo, r1zw, r2zw,                 &
                                       logc_w_s, logc_ratio_w_s,               &
                                       logc_kbounds_w_s, nzt_topo_nestbc_s,    &
                                       kctw, iflo, ifuo, jflo, jfuo, kflw,     &
                                       kfuw, ijkfc_w, 's','w' )

             CALL pmci_interp_1sto_sn( u,  uc,  icu, jco, kco, r1xu, r2xu,     &
                                       r1yo, r2yo, r1zo, r2zo,                 &
                                       logc_u_s, logc_ratio_u_s,               &
                                       logc_kbounds_u_s, nzt_topo_nestbc_s,    &
                                       kcto, iflu, ifuu, jflo, jfuo, kflo,     &
                                       kfuo, ijkfc_u, 's', 'u' )

             IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.   &
                  (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.    &
                     .NOT. constant_diffusion ) )  THEN
!                CALL pmci_interp_1sto_sn( e,  ec,  ico, jco, kco, r1xo, r2xo,  &
!                                          r1yo, r2yo, r1zo, r2zo,              &
!                                          logc_w_s, logc_ratio_w_s,            &
!                                          logc_kbounds_w_s, nzt_topo_nestbc_s, &
!                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
!                                          kfuo, ijkfc_s, 's', 'e' )
!
!--             Interpolation of e is replaced by the Neumann condition. 
                DO jbgp = -nbgp, -1
                   e(nzb:nzt,jbgp,nxl:nxr) = e(nzb:nzt,0,nxl:nxr)
                ENDDO
             ENDIF

             IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
                CALL pmci_interp_1sto_sn( diss, dissc,  ico, jco, kco, r1xo,   &
                                          r2xo, r1yo, r2yo, r1zo, r2zo,        &
                                          logc_w_s, logc_ratio_w_s,            &
                                          logc_kbounds_w_s, nzt_topo_nestbc_s, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 's', 's' )
             ENDIF

             IF ( .NOT. neutral )  THEN
                CALL pmci_interp_1sto_sn( pt, ptc, ico, jco, kco, r1xo, r2xo,  &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_s, logc_ratio_w_s,            &
                                          logc_kbounds_w_s, nzt_topo_nestbc_s, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 's', 's' )
             ENDIF

             IF ( humidity )  THEN
                CALL pmci_interp_1sto_sn( q, q_c, ico, jco, kco, r1xo, r2xo,   &
                                          r1yo,r2yo, r1zo, r2zo,               &
                                          logc_w_s, logc_ratio_w_s,            &
                                          logc_kbounds_w_s, nzt_topo_nestbc_s, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 's', 's' )

                IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN

                   CALL pmci_interp_1sto_sn( qc, qcc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo,r2yo, r1zo, r2zo,      &
                                             logc_w_s, logc_ratio_w_s,         &
                                             logc_kbounds_w_s,                 &
                                             nzt_topo_nestbc_s,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 's', 's' )

                   CALL pmci_interp_1sto_sn( nc, ncc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo,r2yo, r1zo, r2zo,      &
                                             logc_w_s, logc_ratio_w_s,         &
                                             logc_kbounds_w_s,                 &
                                             nzt_topo_nestbc_s,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 's', 's' )

                ENDIF

                IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN

                   CALL pmci_interp_1sto_sn( qr, qrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo,r2yo, r1zo, r2zo,      &
                                             logc_w_s, logc_ratio_w_s,         &
                                             logc_kbounds_w_s,                 &
                                             nzt_topo_nestbc_s,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 's', 's' )

                   CALL pmci_interp_1sto_sn( nr, nrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo,r2yo, r1zo, r2zo,      &
                                             logc_w_s, logc_ratio_w_s,         &
                                             logc_kbounds_w_s,                 &
                                             nzt_topo_nestbc_s,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 's', 's' )

                ENDIF

             ENDIF

             IF ( passive_scalar )  THEN
                CALL pmci_interp_1sto_sn( s, sc, ico, jco, kco, r1xo, r2xo,    &
                                          r1yo,r2yo, r1zo, r2zo,               &
                                          logc_w_s, logc_ratio_w_s,            &
                                          logc_kbounds_w_s,                    &
                                          nzt_topo_nestbc_s,                   &
                                          kcto, iflo, ifuo, jflo, jfuo,        &
                                          kflo, kfuo, ijkfc_s, 's', 's' )
             ENDIF

             IF ( air_chemistry  .AND.  nest_chemistry )  THEN
                DO  n = 1, nspec
                   CALL pmci_interp_1sto_sn( chem_species(n)%conc,             &
                                             chem_spec_c(:,:,:,n),             &
                                             ico, jco, kco, r1xo, r2xo,        &
                                             r1yo, r2yo, r1zo, r2zo,           &
                                             logc_w_s, logc_ratio_w_s,         &
                                             logc_kbounds_w_s,                 &
                                             nzt_topo_nestbc_s,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 's', 's' )
                ENDDO
             ENDIF
          ENDIF
!
!--       North border pe
          IF ( bc_dirichlet_n )  THEN
             
             CALL pmci_interp_1sto_sn( u,  uc,  icu, jco, kco, r1xu, r2xu,     &
                                       r1yo, r2yo, r1zo, r2zo,                 &
                                       logc_u_n, logc_ratio_u_n,               &
                                       logc_kbounds_u_n, nzt_topo_nestbc_n,    &
                                       kcto, iflu, ifuu, jflo, jfuo, kflo,     &
                                       kfuo, ijkfc_u, 'n', 'u' )

             CALL pmci_interp_1sto_sn( v,  vc,  ico, jcv, kco, r1xo, r2xo,     &
                                       r1yv, r2yv, r1zo, r2zo,                 &
                                       logc_v_n, logc_ratio_v_n,               &
                                       logc_kbounds_v_n, nzt_topo_nestbc_n,    &
                                       kcto, iflo, ifuo, jflv, jfuv, kflo,     &
                                       kfuo, ijkfc_v, 'n', 'v' )

             CALL pmci_interp_1sto_sn( w,  wc,  ico, jco, kcw, r1xo, r2xo,     &
                                       r1yo, r2yo, r1zw, r2zw,                 &
                                       logc_w_n, logc_ratio_w_n,               &
                                       logc_kbounds_w_n, nzt_topo_nestbc_n,    &
                                       kctw, iflo, ifuo, jflo, jfuo, kflw,     &
                                       kfuw, ijkfc_w, 'n', 'w' )

             IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.   & 
                  (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.    &
                     .NOT. constant_diffusion ) )  THEN
!                CALL pmci_interp_1sto_sn( e,  ec,  ico, jco, kco, r1xo, r2xo,  &
!                                          r1yo, r2yo, r1zo, r2zo,              &
!                                          logc_w_n, logc_ratio_w_n,            &
!                                          logc_kbounds_w_n, nzt_topo_nestbc_n, &
!                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
!                                          kfuo, ijkfc_s, 'n', 'e' )
!
!--             Interpolation of e is replaced by the Neumann condition. 
                DO jbgp = ny+1, ny+nbgp
                   e(nzb:nzt,jbgp,nxl:nxr) = e(nzb:nzt,ny,nxl:nxr)
                ENDDO
             ENDIF

             IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
                CALL pmci_interp_1sto_sn( diss, dissc,  ico, jco, kco, r1xo,   &
                                          r2xo, r1yo, r2yo, r1zo, r2zo,        &
                                          logc_w_n, logc_ratio_w_n,            &
                                          logc_kbounds_w_n, nzt_topo_nestbc_n, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'n', 's' )

             ENDIF

             IF ( .NOT. neutral )  THEN
                CALL pmci_interp_1sto_sn( pt, ptc, ico, jco, kco, r1xo, r2xo,  &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_n, logc_ratio_w_n,            &
                                          logc_kbounds_w_n, nzt_topo_nestbc_n, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'n', 's' )
             ENDIF

             IF ( humidity )  THEN
                CALL pmci_interp_1sto_sn( q, q_c, ico, jco, kco, r1xo, r2xo,   &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_n, logc_ratio_w_n,            &
                                          logc_kbounds_w_n, nzt_topo_nestbc_n, &
                                          kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                          kfuo, ijkfc_s, 'n', 's' )

                IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN

                   CALL pmci_interp_1sto_sn( qc, qcc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_n, logc_ratio_w_n,         &
                                             logc_kbounds_w_n,                 &
                                             nzt_topo_nestbc_n,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'n', 's' )

                   CALL pmci_interp_1sto_sn( nc, ncc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_u_n, logc_ratio_u_n,         &
                                             logc_kbounds_w_n,                 &
                                             nzt_topo_nestbc_n,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'n', 's' )

                ENDIF

                IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN

                   CALL pmci_interp_1sto_sn( qr, qrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_n, logc_ratio_w_n,         &
                                             logc_kbounds_w_n,                 &
                                             nzt_topo_nestbc_n,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'n', 's' )

                   CALL pmci_interp_1sto_sn( nr, nrc, ico, jco, kco, r1xo,     &
                                             r2xo, r1yo, r2yo, r1zo, r2zo,     &
                                             logc_w_n, logc_ratio_w_n,         &
                                             logc_kbounds_w_n,                 &
                                             nzt_topo_nestbc_n,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'n', 's' )

                ENDIF

             ENDIF

             IF ( passive_scalar )  THEN
                CALL pmci_interp_1sto_sn( s, sc, ico, jco, kco, r1xo, r2xo,    &
                                          r1yo, r2yo, r1zo, r2zo,              &
                                          logc_w_n, logc_ratio_w_n,            &
                                          logc_kbounds_w_n,                    &
                                          nzt_topo_nestbc_n,                   &
                                          kcto, iflo, ifuo, jflo, jfuo,        &
                                          kflo, kfuo, ijkfc_s, 'n', 's' )
             ENDIF

             IF ( air_chemistry  .AND.  nest_chemistry )  THEN
                DO  n = 1, nspec
                   CALL pmci_interp_1sto_sn( chem_species(n)%conc,             &
                                             chem_spec_c(:,:,:,n),             &
                                             ico, jco, kco, r1xo, r2xo,        &
                                             r1yo, r2yo, r1zo, r2zo,           &
                                             logc_w_n, logc_ratio_w_n,         &
                                             logc_kbounds_w_n,                 &
                                             nzt_topo_nestbc_n,                &
                                             kcto, iflo, ifuo, jflo, jfuo,     &
                                             kflo, kfuo, ijkfc_s, 'n', 's' )
                ENDDO
             ENDIF
          ENDIF

       ENDIF       ! IF ( nesting_mode /= 'vertical' )
!
!--    All PEs are top-border PEs
       CALL pmci_interp_1sto_t( w,  wc,  ico, jco, kcw, r1xo, r2xo, r1yo,      &
                                r2yo, r1zw, r2zw, kctw, iflo, ifuo,            &
                                jflo, jfuo, kflw, kfuw, ijkfc_w, 'w' )
       CALL pmci_interp_1sto_t( u,  uc,  icu, jco, kco, r1xu, r2xu, r1yo,      &
                                r2yo, r1zo, r2zo, kcto, iflu, ifuu,            &
                                jflo, jfuo, kflo, kfuo, ijkfc_u, 'u' )
       CALL pmci_interp_1sto_t( v,  vc,  ico, jcv, kco, r1xo, r2xo, r1yv,      &
                                r2yv, r1zo, r2zo, kcto, iflo, ifuo,            &
                                jflv, jfuv, kflo, kfuo, ijkfc_v, 'v' )


       IF ( (        rans_mode_parent  .AND.         rans_mode )  .OR.         &
            (  .NOT. rans_mode_parent  .AND.  .NOT.  rans_mode  .AND.          &
               .NOT. constant_diffusion ) )  THEN
!          CALL pmci_interp_1sto_t( e,  ec,  ico, jco, kco, r1xo, r2xo, r1yo,   &
!                                   r2yo, r1zo, r2zo, kcto, iflo, ifuo,         &
!                                   jflo, jfuo, kflo, kfuo, ijkfc_s, 'e' )
!
!--       Interpolation of e is replaced by the Neumann condition. 
          e(nzt+1,nys:nyn,nxl:nxr) = e(nzt,nys:nyn,nxl:nxr)

       ENDIF

       IF ( rans_mode_parent  .AND.  rans_mode  .AND.  rans_tke_e )  THEN
          CALL pmci_interp_1sto_t( diss, dissc, ico, jco, kco, r1xo, r2xo,     &
                                   r1yo, r2yo, r1zo, r2zo, kcto, iflo, ifuo,   &
                                   jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( .NOT. neutral )  THEN
          CALL pmci_interp_1sto_t( pt, ptc, ico, jco, kco, r1xo, r2xo,         &
                                   r1yo, r2yo, r1zo, r2zo, kcto, iflo, ifuo,   &
                                   jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( humidity )  THEN

          CALL pmci_interp_1sto_t( q, q_c, ico, jco, kco, r1xo, r2xo, r1yo,    &
                                   r2yo, r1zo, r2zo, kcto, iflo, ifuo,         &
                                   jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

          IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN

             CALL pmci_interp_1sto_t( qc, qcc, ico, jco, kco, r1xo, r2xo, r1yo,&
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

             CALL pmci_interp_1sto_t( nc, ncc, ico, jco, kco, r1xo, r2xo, r1yo,&
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

          ENDIF

          IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN


             CALL pmci_interp_1sto_t( qr, qrc, ico, jco, kco, r1xo, r2xo, r1yo,&
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

             CALL pmci_interp_1sto_t( nr, nrc, ico, jco, kco, r1xo, r2xo, r1yo,&
                                      r2yo, r1zo, r2zo, kcto, iflo, ifuo,      &
                                      jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )

          ENDIF

       ENDIF

       IF ( passive_scalar )  THEN
          CALL pmci_interp_1sto_t( s, sc, ico, jco, kco, r1xo, r2xo, r1yo,     &
                                   r2yo, r1zo, r2zo, kcto, iflo, ifuo,         &
                                   jflo, jfuo, kflo, kfuo, ijkfc_s, 's' )
       ENDIF

       IF ( air_chemistry  .AND.  nest_chemistry )  THEN
          DO  n = 1, nspec
             CALL pmci_interp_1sto_t( chem_species(n)%conc,                    &
                                      chem_spec_c(:,:,:,n),                    &
                                      ico, jco, kco, r1xo, r2xo,               &
                                      r1yo, r2yo, r1zo, r2zo,                  &
                                      kcto, iflo, ifuo, jflo, jfuo,            &
                                      kflo, kfuo, ijkfc_s, 's' )
          ENDDO
       ENDIF

   END SUBROUTINE pmci_interpolation



   SUBROUTINE pmci_anterpolation

!
!--   A wrapper routine for all anterpolation actions.
!--   Note that TKE is not anterpolated.
      IMPLICIT NONE

      INTEGER(iwp) ::  n          !< running index for number of chemical species

      CALL pmci_anterp_tophat( u,  uc,  kcto, iflu, ifuu, jflo, jfuo, kflo,    &
                               kfuo, ijkfc_u, 'u' )
      CALL pmci_anterp_tophat( v,  vc,  kcto, iflo, ifuo, jflv, jfuv, kflo,    &
                               kfuo, ijkfc_v, 'v' )
      CALL pmci_anterp_tophat( w,  wc,  kctw, iflo, ifuo, jflo, jfuo, kflw,    &
                               kfuw, ijkfc_w, 'w' )
!
!--   Anterpolation of TKE and dissipation rate if parent and child are in 
!--   RANS mode.
      IF ( rans_mode_parent  .AND.  rans_mode )  THEN
         CALL pmci_anterp_tophat( e, ec, kcto, iflo, ifuo, jflo, jfuo, kflo,   &
                                  kfuo, ijkfc_s, 'e' )
!
!--      Anterpolation of dissipation rate only if TKE-e closure is applied.
         IF ( rans_tke_e )  THEN
            CALL pmci_anterp_tophat( diss, dissc, kcto, iflo, ifuo, jflo, jfuo,&
                                     kflo, kfuo, ijkfc_s, 'diss' )
         ENDIF

      ENDIF

      IF ( .NOT. neutral )  THEN
         CALL pmci_anterp_tophat( pt, ptc, kcto, iflo, ifuo, jflo, jfuo, kflo, &
                                  kfuo, ijkfc_s, 'pt' )
      ENDIF

      IF ( humidity )  THEN

         CALL pmci_anterp_tophat( q, q_c, kcto, iflo, ifuo, jflo, jfuo, kflo,  &
                                  kfuo, ijkfc_s, 'q' )

         IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN

            CALL pmci_anterp_tophat( qc, qcc, kcto, iflo, ifuo, jflo, jfuo,    &
                                     kflo, kfuo, ijkfc_s, 'qc' )

            CALL pmci_anterp_tophat( nc, ncc, kcto, iflo, ifuo, jflo, jfuo,    &
                                     kflo, kfuo, ijkfc_s, 'nc' )

         ENDIF

         IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN

            CALL pmci_anterp_tophat( qr, qrc, kcto, iflo, ifuo, jflo, jfuo,    &
                                     kflo, kfuo, ijkfc_s, 'qr' )

            CALL pmci_anterp_tophat( nr, nrc, kcto, iflo, ifuo, jflo, jfuo,    &
                                     kflo, kfuo, ijkfc_s, 'nr' )

         ENDIF

      ENDIF

      IF ( passive_scalar )  THEN
         CALL pmci_anterp_tophat( s, sc, kcto, iflo, ifuo, jflo, jfuo, kflo,   &
                                  kfuo, ijkfc_s, 's' )
      ENDIF

      IF ( air_chemistry  .AND.  nest_chemistry )  THEN
         DO  n = 1, nspec
            CALL pmci_anterp_tophat( chem_species(n)%conc,                     &
                                     chem_spec_c(:,:,:,n),                     &
                                     kcto, iflo, ifuo, jflo, jfuo, kflo,       &
                                     kfuo, ijkfc_s, 's' )
         ENDDO
      ENDIF

   END SUBROUTINE pmci_anterpolation



   SUBROUTINE pmci_interp_1sto_lr( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y, r1z, &
                                   r2z, logc, logc_ratio, logc_kbounds,        &
                                   nzt_topo_nestbc,                            &
                                   kct, ifl, ifu, jfl, jfu, kfl, kfu, ijkfc,   &
                                   edge, var )
!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the left and right boundaries. It is
!--   based on trilinear interpolation.

      IMPLICIT NONE

      INTEGER(iwp) ::  nzt_topo_nestbc   !<

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f       !<
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc      !<
      REAL(wp), DIMENSION(1:2,0:ncorr-1,nzb:nzt_topo_nestbc,nys:nyn),          &
                                      INTENT(IN)    ::  logc_ratio   !<
!AH      REAL(wp), DIMENSION(nxlg:nxrg), INTENT(IN)    ::  r1x     !<
!AH      REAL(wp), DIMENSION(nxlg:nxrg), INTENT(IN)    ::  r2x     !<
!AH      REAL(wp), DIMENSION(nysg:nyng), INTENT(IN)    ::  r1y     !<
!AH      REAL(wp), DIMENSION(nysg:nyng), INTENT(IN)    ::  r2y     !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x     !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x     !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y     !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y     !<

!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z     !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z     !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z     !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z     !<
      

      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)         ::  ic     !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)         ::  jc     !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)           ::  kc     !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN)        ::  kc     !<
      INTEGER(iwp), DIMENSION(1:2,nzb:nzt_topo_nestbc,nys:nyn),                &
                                          INTENT(IN)           :: logc   !<
      INTEGER(iwp), DIMENSION(1:2,nys:nyn), INTENT(IN)         :: logc_kbounds !<

      INTEGER(iwp) :: kct

!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH

!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction

!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) ::  edge   !< Edge symbol: 'l', 'r', 's' or 'n' 
      CHARACTER(LEN=1), INTENT(IN) ::  var    !< Variable symbol: 'u', 'v', 'w' or 's'

      INTEGER(iwp) ::  i        !< Lower bound of the running index ia
      INTEGER(iwp) ::  ia       !< i-index running over the parent-grid cell on the boundary 
      INTEGER(iwp) ::  iaw      !< Reduced ia-index for workarr_lr
      INTEGER(iwp) ::  iawbc    !< iaw-index pointing to the boundary-value nodes (either 0 or igsr-1)      
      INTEGER(iwp) ::  ib       !< Fixed i-index pointing to the node just behind the boundary-value node
      INTEGER(iwp) ::  ibc      !< Fixed i-index pointing to the boundary-value nodes (either i or iend)
      INTEGER(iwp) ::  ibgp     !< Index running over the redundant boundary ghost points in i-direction
!AH      INTEGER(iwp) ::  ibeg     !< i-index pointing to the starting point of workarr_lr in the i-direction 
      INTEGER(iwp) ::  iend     !< Upper bound of the running index ia
      INTEGER(iwp) ::  ierr     !< MPI error code
      INTEGER(iwp) ::  ijk      !< Running index for all child-grid cells within the anterpolation cell
      INTEGER(iwp) ::  iw       !< i-index for wall_flags_0
      INTEGER(iwp) ::  j        !< Running index in the y-direction
      INTEGER(iwp) ::  jco      !<
      INTEGER(iwp) ::  jcorr    !<
      INTEGER(iwp) ::  jinc     !<
      INTEGER(iwp) ::  jw       !< j-index for wall_flags_0
      INTEGER(iwp) ::  j1       !<
      INTEGER(iwp) ::  k        !< Running index in the z-direction
      INTEGER(iwp) ::  k_wall   !< Vertical index of topography top
      INTEGER(iwp) ::  kco      !<
      INTEGER(iwp) ::  kcorr    !<
      INTEGER(iwp) ::  kw       !< k-index for wall_flags_0
      INTEGER(iwp) ::  k1       !<
      INTEGER(iwp) ::  l        !< Parent-grid running index in the x-direction
      INTEGER(iwp) ::  lbeg     !< l-index pointing to the starting point of workarrc_lr in the l-direction 
      INTEGER(iwp) ::  lp1      !< l+1
      INTEGER(iwp) ::  loff     !< l-offset needed on the right boundary to correctly refer to boundary ghost points
      INTEGER(iwp) ::  lw       !< Reduced l-index for workarrc_lr
      INTEGER(iwp) ::  m        !< Parent-grid running index in the y-direction
      INTEGER(iwp) ::  mnorthv  !< Upshift by one for the upper bound of index m in case of var == 'v'
      INTEGER(iwp) ::  mp1      !< m+1
      INTEGER(iwp) ::  n        !< Parent-grid running index in the z-direction
      INTEGER(iwp) ::  np1      !< n+1
      INTEGER(iwp) ::  ntopw    !< Upshift by one for the upper bound of index n in case of var == 'w'
      INTEGER(iwp) ::  var_flag !< Variable flag for BTEST( wall_flags_0 )

      REAL(wp) ::  cellsum      !< Sum of child-grid node values over the anterpolation cell 
      REAL(wP) ::  cellsumd     !< Sum of differences over the anterpolation cell 
      REAL(wp) ::  fkj          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkjp         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpj         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpjp        !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fk           !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkp          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  rcorr        !< Average reversibility correction for the whole anterpolation cell
      REAL(wp) ::  rcorr_ijk    !< Reversibility correction distributed to the individual child-grid nodes

! 
!--   Check which edge is to be handled
      IF ( edge == 'l' )  THEN
!
!--      For u, nxl is a ghost node, but not for the other variables
         IF ( var == 'u' )  THEN
            i     = nxl
            iend  = nxl
            ibc   = nxl    
            ib    = ibc - 1
            iawbc = 0
            l     = icl + 2
            lw    = 2
            lbeg  = icl
            loff  = 0            
         ELSE
            i     = nxl - igsr
            iend  = nxl - 1
            ibc   = nxl - 1
            ib    = ibc - 1
            iawbc = igsr-1
            l     = icl + 1
            lw    = 1
            lbeg  = icl
            loff  = 0
         ENDIF        
      ELSEIF ( edge == 'r' )  THEN
         IF ( var == 'u' )  THEN
            i     = nxr + 1            
            iend  = nxr + 1
            ibc   = nxr + 1 
            ib    = ibc + 1
            iawbc = 0
            l     = icr - 1
            lw    = 1
            lbeg  = icr - 2
            loff  = 0
         ELSE
            i     = nxr + 1
            iend  = nxr + igsr
            ibc   = nxr + 1
            ib    = ibc + 1
            iawbc = 0
            l     = icr - 1
            lw    = 1
            lbeg  = icr - 2
            loff  = 1
         ENDIF         
      ENDIF

      IF  ( var == 'w' )  THEN
         ntopw = 1
      ELSE
         ntopw = 0
      ENDIF

      IF  ( var == 'v' )  THEN
         mnorthv = 0
      ELSE
         mnorthv = 1
      ENDIF

      IF ( var == 'u' )  THEN 
         var_flag = 1 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2 
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF

      IF  ( var == 'u' )  THEN
!AH!
!AH!--      Substitute the necessary parent-grid data to the work array workarrc_lr. 
!AH         workarrc_lr = 0.0_wp     
!AH         IF  ( pdims(2) > 1 )  THEN
!AH#if defined( __parallel )
!AH            IF  ( nys == 0 )  THEN 
!AH               workarrc_lr(0:cg%nz+1,jcsw:jcnw-1,0:2)                           &
!AH                    = fc(0:cg%nz+1,jcsw:jcnw-1,lbeg:lbeg+2)
!AH            ELSE IF  ( nyn == ny )  THEN
!AH               workarrc_lr(0:cg%nz+1,jcsw+1:jcnw,0:2)                           &
!AH                    = fc(0:cg%nz+1,jcsw+1:jcnw,lbeg:lbeg+2)
!AH            ELSE
!AH               workarrc_lr(0:cg%nz+1,jcsw+1:jcnw-1,0:2)                         &
!AH                    = fc(0:cg%nz+1,jcsw+1:jcnw-1,lbeg:lbeg+2)
!AH            ENDIF
!AH!
!AH!--         South-north exchange if more than one PE subdomain in the y-direction.
!AH!--         Note that in case of 3-D nesting the south (psouth == MPI_PROC_NULL) 
!AH!--         and north (pnorth == MPI_PROC_NULL) boundaries are not exchanged 
!AH!--         because the nest domain is not cyclic.
!AH!--         From south to north
!AH            CALL MPI_SENDRECV( workarrc_lr(0,jcsw+1,0), 1,                      &
!AH                 workarrc_lr_exchange_type, psouth,  0,                         &
!AH                 workarrc_lr(0,jcnw,0), 1,                                      &
!AH                 workarrc_lr_exchange_type, pnorth,  0,                         &
!AH                 comm2d, status, ierr ) 
!AH!
!AH!--         From north to south        
!AH            CALL MPI_SENDRECV( workarrc_lr(0,jcnw-1,0), 1,                      &
!AH                 workarrc_lr_exchange_type, pnorth,  1,                         &
!AH                 workarrc_lr(0,jcsw,0), 1,                                      &
!AH                 workarrc_lr_exchange_type, psouth,  1,                         &
!AH                 comm2d, status, ierr )
!AH#endif
!AH         ELSE
!AH            workarrc_lr(0:cg%nz+1,jcsw:jcnw,0:2)                                &
!AH                 = fc(0:cg%nz+1,jcsw:jcnw,lbeg:lbeg+2)            
!AH         ENDIF

         DO  m = jcsw+1, jcnw-1
            DO n = 0, kct 

               DO  j = jfl(m), jfu(m)
                  DO  k = kfl(n), kfu(n)
                     f(k,j,ibc) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'v' )  THEN
         
         DO  m = jcsw+1, jcnw-1
            DO n = 0, kct 

               DO  j = jfl(m), jfl(m+1)-1
                  DO  k = kfl(n), kfu(n) 
                     f(k,j,ibc) = fc(n,m,l) 
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'w' )  THEN

         DO  m = jcsw+1, jcnw-1
            DO n = 1, kct + 1   ! It is important to go up to kct+1  

               DO  j = jfl(m), jfu(m)
                  f(nzb,j,ibc) = 0.0_wp   ! Because the n-loop starts from n=1 instead of 0
                  DO  k = kfu(n-1)+1, kfu(n) 
                     f(k,j,ibc) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE   ! scalars
         
         DO  m = jcsw+1, jcnw-1
            DO n = 0, kct 
               
               DO  j = jfl(m), jfu(m)
                  DO  k = kfl(n), kfu(n)
                     f(k,j,ibc) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ENDIF  ! var
!
!--   Fill up also the redundant 2nd and 3rd ghost-node layers
      IF ( edge == 'l' )  THEN
         DO  ibgp = -nbgp, ib
            f(0:nzt+1,nysg:nyng,ibgp) = f(0:nzt+1,nysg:nyng,ibc)
         ENDDO
      ELSEIF ( edge == 'r' )  THEN
         DO  ibgp = ib, nx+nbgp
            f(0:nzt+1,nysg:nyng,ibgp) = f(0:nzt+1,nysg:nyng,ibc)
         ENDDO
      ENDIF

   END SUBROUTINE pmci_interp_1sto_lr



   SUBROUTINE pmci_interp_1sto_sn( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y, r1z, &
                                   r2z, logc, logc_ratio, logc_kbounds,        &
                                   nzt_topo_nestbc,                            &
                                   kct, ifl, ifu, jfl, jfu, kfl, kfu, ijkfc,   &
                                   edge, var )

!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the south and north boundaries. 
!--   This subroutine is based on trilinear interpolation.

      IMPLICIT NONE

      INTEGER(iwp) ::  nzt_topo_nestbc   !<

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f             !<
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc            !<
      REAL(wp), DIMENSION(1:2,0:ncorr-1,nzb:nzt_topo_nestbc,nxl:nxr),          &
                                      INTENT(IN)    ::  logc_ratio    !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x           !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x           !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y           !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y           !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z           !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z           !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z           !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z           !<

      
      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)         ::  ic    !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)         ::  jc    !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)           ::  kc    !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN)        ::  kc    !<
      INTEGER(iwp), DIMENSION(1:2,nzb:nzt_topo_nestbc,nxl:nxr),                &
                                          INTENT(IN)           ::  logc  !<
      INTEGER(iwp), DIMENSION(1:2,nxl:nxr), INTENT(IN)         ::  logc_kbounds  !< 

      INTEGER(iwp) :: kct
!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH

!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) ::  edge   !< Edge symbol: 'l', 'r', 's' or 'n' 
      CHARACTER(LEN=1), INTENT(IN) ::  var    !< Variable symbol: 'u', 'v', 'w' or 's'
      
      INTEGER(iwp) ::  i        !< Running index in the x-direction
      INTEGER(iwp) ::  iinc     !<
      INTEGER(iwp) ::  icorr    !<
      INTEGER(iwp) ::  ico      !<
      INTEGER(iwp) ::  ierr     !< MPI error code
      INTEGER(iwp) ::  ijk      !< Running index for all child-grid cells within the anterpolation cell
      INTEGER(iwp) ::  iw       !< i-index for wall_flags_0
      INTEGER(iwp) ::  i1       !<
      INTEGER(iwp) ::  j        !< Lower bound of the running index ja
      INTEGER(iwp) ::  ja       !< Index in y-direction running over the parent-grid cell on the boundary 
      INTEGER(iwp) ::  jaw      !< Reduced ja-index for workarr_sn
      INTEGER(iwp) ::  jawbc    !< jaw-index pointing to the boundary-value nodes (either 0 or jgsr-1)
!AH      INTEGER(iwp) ::  jbeg     !< j-index pointing to the starting point of workarr_sn in the j-direction
      INTEGER(iwp) ::  jb       !< Fixed j-index pointing to the node just behind the boundary-value node
      INTEGER(iwp) ::  jbc      !< Fixed j-index pointing to the boundary-value nodes (either j or jend)
      INTEGER(iwp) ::  jbgp     !< Index running over the redundant boundary ghost points in j-direction
      INTEGER(iwp) ::  jend     !< Upper bound of the running index ja
      INTEGER(iwp) ::  jw       !< j-index for wall_flags_0
      INTEGER(iwp) ::  k        !< Running index in the z-direction
      INTEGER(iwp) ::  k_wall   !< Vertical index of topography top
      INTEGER(iwp) ::  kcorr    !<
      INTEGER(iwp) ::  kco      !<
      INTEGER(iwp) ::  kw       !< k-index for wall_flags_0
      INTEGER(iwp) ::  k1       !<
      INTEGER(iwp) ::  l        !< Parent-grid running index in the x-direction
      INTEGER(iwp) ::  lp1      !< l+1
      INTEGER(iwp) ::  lrightu  !< Upshift by one for the upper bound of index l in case of var == 'u'
      INTEGER(iwp) ::  m        !< Parent-grid running index in the y-direction
      INTEGER(iwp) ::  mbeg     !< m-index pointing to the starting point of workarrc_sn in the m-direction 
      INTEGER(iwp) ::  moff     !< m-offset needed on the north boundary to correctly refer to boundary ghost points
      INTEGER(iwp) ::  mp1      !< m+1
      INTEGER(iwp) ::  mw       !< Reduced m-index for workarrc_sn
      INTEGER(iwp) ::  n        !< Parent-grid running index in the z-direction
      INTEGER(iwp) ::  np1      !< n+1
      INTEGER(iwp) ::  ntopw    !< Upshift by one for the upper bound of index n in case of var == 'w'
      INTEGER(iwp) ::  var_flag !< Variable flag for BTEST( wall_flags_0 )

      REAL(wp) ::  cellsum      !< Sum of child-grid node values over the anterpolation cell 
      REAL(wp) ::  cellsumd     !< Sum of differences over the anterpolation cell 
      REAL(wp) ::  fk           !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkj          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkjp         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpj         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpjp        !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkp          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  rcorr        !< Average reversibility correction for the whole anterpolation cell
      REAL(wp) ::  rcorr_ijk    !< Reversibility correction distributed to the individual child-grid nodes

!
!--   Check which edge is to be handled: south or north
      IF ( edge == 's' )  THEN
!
!--      For v, nys is a ghost node, but not for the other variables
         IF ( var == 'v' )  THEN
            j     = nys
            jend  = nys 
            jawbc = 0
            jbc   = nys
            jb    = jbc - 1
            m     = jcs + 2
            mw    = 2
            mbeg  = jcs
            moff  = 0
         ELSE
            j     = nys - jgsr
            jend  = nys - 1
            jawbc = jgsr - 1
            jbc   = nys - 1
            jb    = jbc - 1
            m     = jcs + 1
            mw    = 1
            mbeg  = jcs
            moff  = 0
         ENDIF
      ELSEIF ( edge == 'n' )  THEN
         IF ( var == 'v' )  THEN
            j     = nyn + 1
            jend  = nyn + 1
            jawbc = 0
            jbc   = nyn + 1
            jb    = jbc + 1
            m     = jcn - 1
            mw    = 1
            mbeg  = jcn - 2
            moff  = 0
         ELSE
            j     = nyn + 1
            jend  = nyn + jgsr
            jawbc = 0
            jbc   = nyn + 1
            jb    = jbc + 1
            m     = jcn - 1
            mw    = 1
            mbeg  = jcn - 2
            moff  = 1
         ENDIF
      ENDIF

      IF  ( var == 'w' )  THEN
         ntopw = 1
      ELSE
         ntopw = 0
      ENDIF

      IF  ( var == 'u' )  THEN
         lrightu = 0
      ELSE
         lrightu = 1
      ENDIF

      IF ( var == 'u' )  THEN 
         var_flag = 1 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2 
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF

      IF  ( var == 'v' )  THEN
!AH!
!AH!--      Substitute the necessary parent-grid data to the work array workarrc_sn. 
!AH         workarrc_sn = 0.0_wp     
!AH         IF  ( pdims(1) > 1 )  THEN
!AH#if defined( __parallel )
!AH            IF  ( nxl == 0 )  THEN   ! if ( bc_dirichlet_l )
!AH               workarrc_sn(0:cg%nz+1,0:2,iclw:icrw-1)                           &
!AH                    = fc(0:cg%nz+1,mbeg:mbeg+2,iclw:icrw-1)
!AH            ELSE IF  ( nxr == nx )  THEN    ! if ( bc_dirichlet_r )
!AH               workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw)                           &
!AH                    = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw)
!AH            ELSE
!AH               workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw-1)                         &
!AH                    = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw-1)
!AH            ENDIF
!AH!
!AH!--         Left-right exchange if more than one PE subdomain in the x-direction.
!AH!--         Note that in case of 3-D nesting the left (pleft == MPI_PROC_NULL) and 
!AH!--         right (pright == MPI_PROC_NULL) boundaries are not exchanged because 
!AH!--         the nest domain is not cyclic.
!AH!--         From left to right
!AH            CALL MPI_SENDRECV( workarrc_sn(0,0,iclw+1), 1,                      &
!AH                 workarrc_sn_exchange_type, pleft,   0,                         &
!AH                 workarrc_sn(0,0,icrw), 1,                                      &
!AH                 workarrc_sn_exchange_type, pright,  0,                         &
!AH                 comm2d, status, ierr ) 
!AH!
!AH!--         From right to left        
!AH            CALL MPI_SENDRECV( workarrc_sn(0,0,icrw-1), 1,                      &
!AH                 workarrc_sn_exchange_type, pright,  1,                         &
!AH                 workarrc_sn(0,0,iclw), 1,                                      &
!AH                 workarrc_sn_exchange_type, pleft,   1,                         &
!AH                 comm2d, status, ierr )
!AH#endif      
!AH         ELSE
!AH            workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw-1)                            &
!AH                 = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw-1)
!AH         ENDIF

         DO  l = iclw+1, icrw-1
            DO n = 0, kct 

               DO  i = ifl(l), ifu(l)
                  DO  k = kfl(n), kfu(n)
                     f(k,jbc,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'u' )  THEN
         
         DO  l = iclw+1, icrw-1
            DO n = 0, kct 
               
               DO  i = ifl(l), ifl(l+1)-1
                  DO  k = kfl(n), kfu(n) 
                     f(k,jbc,i) = fc(n,m,l) 
!AH
!                     write(9,"('sn u: ',6(i3,2x),2(e12.5,2x))") n, m, l, k, jbc, i, fc(n,m,l), f(k,jbc,i)
!                     flush(9)
!AH
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'w' )  THEN

         DO  l = iclw+1, icrw-1
            DO n = 1, kct + 1   ! It is important to go up to kct+1  
               
               DO  i = ifl(l), ifu(l)
                  f(nzb,jbc,i) = 0.0_wp   ! Because the n-loop starts from n=1 instead of 0
                  DO  k = kfu(n-1)+1, kfu(n) 
                     f(k,jbc,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE   ! scalars
         
         DO  l = iclw+1, icrw-1
            DO n = 0, kct 
               
               DO  i = ifl(l), ifu(l)
                  DO  k = kfl(n), kfu(n)
                     f(k,jbc,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ENDIF  ! var
!
!--   Fill up also the redundant 2nd and 3rd ghost-node layers
      IF ( edge == 's' )  THEN
         DO  jbgp = -nbgp, jb
            f(0:nzt+1,jbgp,nxlg:nxrg) = f(0:nzt+1,jbc,nxlg:nxrg)
         ENDDO
      ELSEIF ( edge == 'n' )  THEN
         DO  jbgp = jb, ny+nbgp
            f(0:nzt+1,jbgp,nxlg:nxrg) = f(0:nzt+1,jbc,nxlg:nxrg)
         ENDDO
      ENDIF

   END SUBROUTINE pmci_interp_1sto_sn



   SUBROUTINE pmci_interp_1sto_t( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y,       &
                                  r1z, r2z, kct, ifl, ifu, jfl, jfu, kfl, kfu, &
                                  ijkfc, var )

!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the top boundary. 
!--   This subroutine is based on trilinear interpolation.

      IMPLICIT NONE

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f     !< Child-grid array
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc    !< Parent-grid array
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x   !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x   !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y   !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y   !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z   !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z   !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z   !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z   !<

      
      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  ic    !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  jc    !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  kc    !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  kc    !<

      INTEGER(iwp) :: kct
!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) :: var   !<

      INTEGER(iwp) ::  i    !<
      INTEGER(iwp) ::  ib   !<
      INTEGER(iwp) ::  iclc !< Lower i-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  icrc !< Upper i-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  ie   !<
      INTEGER(iwp) ::  ierr !< MPI error code
      INTEGER(iwp) ::  ijk  !<
      INTEGER(iwp) ::  iw   !<
      INTEGER(iwp) ::  j    !<
      INTEGER(iwp) ::  jb   !<
      INTEGER(iwp) ::  jcsc !< Lower j-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  jcnc !< Upper j-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  je   !< 
      INTEGER(iwp) ::  jw   !<     
      INTEGER(iwp) ::  k    !< Vertical child-grid index fixed to the boundary-value level 
      INTEGER(iwp) ::  ka   !< Running vertical child-grid index
      INTEGER(iwp) ::  kw   !< 
      INTEGER(iwp) ::  l    !< Parent-grid index in x-direction
      INTEGER(iwp) ::  lp1  !< l+1
      INTEGER(iwp) ::  m    !< Parent-grid index in y-direction
      INTEGER(iwp) ::  mp1  !< m+1
      INTEGER(iwp) ::  n    !< Parent-grid work array index in z-direction
      INTEGER(iwp) ::  np1  !< n+1
      INTEGER(iwp) ::  noff !< n-offset needed on the top boundary to correctly refer to boundary ghost points  
      INTEGER(iwp) ::  nw   !< n-index for workarrc_t
      INTEGER(iwp) ::  var_flag  !<
      
      REAL(wp) ::  cellsum     !<
      REAL(wp) ::  cellsumd    !<
      REAL(wp) ::  fac
      REAL(wp) ::  fk          !<
      REAL(wp) ::  fkj         !<
      REAL(wp) ::  fkjp        !<
      REAL(wp) ::  fkpj        !<
      REAL(wp) ::  fkpjp       !<
      REAL(wp) ::  fkp         !<
      REAL(wp) ::  rcorr       !<
      REAL(wp) ::  rcorr_ijk   !<


      IF ( var == 'w' )  THEN
         k    = nzt
         noff = 0
      ELSE
         k    = nzt + 1
         noff = 1
      ENDIF

      IF ( var == 'u' )  THEN 
         var_flag = 1 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2      
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF
      n  = kc(k) + noff
      nw = 1 

      IF  ( var == 'w' )  THEN
!AH!
!AH!--      Substitute the necessary parent-grid data to the work array. 
!AH!--      Note that the dimension of workarrc_t is (0:2,jcsw:jcnw,iclw:icrw),
!AH!--      And the jc?w and ic?w-index bounds depend on the location of the PE-
!AH!--      subdomain relative to the side boundaries. 
!AH         iclc = iclw + 1
!AH         icrc = icrw - 1      
!AH         jcsc = jcsw + 1
!AH         jcnc = jcnw - 1
!AH         IF  ( bc_dirichlet_l )  THEN
!AH            iclc = iclw
!AH         ENDIF
!AH         IF  ( bc_dirichlet_r )  THEN
!AH            icrc = icrw
!AH         ENDIF
!AH         IF  ( bc_dirichlet_s )  THEN
!AH            jcsc = jcsw
!AH         ENDIF
!AH         IF  ( bc_dirichlet_n )  THEN
!AH            jcnc = jcnw
!AH         ENDIF
!AH         workarrc_t = 0.0_wp
!AH         workarrc_t(0:2,jcsc:jcnc,iclc:icrc)                                    &
!AH              = fc(kc(k):kc(k)+2,jcsc:jcnc,iclc:icrc)
!AH!
!AH!--      Left-right exchange if more than one PE subdomain in the x-direction.
!AH!--      Note that in case of 3-D nesting the left and right boundaries are 
!AH!--      not exchanged because the nest domain is not cyclic.
!AH#if defined( __parallel )
!AH         IF  ( pdims(1) > 1 )  THEN
!AH!
!AH!--         From left to right
!AH            CALL MPI_SENDRECV( workarrc_t(0,jcsw,iclw+1), 1,                    &
!AH                 workarrc_t_exchange_type_y, pleft,  0,                         &
!AH                 workarrc_t(0,jcsw,icrw), 1,                                    &
!AH                 workarrc_t_exchange_type_y, pright, 0,                         &
!AH                 comm2d, status, ierr ) 
!AH!
!AH!--         From right to left        
!AH            CALL MPI_SENDRECV( workarrc_t(0,jcsw,icrw-1), 1,                    &
!AH                 workarrc_t_exchange_type_y, pright, 1,                         &
!AH                 workarrc_t(0,jcsw,iclw), 1,                                    &
!AH                 workarrc_t_exchange_type_y, pleft,  1,                         &
!AH                 comm2d, status, ierr )
!AH         ENDIF
!AH!
!AH!--      South-north exchange if more than one PE subdomain in the y-direction.
!AH!--      Note that in case of 3-D nesting the south and north boundaries are 
!AH!--      not exchanged because the nest domain is not cyclic.
!AH         IF  ( pdims(2) > 1 )  THEN
!AH!
!AH!--         From south to north         
!AH            CALL MPI_SENDRECV( workarrc_t(0,jcsw+1,iclw), 1,                    &
!AH                 workarrc_t_exchange_type_x, psouth, 2,                         &
!AH                 workarrc_t(0,jcnw,iclw), 1,                                    &
!AH                 workarrc_t_exchange_type_x, pnorth, 2,                         &
!AH                 comm2d, status, ierr ) 
!AH!
!AH!--         From north to south        
!AH            CALL MPI_SENDRECV( workarrc_t(0,jcnw-1,iclw), 1,                    &
!AH                 workarrc_t_exchange_type_x, pnorth, 3,                         &
!AH                 workarrc_t(0,jcsw,iclw), 1,                                    &
!AH                 workarrc_t_exchange_type_x, psouth, 3,                         &
!AH                 comm2d, status, ierr )
!AH         ENDIF
!AH#endif      
!AH

         DO  l = iclw+1, icrw-1
            DO  m = jcsw+1, jcnw-1

               DO  i = ifl(l), ifu(l)
                  DO  j = jfl(m), jfu(m)
                     f(k,j,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'u' )  THEN
         
         DO  l = iclw+1, icrw-1
            DO  m = jcsw+1, jcnw-1
               
               DO  i = ifl(l), ifl(l+1)-1
                  DO  j = jfl(m), jfu(m) 
                     f(k,j,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE IF  ( var == 'v' )  THEN

         DO  l = iclw+1, icrw-1
            DO  m = jcsw+1, jcnw-1

               DO  i = ifl(l), ifu(l)
                  DO  j = jfl(m), jfl(m+1)-1
                     f(k,j,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ELSE   ! scalars
         
         DO  l = iclw+1, icrw-1
            DO  m = jcsw+1, jcnw-1

               DO  i = ifl(l), ifu(l)
                  DO  j = jfl(m), jfu(m)
                     f(k,j,i) = fc(n,m,l)
                  ENDDO
               ENDDO

            ENDDO
         ENDDO

      ENDIF  ! var
!
!--   Just fill up the redundant second ghost-node layer in case of var == w.
      IF ( var == 'w' )  THEN
         f(nzt+1,:,:) = f(nzt,:,:)
      ENDIF

   END SUBROUTINE pmci_interp_1sto_t



   SUBROUTINE pmci_interp_tril_lr( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y, r1z, &
                                   r2z, logc, logc_ratio, logc_kbounds,        &
                                   nzt_topo_nestbc,                            &
                                   kct, ifl, ifu, jfl, jfu, kfl, kfu, ijkfc,   &
                                   edge, var )
!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the left and right boundaries. It is
!--   based on trilinear interpolation.

      IMPLICIT NONE

      INTEGER(iwp) ::  nzt_topo_nestbc   !<

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f       !<
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc      !<
      REAL(wp), DIMENSION(1:2,0:ncorr-1,nzb:nzt_topo_nestbc,nys:nyn),          &
                                      INTENT(IN)    ::  logc_ratio   !<
!AH      REAL(wp), DIMENSION(nxlg:nxrg), INTENT(IN)    ::  r1x     !<
!AH      REAL(wp), DIMENSION(nxlg:nxrg), INTENT(IN)    ::  r2x     !<
!AH      REAL(wp), DIMENSION(nysg:nyng), INTENT(IN)    ::  r1y     !<
!AH      REAL(wp), DIMENSION(nysg:nyng), INTENT(IN)    ::  r2y     !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x     !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x     !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y     !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y     !<

!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z     !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z     !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z     !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z     !<
      

      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)         ::  ic     !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)         ::  jc     !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)           ::  kc     !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN)        ::  kc     !<
      INTEGER(iwp), DIMENSION(1:2,nzb:nzt_topo_nestbc,nys:nyn),                &
                                          INTENT(IN)           :: logc   !<
      INTEGER(iwp), DIMENSION(1:2,nys:nyn), INTENT(IN)         :: logc_kbounds !<

      INTEGER(iwp) :: kct

!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH

!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction

!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) ::  edge   !< Edge symbol: 'l', 'r', 's' or 'n' 
      CHARACTER(LEN=1), INTENT(IN) ::  var    !< Variable symbol: 'u', 'v', 'w' or 's'

      INTEGER(iwp) ::  i        !< Lower bound of the running index ia
      INTEGER(iwp) ::  ia       !< i-index running over the parent-grid cell on the boundary 
      INTEGER(iwp) ::  iaw      !< Reduced ia-index for workarr_lr
      INTEGER(iwp) ::  iawbc    !< iaw-index pointing to the boundary-value nodes (either 0 or igsr-1)      
      INTEGER(iwp) ::  ibc      !< Fixed i-index pointing to the boundary-value nodes (either i or iend)
!AH      INTEGER(iwp) ::  ibeg     !< i-index pointing to the starting point of workarr_lr in the i-direction 
      INTEGER(iwp) ::  iend     !< Upper bound of the running index ia
      INTEGER(iwp) ::  ierr     !< MPI error code
      INTEGER(iwp) ::  ijk      !< Running index for all child-grid cells within the anterpolation cell
      INTEGER(iwp) ::  iw       !< i-index for wall_flags_0
      INTEGER(iwp) ::  j        !< Running index in the y-direction
      INTEGER(iwp) ::  jco      !<
      INTEGER(iwp) ::  jcorr    !<
      INTEGER(iwp) ::  jinc     !<
      INTEGER(iwp) ::  jw       !< j-index for wall_flags_0
      INTEGER(iwp) ::  j1       !<
      INTEGER(iwp) ::  k        !< Running index in the z-direction
      INTEGER(iwp) ::  k_wall   !< Vertical index of topography top
      INTEGER(iwp) ::  kco      !<
      INTEGER(iwp) ::  kcorr    !<
      INTEGER(iwp) ::  kw       !< k-index for wall_flags_0
      INTEGER(iwp) ::  k1       !<
      INTEGER(iwp) ::  l        !< Parent-grid running index in the x-direction
      INTEGER(iwp) ::  lbeg     !< l-index pointing to the starting point of workarrc_lr in the l-direction 
      INTEGER(iwp) ::  lp1      !< l+1
      INTEGER(iwp) ::  loff     !< l-offset needed on the right boundary to correctly refer to boundary ghost points
      INTEGER(iwp) ::  lw       !< Reduced l-index for workarrc_lr
      INTEGER(iwp) ::  m        !< Parent-grid running index in the y-direction
      INTEGER(iwp) ::  mnorthv  !< Upshift by one for the upper bound of index m in case of var == 'v'
      INTEGER(iwp) ::  mp1      !< m+1
      INTEGER(iwp) ::  n        !< Parent-grid running index in the z-direction
      INTEGER(iwp) ::  np1      !< n+1
      INTEGER(iwp) ::  ntopw    !< Upshift by one for the upper bound of index n in case of var == 'w'
      INTEGER(iwp) ::  var_flag !< Variable flag for BTEST( wall_flags_0 )

      REAL(wp) ::  cellsum      !< Sum of child-grid node values over the anterpolation cell 
      REAL(wP) ::  cellsumd     !< Sum of differences over the anterpolation cell 
      REAL(wp) ::  fkj          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkjp         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpj         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpjp        !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fk           !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkp          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  rcorr        !< Average reversibility correction for the whole anterpolation cell
      REAL(wp) ::  rcorr_ijk    !< Reversibility correction distributed to the individual child-grid nodes
 
! 
!--   Check which edge is to be handled
      IF ( edge == 'l' )  THEN
!
!--      For u, nxl is a ghost node, but not for the other variables
         IF ( var == 'u' )  THEN
            i     = nxl
            iend  = nxl
            ibc   = nxl
            iawbc = 0
            lbeg  = icl
            loff  = 0
         ELSE
            i     = nxl - igsr
            iend  = nxl - 1
            ibc   = nxl - 1
            iawbc = igsr-1
            lbeg  = icl
            loff  = 0
         ENDIF
      ELSEIF ( edge == 'r' )  THEN
         IF ( var == 'u' )  THEN
            i     = nxr + 1            
            iend  = nxr + 1
            ibc   = nxr + 1 
            iawbc = 0
            lbeg  = icr - 2
            loff  = 0
         ELSE
            i     = nxr + 1
            iend  = nxr + igsr
            ibc   = nxr + 1
            iawbc = 0
            lbeg  = icr - 2
            loff  = 1
         ENDIF

      ENDIF

      IF  ( var == 'w' )  THEN
         ntopw = 1
      ELSE
         ntopw = 0
      ENDIF

      IF  ( var == 'v' )  THEN
         mnorthv = 0
      ELSE
         mnorthv = 1
      ENDIF

      IF ( var == 'u' )  THEN 
         var_flag = 1 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2 
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF

!AH
!
!--   Substitute the necessary parent-grid data to the work array workarrc_lr. 
      workarrc_lr = 0.0_wp     
      IF  ( pdims(2) > 1 )  THEN
#if defined( __parallel )
         IF  ( nys == 0 )  THEN 
            workarrc_lr(0:cg%nz+1,jcsw:jcnw-1,0:2)                              &
                 = fc(0:cg%nz+1,jcsw:jcnw-1,lbeg:lbeg+2)
         ELSE IF  ( nyn == ny )  THEN
            workarrc_lr(0:cg%nz+1,jcsw+1:jcnw,0:2)                              &
                 = fc(0:cg%nz+1,jcsw+1:jcnw,lbeg:lbeg+2)
         ELSE
            workarrc_lr(0:cg%nz+1,jcsw+1:jcnw-1,0:2)                            &
                 = fc(0:cg%nz+1,jcsw+1:jcnw-1,lbeg:lbeg+2)
         ENDIF
!
!--      South-north exchange if more than one PE subdomain in the y-direction.
!--      Note that in case of 3-D nesting the south (psouth == MPI_PROC_NULL) 
!--      and north (pnorth == MPI_PROC_NULL) boundaries are not exchanged 
!--      because the nest domain is not cyclic.
!--      From south to north
         CALL MPI_SENDRECV( workarrc_lr(0,jcsw+1,0), 1,                         &
              workarrc_lr_exchange_type, psouth,  0,                            &
              workarrc_lr(0,jcnw,0), 1,                                         &
              workarrc_lr_exchange_type, pnorth,  0,                            &
              comm2d, status, ierr ) 
!
!--      From north to south        
         CALL MPI_SENDRECV( workarrc_lr(0,jcnw-1,0), 1,                         &
              workarrc_lr_exchange_type, pnorth,  1,                            &
              workarrc_lr(0,jcsw,0), 1,                                         &
              workarrc_lr_exchange_type, psouth,  1,                            &
              comm2d, status, ierr )
#endif
      ELSE
         workarrc_lr(0:cg%nz+1,jcsw:jcnw,0:2)                                   &
              = fc(0:cg%nz+1,jcsw:jcnw,lbeg:lbeg+2)            
      ENDIF
!
!AH
      workarr_lr = 0.0_wp
      
      DO  ia = i, iend
         iaw = ia - i
         DO  j = nys-1, nyn+1
!AH         DO  j = nys, nyn
!
!--         Determine vertical index of topography top at grid point (j,i)
!AH         k_wall = get_topography_top_index_ji( j, i, TRIM( var ) )
            DO  k = nzb, nzt+1 !k_wall, nzt+1
!AH               l   = ic(ia) - ic(i)
               l   = ic(ia) - lbeg
               lp1 = MIN( l + 1, 2 )  ! If l+1 > 2 (l=ic(nxr+1)-lbeg), r1x = 1 and r2x = 0
               m   = jc(j)
               mp1 = MIN( m + 1, jcnw )  ! If m+1 > jcn (m=jc(nyn+1)), r1y = 1 and r2y = 0
               n   = kc(k)
               np1 = n + 1            

!AH
!               fkj      = r1x(i) * fc(n,m,l)     + r2x(i) * fc(n,m,l+1)
!               fkjp     = r1x(i) * fc(n,m+1,l)   + r2x(i) * fc(n,m+1,l+1)
!               fkpj     = r1x(i) * fc(n+1,m,l)   + r2x(i) * fc(n+1,m,l+1)
!               fkpjp    = r1x(i) * fc(n+1,m+1,l) + r2x(i) * fc(n+1,m+1,l+1)
!AH
               
               fkj      = r1x(ia) * workarrc_lr(n,m,l)     + r2x(ia) * workarrc_lr(n,m,lp1)
               fkjp     = r1x(ia) * workarrc_lr(n,mp1,l)   + r2x(ia) * workarrc_lr(n,mp1,lp1)
               fkpj     = r1x(ia) * workarrc_lr(np1,m,l)   + r2x(ia) * workarrc_lr(np1,m,lp1)
               fkpjp    = r1x(ia) * workarrc_lr(np1,mp1,l) + r2x(ia) * workarrc_lr(np1,mp1,lp1)
               
               fk       = r1y(j) * fkj  + r2y(j) * fkjp
               fkp      = r1y(j) * fkpj + r2y(j) * fkpjp

!AH
!               f(k,j,i) = r1z(k) * fk + r2z(k) * fkp               
               workarr_lr(k,j,iaw) = r1z(k) * fk + r2z(k) * fkp
               
!               if  ( ( edge == 'l' ) .and. ( ia  == ibc ) ) then
!                  write(9,"('pmci_interp_tril_lr: ',a2,2x,11(i4,2x),7(e12.5,2x))") var, k, j, ia, ibc, iaw, n, m, l, np1, mp1, lp1, &
!                       workarr_lr(k,j,iaw), r1x(ia), r2x(ia), workarrc_lr(n,m,l), workarrc_lr(n,mp1,l), workarrc_lr(np1,m,l), workarrc_lr(np1,mp1,l)
!                  flush(9)
!               endif

!AH
               
            ENDDO
         ENDDO
      ENDDO
!
!--   Generalized log-law-correction algorithm.
!--   Doubly two-dimensional index arrays logc(1:2,:,:) and log-ratio arrays 
!--   logc_ratio(1:2,0:ncorr-1,:,:) have been precomputed in subroutine
!--   pmci_init_loglaw_correction.
!
!--   Solid surface below the node 
      IF ( constant_flux_layer .AND. ( var == 'u' .OR. var == 'v' ) )  THEN
         DO  j = nys, nyn
!
!--         Determine vertical index of topography top at grid point (j,i)
            k_wall = get_topography_top_index_ji( j, i, TRIM ( var ) )

            k = k_wall+1
            IF ( ( logc(1,k,j) /= 0 )  .AND.  ( logc(2,k,j) == 0 ) )  THEN
               k1 = logc(1,k,j)
               DO  kcorr = 0, ncorr - 1
                  kco = k + kcorr
!AH                  f(kco,j,i) = logc_ratio(1,kcorr,k,j) * f(k1,j,i)
               ENDDO
            ENDIF
         ENDDO
      ENDIF
!
!--   In case of non-flat topography, also vertical walls and corners need to be
!--   treated. Only single and double wall nodes are corrected. Triple and
!--   higher-multiple wall nodes are not corrected as the log law would not be
!--   valid anyway in such locations.
      IF ( topography /= 'flat' )  THEN

         IF ( constant_flux_layer .AND. ( var == 'u' .OR. var == 'w' ) )  THEN           
!
!--         Solid surface only on south/north side of the node                   
            DO  j = nys, nyn
               DO  k = logc_kbounds(1,j), logc_kbounds(2,j)   
                  IF ( ( logc(2,k,j) /= 0 )  .AND.  ( logc(1,k,j) == 0 ) )  THEN
!
!--                  Direction of the wall-normal index is carried in as the
!--                  sign of logc
                     jinc = SIGN( 1, logc(2,k,j) )
                     j1   = ABS( logc(2,k,j) )
                     DO  jcorr = 0, ncorr-1
                        jco = j + jinc * jcorr
                        IF ( jco >= nys .AND. jco <= nyn )  THEN
!AH                           f(k,jco,i) = logc_ratio(2,jcorr,k,j) * f(k,j1,i)
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDDO
         ENDIF
!
!--      Solid surface on both below and on south/north side of the node           
         IF ( constant_flux_layer .AND. var == 'u' )  THEN
            DO  j = nys, nyn
               k = logc_kbounds(1,j)
               IF ( ( logc(2,k,j) /= 0 )  .AND.  ( logc(1,k,j) /= 0 ) )  THEN
                  k1   = logc(1,k,j)                 
                  jinc = SIGN( 1, logc(2,k,j) )
                  j1   = ABS( logc(2,k,j) )                 
                  DO  jcorr = 0, ncorr-1
                     jco = j + jinc * jcorr
                     IF ( jco >= nys .AND. jco <= nyn )  THEN
                        DO  kcorr = 0, ncorr-1
                           kco = k + kcorr
!AH                           f(kco,jco,i) = 0.5_wp * ( logc_ratio(1,kcorr,k,j) * &
!AH                                                     f(k1,j,i)                 &
!AH                                                   + logc_ratio(2,jcorr,k,j) * &
!AH                                                     f(k,j1,i) )
                        ENDDO
                     ENDIF
                  ENDDO
               ENDIF
            ENDDO
         ENDIF

      ENDIF  ! ( topography /= 'flat' )
!
!--   Apply the reversibility correction. 

!      if  ( var == 'u' )  then

      l  = ic(ibc) + loff
      lw = 1
!      write(9,"('pmci_interp_tril_lr: edge, var, l, i, iend, ifl(l), ifu(l) = ',2(a2,2x),5(i4,2x))") &
!           edge, var, l, i, iend, ifl(l), ifu(l)
!      flush(9)
!AH      DO  m = jcsw+1, jcnw-1       
      DO  m = jcsw + 1, jcnw - mnorthv   ! mnorthv = 0 for v and 1 for all others
         DO  n = 0, kct + ntopw            ! ntopw = 1 for w and 0 for all others
            ijk = 1
            cellsum   = 0.0_wp
            cellsumd  = 0.0_wp
!
!--         Note that the index name i must not be used here as a loop 
!--         index name since i is the constant boundary index, hence 
!--         the name ia. 
            DO  ia = ifl(l), ifu(l)
               iaw = ia - i
               iw  = MAX( MIN( ia, nx+1 ), -1 )
               DO  j = jfl(m), jfu(m)
                  jw = MAX( MIN( j, ny+1 ), -1 )
                  DO  k = kfl(n), kfu(n)
                     kw = MIN( k, nzt+1 )
!AH
!                     cellsum = cellsum + MERGE( f(k,j,ia), 0.0_wp,              &
!                          BTEST( wall_flags_0(kw,jw,iw), var_flag ) )
                     cellsum = cellsum + MERGE( workarr_lr(k,j,iaw), 0.0_wp,     &
                          BTEST( wall_flags_0(kw,jw,iw), var_flag ) )
!AH
                     
!AH                     celltmpd(ijk) = ABS( fc(n,m,l) - f(k,j,ia) )
!AH                     celltmpd(ijk) = ABS( workarrc_lr(n,m,lw) - f(k,j,ia) )
                     celltmpd(ijk) = ABS( workarrc_lr(n,m,lw) - workarr_lr(k,j,iaw) )
                     cellsumd      = cellsumd  + MERGE( celltmpd(ijk),          &
                          0.0_wp, BTEST( wall_flags_0(kw,jw,iw), var_flag ) )

!                     write(9,"('lr1: ',a1,2x,8(i4,2x),5(e12.5,2x))") var, n, m, l, k, j, ia, iaw, ijk,  &
!                          workarrc_lr(n,m,lw), workarr_lr(k,j,iaw), cellsum, celltmpd(ijk), cellsumd
!                     flush(9)

                     ijk = ijk + 1
                  ENDDO
               ENDDO
            ENDDO

            IF ( ijkfc(n,m,l) /= 0 )  THEN
               cellsum   = cellsum / REAL( ijkfc(n,m,l), KIND=wp )
!               rcorr     = fc(n,m,l) - cellsum
               rcorr     = workarrc_lr(n,m,lw) - cellsum
               cellsumd  = cellsumd / REAL( ijkfc(n,m,l), KIND=wp )
            ELSE
               cellsum   = 0.0_wp                  
               rcorr     = 0.0_wp
               cellsumd  = 1.0_wp
               celltmpd  = 1.0_wp
            ENDIF
! 
!--         Distribute the correction term to the child nodes according to 
!--         their relative difference to the parent value such that the
!--         node with the largest difference gets the largest share of the 
!--         correction. The distribution is skipped if rcorr is negligibly 
!--         small in order to avoid division by zero. 
            IF ( ABS(rcorr) < 0.000001_wp )  THEN                  
               cellsumd  = 1.0_wp
               celltmpd  = 1.0_wp
            ENDIF

            ijk = 1
            DO  ia = ifl(l), ifu(l)
!AH               iaw = ia - ifl(l)
               iaw = ia - i
               DO  j = jfl(m), jfu(m)
                  DO  k = kfl(n), kfu(n)
                     rcorr_ijk = rcorr * celltmpd(ijk) / cellsumd 
!AH                     f(k,j,ia) = f(k,j,ia) + rcorr_ijk
                     workarr_lr(k,j,iaw) = workarr_lr(k,j,iaw) + rcorr_ijk

!                     write(9,"('lr2: ',a1,2x,9(i4,2x),4(e12.5,2x))") var, n, m, l, k, j, ia, iaw, ijk, ijkfc(n,m,l), &
!                          rcorr, rcorr_ijk, workarr_lr(k,j,iaw), workarrc_lr(n,m,lw)
!                     flush(9)

                     ijk = ijk + 1
                  ENDDO
               ENDDO
            ENDDO
!
!--         Velocity components tangential to the boundary need special 
!--         treatment because their anterpolation cells are flat in their 
!--         direction and hence do not cover all the child-grid boundary nodes. 
!--         These components are next linearly interpolated to those child-grid 
!--         boundary nodes which are not covered by the anterpolation cells.
            IF  ( ( var == 'w' ) .AND. ( n > 0 ) )  THEN
               DO  k = kfu(n-1)+1, kfl(n)-1
                  IF  ( k <= nzt )  THEN
                     DO  j = jfl(m), jfu(m)
                        IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
!AH                              
!                           f(k,j,ia) = r1z(k) * f(kfu(n-1),j,ia)                &
!                                   + r2z(k) * f(kfl(n),j,ia)
                           workarr_lr(k,j,iawbc) =                              &
                                  r1z(k) * workarr_lr(kfu(n-1),j,iawbc)         &
                                + r2z(k) * workarr_lr(kfl(n),j,iawbc)

!                           write(9,"('lr3: ',a1,2x,7(i4,2x),3(e12.5,2x))") var, n, m, l, k, j, iawbc, ibc, &
!                                workarr_lr(k-1,j,iawbc), workarr_lr(k,j,iawbc), workarr_lr(k+1,j,iawbc)
!                           flush(9)

!AH                              
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDIF

            IF  ( ( var == 'v' ) .AND. ( m > jcsw ) )  THEN
               DO  j = jfu(m-1)+1, jfl(m)-1
                  IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
                     DO  k = kfl(n), kfu(n)
                        IF  ( k <= nzt+1 )  THEN                              
!AH                              
!                           f(k,j,ia) = r1y(j) * f(k,jfu(m-1),ia)                &
!                                   + r2y(j) * f(k,jfl(m),ia)
                           workarr_lr(k,j,iawbc) =                              &
                                  r1y(j) * workarr_lr(k,jfu(m-1),iawbc)         &
                                + r2y(j) * workarr_lr(k,jfl(m),iawbc)

!                           write(9,"('lr4: ',a1,2x,7(i4,2x),3(e12.5,2x))") var, n, m, l, k, j, iawbc, ibc, &
!                                workarr_lr(k,j-1,iawbc), workarr_lr(k,j,iawbc), workarr_lr(k,j+1,iawbc)
!                           flush(9)

!AH                              
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDIF
               
         ENDDO  ! n
      ENDDO  ! m

!      endif  ! var

!
!--   Finally substitute the boundary values.
      f(nzb:nzt+1,nys:nyn,ibc) = workarr_lr(nzb:nzt+1,nys:nyn,iawbc)

!      do  k = 0, 2
!         do  j = nys, nyn 
!            if  ( edge == 'l' )  then
!               write(9,"('lr5: ',2(a2,2x),4(i4,2x),4(e12.5,2x))") edge, var, k, j, ibc, iawbc,  &
!                    workarr_lr(k,j,iawbc), f(k,j,ibc), f(k,j,ibc+1), f(k,j,ibc+2)
!            else if  ( edge == 'r' )  then
!               write(9,"('lr5: ',2(a2,2x),4(i4,2x),4(e12.5,2x))") edge, var, k, j, ibc, iawbc,  &
!                    f(k,j,ibc-2), f(k,j,ibc-1), f(k,j,ibc), workarr_lr(k,j,iawbc)
!            endif
!         enddo
!      enddo
!      flush(9)

   END SUBROUTINE pmci_interp_tril_lr



   SUBROUTINE pmci_interp_tril_sn( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y, r1z, &
                                   r2z, logc, logc_ratio, logc_kbounds,        &
                                   nzt_topo_nestbc,                            &
                                   kct, ifl, ifu, jfl, jfu, kfl, kfu, ijkfc,   &
                                   edge, var )

!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the south and north boundaries. 
!--   This subroutine is based on trilinear interpolation.

      IMPLICIT NONE

      INTEGER(iwp) ::  nzt_topo_nestbc   !<

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f             !<
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc            !<
      REAL(wp), DIMENSION(1:2,0:ncorr-1,nzb:nzt_topo_nestbc,nxl:nxr),          &
                                      INTENT(IN)    ::  logc_ratio    !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x           !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x           !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y           !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y           !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z           !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z           !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z           !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z           !<

      
      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)         ::  ic    !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)         ::  jc    !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)           ::  kc    !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN)        ::  kc    !<
      INTEGER(iwp), DIMENSION(1:2,nzb:nzt_topo_nestbc,nxl:nxr),                &
                                          INTENT(IN)           ::  logc  !<
      INTEGER(iwp), DIMENSION(1:2,nxl:nxr), INTENT(IN)         ::  logc_kbounds  !< 

      INTEGER(iwp) :: kct
!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH

!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) ::  edge   !< Edge symbol: 'l', 'r', 's' or 'n' 
      CHARACTER(LEN=1), INTENT(IN) ::  var    !< Variable symbol: 'u', 'v', 'w' or 's'
      
      INTEGER(iwp) ::  i        !< Running index in the x-direction
      INTEGER(iwp) ::  iinc     !<
      INTEGER(iwp) ::  icorr    !<
      INTEGER(iwp) ::  ico      !<
      INTEGER(iwp) ::  ierr     !< MPI error code
      INTEGER(iwp) ::  ijk      !< Running index for all child-grid cells within the anterpolation cell
      INTEGER(iwp) ::  iw       !< i-index for wall_flags_0
      INTEGER(iwp) ::  i1       !<
      INTEGER(iwp) ::  j        !< Lower bound of the running index ja
      INTEGER(iwp) ::  ja       !< Index in y-direction running over the parent-grid cell on the boundary 
      INTEGER(iwp) ::  jaw      !< Reduced ja-index for workarr_sn
      INTEGER(iwp) ::  jawbc    !< jaw-index pointing to the boundary-value nodes (either 0 or jgsr-1)
!AH      INTEGER(iwp) ::  jbeg     !< j-index pointing to the starting point of workarr_sn in the j-direction
      INTEGER(iwp) ::  jbc      !< Fixed j-index pointing to the boundary-value nodes (either j or jend)
      INTEGER(iwp) ::  jend     !< Upper bound of the running index ja
      INTEGER(iwp) ::  jw       !< j-index for wall_flags_0
      INTEGER(iwp) ::  k        !< Running index in the z-direction
      INTEGER(iwp) ::  k_wall   !< Vertical index of topography top
      INTEGER(iwp) ::  kcorr    !<
      INTEGER(iwp) ::  kco      !<
      INTEGER(iwp) ::  kw       !< k-index for wall_flags_0
      INTEGER(iwp) ::  k1       !<
      INTEGER(iwp) ::  l        !< Parent-grid running index in the x-direction
      INTEGER(iwp) ::  lp1      !< l+1
      INTEGER(iwp) ::  lrightu  !< Upshift by one for the upper bound of index l in case of var == 'u'
      INTEGER(iwp) ::  m        !< Parent-grid running index in the y-direction
      INTEGER(iwp) ::  mbeg     !< m-index pointing to the starting point of workarrc_sn in the m-direction 
      INTEGER(iwp) ::  moff     !< m-offset needed on the north boundary to correctly refer to boundary ghost points
      INTEGER(iwp) ::  mp1      !< m+1
      INTEGER(iwp) ::  mw       !< Reduced m-index for workarrc_sn
      INTEGER(iwp) ::  n        !< Parent-grid running index in the z-direction
      INTEGER(iwp) ::  np1      !< n+1
      INTEGER(iwp) ::  ntopw    !< Upshift by one for the upper bound of index n in case of var == 'w'
      INTEGER(iwp) ::  var_flag !< Variable flag for BTEST( wall_flags_0 )

      REAL(wp) ::  cellsum      !< Sum of child-grid node values over the anterpolation cell 
      REAL(wp) ::  cellsumd     !< Sum of differences over the anterpolation cell 
      REAL(wp) ::  fk           !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkj          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkjp         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpj         !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkpjp        !< Intermediate result in trilinear interpolation
      REAL(wp) ::  fkp          !< Intermediate result in trilinear interpolation
      REAL(wp) ::  rcorr        !< Average reversibility correction for the whole anterpolation cell
      REAL(wp) ::  rcorr_ijk    !< Reversibility correction distributed to the individual child-grid nodes

!
!--   Check which edge is to be handled: south or north
      IF ( edge == 's' )  THEN
!
!--      For v, nys is a ghost node, but not for the other variables
         IF ( var == 'v' )  THEN
            j     = nys
            jend  = nys 
            jawbc = 0
            jbc   = nys       
            mbeg  = jcs
            moff  = 0
         ELSE
            j     = nys - jgsr
            jend  = nys - 1
            jawbc = jgsr - 1
            jbc   = nys - 1
            mbeg  = jcs
            moff  = 0
         ENDIF
      ELSEIF ( edge == 'n' )  THEN
         IF ( var == 'v' )  THEN
            j     = nyn + 1
            jend  = nyn + 1
            jawbc = 0
            jbc   = nyn + 1
            mbeg  = jcn - 2
            moff  = 0
         ELSE
            j     = nyn + 1
            jend  = nyn + jgsr
            jawbc = 0
            jbc   = nyn + 1
            mbeg  = jcn - 2
            moff  = 1
         ENDIF
      ENDIF

      IF  ( var == 'w' )  THEN
         ntopw = 1
      ELSE
         ntopw = 0
      ENDIF

      IF  ( var == 'u' )  THEN
         lrightu = 0
      ELSE
         lrightu = 1
      ENDIF

      IF ( var == 'u' )  THEN 
         var_flag = 1 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2 
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF
!AH
!
!--   Substitute the necessary parent-grid data to the work array workarrc_sn. 
      workarrc_sn = 0.0_wp     
      IF  ( pdims(1) > 1 )  THEN
#if defined( __parallel )
         IF  ( nxl == 0 )  THEN   ! if ( bc_dirichlet_l )
            workarrc_sn(0:cg%nz+1,0:2,iclw:icrw-1)                              &
                 = fc(0:cg%nz+1,mbeg:mbeg+2,iclw:icrw-1)
         ELSE IF  ( nxr == nx )  THEN    ! if ( bc_dirichlet_r )
            workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw)                              &
                 = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw)
         ELSE
            workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw-1)                            &
                 = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw-1)
         ENDIF
!
!--      Left-right exchange if more than one PE subdomain in the x-direction.
!--      Note that in case of 3-D nesting the left (pleft == MPI_PROC_NULL) and 
!--      right (pright == MPI_PROC_NULL) boundaries are not exchanged because 
!--      the nest domain is not cyclic.
!--      From left to right
         CALL MPI_SENDRECV( workarrc_sn(0,0,iclw+1), 1,                         &
              workarrc_sn_exchange_type, pleft,   0,                            &
              workarrc_sn(0,0,icrw), 1,                                         &
              workarrc_sn_exchange_type, pright,  0,                            &
              comm2d, status, ierr ) 
!
!--      From right to left        
         CALL MPI_SENDRECV( workarrc_sn(0,0,icrw-1), 1,                         &
              workarrc_sn_exchange_type, pright,  1,                            &
              workarrc_sn(0,0,iclw), 1,                                         &
              workarrc_sn_exchange_type, pleft,   1,                            &
              comm2d, status, ierr )
#endif      
      ELSE
         workarrc_sn(0:cg%nz+1,0:2,iclw+1:icrw-1)                               &
                 = fc(0:cg%nz+1,mbeg:mbeg+2,iclw+1:icrw-1)
      ENDIF
!
!AH

      workarr_sn = 0.0_wp
      
      DO  i = nxl-1, nxr+1
!AH      DO  i = nxl, nxr
         DO  ja = j, jend
            jaw = ja - j
            DO  k = nzb, nzt+1
               l   = ic(i)
               lp1 = MIN( l + 1, icrw )  ! If l+1 > icr (l=ic(nxr+1)), r1x = 1 and r2x = 0
!AH               m   = jc(ja) - jc(j)
               m   = jc(ja) - mbeg
               mp1 = MIN( m + 1, 2 )  ! If m+1 > 2 (m=jc(nyn+1)-mbeg), r1y = 1 and r2y = 0
               n   = kc(k)
               np1 = n + 1
!AH
!               fkj      = r1x(i) * fc(n,m,l)     + r2x(i) * fc(n,m,l+1)
!               fkjp     = r1x(i) * fc(n,m+1,l)   + r2x(i) * fc(n,m+1,l+1)
!               fkpj     = r1x(i) * fc(n+1,m,l)   + r2x(i) * fc(n+1,m,l+1)
!               fkpjp    = r1x(i) * fc(n+1,m+1,l) + r2x(i) * fc(n+1,m+1,l+1)
!AH
               fkj      = r1x(i) * workarrc_sn(n,m,l)     + r2x(i) * workarrc_sn(n,m,lp1)
               fkjp     = r1x(i) * workarrc_sn(n,mp1,l)   + r2x(i) * workarrc_sn(n,mp1,lp1)
               fkpj     = r1x(i) * workarrc_sn(np1,m,l)   + r2x(i) * workarrc_sn(np1,m,lp1)
               fkpjp    = r1x(i) * workarrc_sn(np1,mp1,l) + r2x(i) * workarrc_sn(np1,mp1,lp1)

               fk       = r1y(ja) * fkj  + r2y(ja) * fkjp
               fkp      = r1y(ja) * fkpj + r2y(ja) * fkpjp
!AH
!               f(k,j,i) = r1z(k) * fk   + r2z(k) * fkp
               workarr_sn(k,jaw,i) = r1z(k) * fk   + r2z(k) * fkp

!               if  ( ( edge == 's' ) .and. ( ja  == jbc ) ) then
!                  write(9,"('pmci_interp_tril_sn: ',a2,2x,11(i4,2x),1(e12.5,2x))") var, k, ja, i, jbc, jaw, n, m, l, np1, mp1, lp1, &
!                       workarr_sn(k,jaw,i)
!                  flush(9)
!               endif

!AH               
            ENDDO
         ENDDO
      ENDDO      
!
!--   Generalized log-law-correction algorithm.
!--   Multiply two-dimensional index arrays logc(1:2,:,:) and log-ratio arrays 
!--   logc_ratio(1:2,0:ncorr-1,:,:) have been precomputed in subroutine
!--   pmci_init_loglaw_correction.
!
!--   Solid surface below the node 
      IF ( constant_flux_layer .AND. ( var == 'u'  .OR.  var == 'v' ) )  THEN           
         DO  i = nxl, nxr
!
!--         Determine vertical index of topography top at grid point (j,i)
            k_wall = get_topography_top_index_ji( j, i, TRIM( var ) )

            k = k_wall + 1
            IF ( ( logc(1,k,i) /= 0 )  .AND.  ( logc(2,k,i) == 0 ) )  THEN
               k1 = logc(1,k,i)
               DO  kcorr = 0, ncorr-1
                  kco = k + kcorr
!AH                  f(kco,j,i) = logc_ratio(1,kcorr,k,i) * f(k1,j,i)
               ENDDO
            ENDIF
         ENDDO
      ENDIF
!
!--   In case of non-flat topography, also vertical walls and corners need to be
!--   treated. Only single and double wall nodes are corrected.
!--   Triple and higher-multiple wall nodes are not corrected as it would be
!--   extremely complicated and the log law would not be valid anyway in such
!--   locations.
      IF ( topography /= 'flat' )  THEN

         IF ( constant_flux_layer .AND. ( var == 'v' .OR. var == 'w' ) )  THEN
            DO  i = nxl, nxr
               DO  k = logc_kbounds(1,i), logc_kbounds(2,i)
!
!--               Solid surface only on left/right side of the node           
                  IF ( ( logc(2,k,i) /= 0 )  .AND.  ( logc(1,k,i) == 0 ) )  THEN
!
!--                  Direction of the wall-normal index is carried in as the
!--                  sign of logc
                     iinc = SIGN( 1, logc(2,k,i) )
                     i1  = ABS( logc(2,k,i) )
                     DO  icorr = 0, ncorr-1
                        ico = i + iinc * icorr
                        IF ( ico >= nxl .AND. ico <= nxr )  THEN
!AH                           f(k,j,ico) = logc_ratio(2,icorr,k,i) * f(k,j,i1)
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDDO
         ENDIF
!
!--      Solid surface on both below and on left/right side of the node           
         IF ( constant_flux_layer .AND. var == 'v' )  THEN
            DO  i = nxl, nxr
               k = logc_kbounds(1,i)
               IF ( ( logc(2,k,i) /= 0 )  .AND.  ( logc(1,k,i) /= 0 ) )  THEN
                  k1   = logc(1,k,i)         
                  iinc = SIGN( 1, logc(2,k,i) )
                  i1   = ABS( logc(2,k,i) )
                  DO  icorr = 0, ncorr-1
                     ico = i + iinc * icorr
                     IF ( ico >= nxl .AND. ico <= nxr )  THEN
                        DO  kcorr = 0, ncorr-1
                           kco = k + kcorr
!AH                           f(kco,j,ico) = 0.5_wp * ( logc_ratio(1,kcorr,k,i) * &
!AH                                                     f(k1,j,i)                 &
!AH                                                   + logc_ratio(2,icorr,k,i) * &
!AH                                                     f(k,j,i1) )
                        ENDDO
                     ENDIF
                  ENDDO
               ENDIF
            ENDDO
         ENDIF
         
      ENDIF  ! ( topography /= 'flat' )
!
!--   Apply the reversibility correction. 
      
!      if  ( var == 'u' )  then

      m  = jc(jbc) + moff
      mw = 1
!      write(9,"('pmci_interp_tril_sn: edge, var, m, j, jend, jfl(m), jfu(m) = ',2(a2,2x),5(i4,2x))") &
!           edge, var, m, j, jend, jfl(m), jfu(m)
!      flush(9)
!AH      DO  l = iclw+1, icrw-1
      DO  l = iclw + 1, icrw - lrightu   ! lrightu = 0 for u and 1 for all others
         DO  n = 0, kct + ntopw            ! ntopw = 1 for w and 0 for all others
            ijk = 1
            cellsum   = 0.0_wp 
            cellsumd  = 0.0_wp
            DO  i = ifl(l), ifu(l)
               iw = MAX( MIN( i, nx+1 ), -1 )
               DO  ja = jfl(m), jfu(m)
                  jaw = ja - j
                  jw  = MAX( MIN( ja, ny+1 ), -1 )
                  DO  k = kfl(n), kfu(n)
                     kw = MIN( k, nzt+1 )
!AH                     cellsum = cellsum + MERGE( f(k,ja,i), 0.0_wp,              &
!AH                          BTEST( wall_flags_0(kw,jw,iw), var_flag ) )
                     cellsum = cellsum + MERGE( workarr_sn(k,jaw,i), 0.0_wp,       &
                          BTEST( wall_flags_0(kw,jw,iw), var_flag ) )                    
!AH                     celltmpd(ijk) = ABS( fc(n,m,l) - f(k,ja,i) )
!AH                     celltmpd(ijk) = ABS( workarrc_sn(n,mw,l) - f(k,ja,i) )
                     celltmpd(ijk) = ABS( workarrc_sn(n,mw,l) - workarr_sn(k,jaw,i) )
                     cellsumd      = cellsumd  + MERGE( celltmpd(ijk),          &
                          0.0_wp, BTEST( wall_flags_0(kw,jw,iw), var_flag ) )

!                     write(9,"('sn1: ',a1,2x,8(i4,2x),5(e12.5,2x))") var, n, m, l, k, ja, i, jaw, ijk,  &
!                          workarrc_sn(n,mw,l), workarr_sn(k,jaw,i), cellsum, celltmpd(ijk), cellsumd
!                     flush(9)

                     ijk = ijk + 1
                  ENDDO
               ENDDO
            ENDDO

            IF ( ijkfc(n,m,l) /= 0 )  THEN
               cellsum   = cellsum / REAL( ijkfc(n,m,l), KIND=wp )
!AH               rcorr     = fc(n,m,l) - cellsum
               rcorr     = workarrc_sn(n,mw,l) - cellsum
               cellsumd  = cellsumd / REAL( ijkfc(n,m,l), KIND=wp )
            ELSE
               cellsum   = 0.0_wp                  
               rcorr     = 0.0_wp
               cellsumd  = 1.0_wp
               celltmpd  = 1.0_wp
            ENDIF
! 
!--         Distribute the correction term to the child nodes according to 
!--         their relative difference to the parent value such that the
!--         node with the largest difference gets the largest share of the 
!--         correction. The distribution is skipped if rcorr is negligibly 
!--         small in order to avoid division by zero. 
            IF ( ABS(rcorr) < 0.000001_wp )  THEN                  
               cellsumd  = 1.0_wp
               celltmpd  = 1.0_wp
            ENDIF
            
            ijk = 1           
            DO  i = ifl(l), ifu(l)
               DO  ja = jfl(m), jfu(m)
!AH                  jaw = ja - jfl(m)
                  jaw = ja - j
                  DO  k = kfl(n), kfu(n)
                     rcorr_ijk = rcorr * celltmpd(ijk) / cellsumd 
!AH                     f(k,ja,i) = f(k,ja,i) + rcorr_ijk
                     workarr_sn(k,jaw,i) = workarr_sn(k,jaw,i) + rcorr_ijk

!                     write(9,"('sn2: ',a1,2x,9(i4,2x),4(e12.5,2x))") var, n, m, l, k, ja, i, jaw, ijk, ijkfc(n,m,l), &
!                          rcorr, rcorr_ijk, workarr_sn(k,jaw,i), workarrc_sn(n,mw,l)
!                     flush(9)

                     ijk = ijk + 1
                  ENDDO
               ENDDO
            ENDDO
!
!--         Velocity components tangential to the boundary need special 
!--         treatment because their anterpolation cells are flat in their 
!--         direction and hence do not cover all the child-grid boundary nodes. 
!--         These components are next linearly interpolated to those child-grid 
!--         boundary nodes which are not covered by the anterpolation cells.     
            IF  ( ( var == 'w' ) .AND. ( n > 0 ) )  THEN
               DO  k = kfu(n-1)+1, kfl(n)-1
                  IF  ( k <= nzt )  THEN
                     DO  i = ifl(l), ifu(l)
                        IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
!AH                         f(k,ja,i) = r1z(k) * f(kfu(n-1),ja,i)                & 
!AH                               + r2z(k) * f(kfl(n),ja,i)
                           workarr_sn(k,jawbc,i) =                              &
                                r1z(k) * workarr_sn(kfu(n-1),jawbc,i) +         & 
                                r2z(k) * workarr_sn(kfl(n),jawbc,i)
   
!                           write(9,"('sn3: ',a1,2x,7(i4,2x),3(e12.5,2x))") var, n, m, l, k, jawbc, i, jbc,  &
!                                workarr_sn(k-1,jawbc,i), workarr_sn(k,jawbc,i), workarr_sn(k+1,jawbc,i)
!                           flush(9)

                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDIF
               
            IF  ( ( var == 'u' ) .AND. ( l > iclw ) )  THEN
               DO  i = ifu(l-1)+1, ifl(l)-1
                  IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
                     DO  k = kfl(n), kfu(n)
                        IF  ( k <= nzt+1 )  THEN
!AH                         f(k,ja,i) = r1x(i) * f(k,ja,ifu(l-1))                &
!AH                               + r2x(i) * f(k,ja,ifl(l))
                           workarr_sn(k,jawbc,i) =                              &
                                r1x(i) * workarr_sn(k,jawbc,ifu(l-1)) +         &
                                r2x(i) * workarr_sn(k,jawbc,ifl(l))
                           
!                           write(9,"('sn4: ',a1,2x,7(i4,2x),3(e12.5,2x))") var, n, m, l, k, jawbc, i, jbc, &
!                                workarr_sn(k,jawbc,i-1), workarr_sn(k,jawbc,i), workarr_sn(k,jawbc,i+1)
!                           flush(9)

                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
            ENDIF
               
         ENDDO  ! n
      ENDDO  ! l 

!      endif  ! var

!
!--   Finally substitute the boundary values.
      f(nzb:nzt+1,jbc,nxl:nxr) = workarr_sn(nzb:nzt+1,jawbc,nxl:nxr)

!      do  k = 0, 2
!         do  i = nxl, nxr 
!            if  ( edge == 's' )  then
!               write(9,"('sn5: ',2(a2,2x),4(i4,2x),4(e12.5,2x))") edge, var, k, i, jbc, jawbc,  &
!                    workarr_sn(k,jawbc,i), f(k,jbc,i), f(k,jbc+1,i), f(k,jbc+2,i)
!            else if  ( edge == 'n' )  then
!               write(9,"('sn5: ',2(a2,2x),4(i4,2x),4(e12.5,2x))") edge, var, k, i, jbc, jawbc,  &
!                    f(k,jbc-2,i), f(k,jbc-1,i), f(k,jbc,i), workarr_sn(k,jawbc,i)
!            endif
!         enddo
!      enddo
!      flush(9)

   END SUBROUTINE pmci_interp_tril_sn

 

   SUBROUTINE pmci_interp_tril_t( f, fc, ic, jc, kc, r1x, r2x, r1y, r2y,       &
                                  r1z, r2z, kct, ifl, ifu, jfl, jfu, kfl, kfu, &
                                  ijkfc, var )

!
!--   Interpolation of ghost-node values used as the child-domain boundary
!--   conditions. This subroutine handles the top boundary. 
!--   This subroutine is based on trilinear interpolation.

      IMPLICIT NONE

      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                      &
                                      INTENT(INOUT) ::  f     !< Child-grid array
      REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr),                          &
                                      INTENT(IN)    ::  fc    !< Parent-grid array
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r1x   !<
      REAL(wp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  r2x   !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r1y   !<
      REAL(wp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  r2y   !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r1z   !<
!AH      REAL(wp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  r2z   !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r1z   !<
      REAL(wp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  r2z   !<

      
      INTEGER(iwp), DIMENSION(nxlfc:nxrfc), INTENT(IN)  ::  ic    !<
      INTEGER(iwp), DIMENSION(nysfc:nynfc), INTENT(IN)  ::  jc    !<
!AH      INTEGER(iwp), DIMENSION(nzb:nzt+1), INTENT(IN)    ::  kc    !<
      INTEGER(iwp), DIMENSION(nzb:nzt+kgsr), INTENT(IN) ::  kc    !<

      INTEGER(iwp) :: kct
!AH
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!      INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
      INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
      INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH      INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
      INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

      CHARACTER(LEN=1), INTENT(IN) :: var   !<

      INTEGER(iwp) ::  i    !<
      INTEGER(iwp) ::  ib   !<
      INTEGER(iwp) ::  iclc !< Lower i-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  icrc !< Upper i-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  ie   !<
      INTEGER(iwp) ::  ierr !< MPI error code
      INTEGER(iwp) ::  ijk  !<
      INTEGER(iwp) ::  iw   !<
      INTEGER(iwp) ::  j    !<
      INTEGER(iwp) ::  jb   !<
      INTEGER(iwp) ::  jcsc !< Lower j-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  jcnc !< Upper j-index limit for copying fc-data to workarrc_t
      INTEGER(iwp) ::  je   !< 
      INTEGER(iwp) ::  jw   !<     
      INTEGER(iwp) ::  k    !< Vertical child-grid index fixed to the boundary-value level 
      INTEGER(iwp) ::  ka   !< Running vertical child-grid index
      INTEGER(iwp) ::  kw   !< 
      INTEGER(iwp) ::  l    !< Parent-grid index in x-direction
      INTEGER(iwp) ::  lp1  !< l+1
      INTEGER(iwp) ::  m    !< Parent-grid index in y-direction
      INTEGER(iwp) ::  mp1  !< m+1
      INTEGER(iwp) ::  n    !< Parent-grid work array index in z-direction
      INTEGER(iwp) ::  np1  !< n+1
      INTEGER(iwp) ::  noff !< n-offset needed on the top boundary to correctly refer to boundary ghost points  
      INTEGER(iwp) ::  nw   !< n-index for workarrc_t
      INTEGER(iwp) ::  var_flag  !<
      
      REAL(wp) ::  cellsum     !<
      REAL(wp) ::  cellsumd    !<
      REAL(wp) ::  fk          !<
      REAL(wp) ::  fkj         !<
      REAL(wp) ::  fkjp        !<
      REAL(wp) ::  fkpj        !<
      REAL(wp) ::  fkpjp       !<
      REAL(wp) ::  fkp         !<
      REAL(wp) ::  rcorr       !<
      REAL(wp) ::  rcorr_ijk   !<

      integer(iwp) :: kend


      IF ( var == 'w' )  THEN
         k    = nzt
         noff = 0
         kend = nzt
      ELSE
         k    = nzt + 1
         noff = 1
         kend = nzt+kgsr
      ENDIF
!
!--   These exceedings by one are needed only to avoid stripes
!--   and spots in visualization. They have no effect on the 
!--   actual solution.     
!
!AH   These loop bounds lead to overflows with the current reduced icl, icr, jcs, jcn
!AH      ib = nxl-1
!AH      ie = nxr+1
!AH      jb = nys-1
!AH      je = nyn+1
      ib = nxl
      ie = nxr
      jb = nys
      je = nyn

      IF ( var == 'u' )  THEN 
         var_flag = 1 
         ie       = nxr + 1 ! Needed for finishing interpolation for u 
      ELSEIF ( var == 'v' )  THEN
         var_flag = 2      
         je       = nyn + 1 ! Needed for finishing interpolation for v 
      ELSEIF ( var == 'w' )  THEN
         var_flag = 3
      ELSE
         var_flag = 0
      ENDIF
         
!AH      DO  i = ib, ie
!AH         DO  j = jb, je
!AH            l = ic(i)
!AH            m = jc(j)
!AH            n = kc(k)            
!AH            fkj      = r1x(i) * fc(n,m,l)     + r2x(i) * fc(n,m,l+1)
!AH            fkjp     = r1x(i) * fc(n,m+1,l)   + r2x(i) * fc(n,m+1,l+1)
!AH            fkpj     = r1x(i) * fc(n+1,m,l)   + r2x(i) * fc(n+1,m,l+1)
!AH            fkpjp    = r1x(i) * fc(n+1,m+1,l) + r2x(i) * fc(n+1,m+1,l+1)
!AH            fk       = r1y(j) * fkj  + r2y(j) * fkjp
!AH            fkp      = r1y(j) * fkpj + r2y(j) * fkpjp
!AH            f(k,j,i) = r1z(k) * fk   + r2z(k) * fkp
!AH         ENDDO
!AH      ENDDO

!      write(9,"('workarrc_t kbounds: ',a2,2x,5(i3,2x))") var, k, kend, kc(k), kc(kend), cg%nz+1
!      flush(9)
!AH
!
!--   Substitute the necessary parent-grid data to the work array. 
!--   Note that the dimension of workarrc_t is (0:2,jcsw:jcnw,iclw:icrw),
!--   And the jc?w and ic?w-index bounds depend on the location of the PE-
!--   subdomain relative to the side boundaries. 
      iclc = iclw + 1
      icrc = icrw - 1      
      jcsc = jcsw + 1
      jcnc = jcnw - 1
      IF  ( bc_dirichlet_l )  THEN
         iclc = iclw
      ENDIF
      IF  ( bc_dirichlet_r )  THEN
         icrc = icrw
      ENDIF
      IF  ( bc_dirichlet_s )  THEN
         jcsc = jcsw
      ENDIF
      IF  ( bc_dirichlet_n )  THEN
         jcnc = jcnw
      ENDIF
      workarrc_t = 0.0_wp
      workarrc_t(0:2,jcsc:jcnc,iclc:icrc)                                       &
           = fc(kc(k):kc(k)+2,jcsc:jcnc,iclc:icrc)
!
!--   Left-right exchange if more than one PE subdomain in the x-direction.
!--   Note that in case of 3-D nesting the left and right boundaries are 
!--   not exchanged because the nest domain is not cyclic.
#if defined( __parallel )
      IF  ( pdims(1) > 1 )  THEN
!
!--      From left to right
         CALL MPI_SENDRECV( workarrc_t(0,jcsw,iclw+1), 1,                       &
              workarrc_t_exchange_type_y, pleft,  0,                            &
              workarrc_t(0,jcsw,icrw), 1,                                       &
              workarrc_t_exchange_type_y, pright, 0,                            &
              comm2d, status, ierr ) 
!
!--      From right to left        
         CALL MPI_SENDRECV( workarrc_t(0,jcsw,icrw-1), 1,                       &
              workarrc_t_exchange_type_y, pright, 1,                            &
              workarrc_t(0,jcsw,iclw), 1,                                       &
              workarrc_t_exchange_type_y, pleft,  1,                            &
              comm2d, status, ierr )
      ENDIF
!
!--   South-north exchange if more than one PE subdomain in the y-direction.
!--   Note that in case of 3-D nesting the south and north boundaries are 
!--   not exchanged because the nest domain is not cyclic.
      IF  ( pdims(2) > 1 )  THEN
!
!--      From south to north         
         CALL MPI_SENDRECV( workarrc_t(0,jcsw+1,iclw), 1,                       &
              workarrc_t_exchange_type_x, psouth, 2,                            &
              workarrc_t(0,jcnw,iclw), 1,                                       &
              workarrc_t_exchange_type_x, pnorth, 2,                            &
              comm2d, status, ierr ) 
!
!--      From north to south        
         CALL MPI_SENDRECV( workarrc_t(0,jcnw-1,iclw), 1,                       &
              workarrc_t_exchange_type_x, pnorth, 3,                            &
              workarrc_t(0,jcsw,iclw), 1,                                       &
              workarrc_t_exchange_type_x, psouth, 3,                            &
              comm2d, status, ierr )
      ENDIF
#endif      
!
!AH

      workarr_t = 0.0_wp

      DO  i = ib, ie
         DO  j = jb, je
            DO ka = k, kend
               l    = ic(i)
               lp1  = MIN( l + 1, icrw ) ! If l+1 > icr (l=ic(nxr+1)), r1x = 1 and r2x = 0
               m    = jc(j)
               mp1  = MIN( m + 1, jcnw ) ! If m+1 > jcn (m=jc(nyn+1)), r1y = 1 and r2y = 0
!AH
!               n    = kc(ka)            
               n   = kc(ka) - kc(k)
               np1 = n + 1
!AH
!               fkj       = r1x(i)  * fc(n,m,l)     + r2x(i) * fc(n,m,l+1)
!               fkjp      = r1x(i)  * fc(n,m+1,l)   + r2x(i) * fc(n,m+1,l+1)
!               fkpj      = r1x(i)  * fc(n+1,m,l)   + r2x(i) * fc(n+1,m,l+1)
!               fkpjp     = r1x(i)  * fc(n+1,m+1,l) + r2x(i) * fc(n+1,m+1,l+1)
!AH
               fkj       = r1x(i)  * workarrc_t(n,m,l)     + r2x(i) * workarrc_t(n,m,lp1)
               fkjp      = r1x(i)  * workarrc_t(n,mp1,l)   + r2x(i) * workarrc_t(n,mp1,lp1)
               fkpj      = r1x(i)  * workarrc_t(np1,m,l)   + r2x(i) * workarrc_t(np1,m,lp1)
               fkpjp     = r1x(i)  * workarrc_t(np1,mp1,l) + r2x(i) * workarrc_t(np1,mp1,lp1)
!AH
               fk        = r1y(j)  * fkj  + r2y(j) * fkjp
               fkp       = r1y(j)  * fkpj + r2y(j) * fkpjp
               workarr_t(ka,j,i) = r1z(ka) * fk + r2z(ka) * fkp
          
            ENDDO
         ENDDO
      ENDDO

!AH!
!AH!--   Just fill up the redundant second ghost-node layer for w.
!AH      IF ( var == 'w' )  THEN
!AH         f(nzt+1,:,:) = f(nzt,:,:)
!AH      ENDIF
!
!--      Apply the reversibility correction. 
         n  = kc(k) + noff  
         nw = 0
!AH         DO  l = icl-1, icr+1
         DO  l = iclw, icrw
!AH            DO  m = jcs-1, jcn+1
            DO  m = jcsw, jcnw
               ijk = 1
               cellsum   = 0.0_wp
               cellsumd  = 0.0_wp
               DO  i = ifl(l), ifu(l)
                  iw = MAX( MIN( i, nx+1 ), -1 )
                  IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
                     DO  j = jfl(m), jfu(m)
                        jw = MAX( MIN( j, ny+1 ), -1 ) 
                        IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
                           DO  ka = kfl(n), kfu(n)
                              kw = MIN( ka, nzt+1 )
!AH                            cellsum = cellsum + MERGE( f(ka,j,i), 0.0_wp,           &
                              cellsum = cellsum + MERGE( workarr_t(ka,j,i), 0.0_wp,   &
                                   BTEST( wall_flags_0(kw,jw,iw), var_flag ) )
!                              cellsum = cellsum + workarr_t(ka,j,i)
!AH                            celltmpd(ijk) = ABS( fc(n,m,l) - f(ka,j,i) )
!AH
!                              celltmpd(ijk) = ABS( fc(n,m,l) - workarr_t(ka,j,i) )
                              celltmpd(ijk) = ABS( workarrc_t(nw,m,l) - workarr_t(ka,j,i) )
!AH
                              cellsumd      = cellsumd  + MERGE( celltmpd(ijk),       &
                                   0.0_wp, BTEST( wall_flags_0(kw,jw,iw), var_flag ) ) 
                              ijk = ijk + 1
                           ENDDO
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO

               IF ( ijkfc(n,m,l) /= 0 )  THEN
                  cellsum   = cellsum / REAL( ijkfc(n,m,l), KIND=wp )
!AH
!                  rcorr     = fc(n,m,l) - cellsum
                  rcorr     = workarrc_t(nw,m,l) - cellsum
!AH                  
                  cellsumd  = cellsumd / REAL( ijkfc(n,m,l), KIND=wp )
               ELSE
                  cellsum   = 0.0_wp
                  rcorr     = 0.0_wp
                  cellsumd  = 1.0_wp
                  celltmpd  = 1.0_wp
               ENDIF
! 
!--            Distribute the correction term to the child nodes according to 
!--            their relative difference to the parent value such that the
!--            node with the largest difference gets the largest share of the 
!--            correction. The distribution is skipped if rcorr is negligibly 
!--            small in order to avoid division by zero. 
               IF ( ABS(rcorr) < 0.000001_wp )  THEN                  
                  cellsumd  = 1.0_wp
                  celltmpd  = 1.0_wp
               ENDIF
            
               ijk = 1
               DO  i = ifl(l), ifu(l)
                  IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
                     DO  j = jfl(m), jfu(m)
                        IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
                           DO  ka = kfl(n), kfu(n)
                              rcorr_ijk = rcorr * celltmpd(ijk) / cellsumd 
!AH                            f(ka,j,i) = f(ka,j,i) + rcorr_ijk
                              workarr_t(ka,j,i) = workarr_t(ka,j,i) + rcorr_ijk

!                                 if  ( i == 128 .and. var == 'u' )  then
!                                    write(9,"('t2: ', 8(i4,2x),3(e12.5,2x))") nw, m, l, ka, j, i, ifl(l), ifu(l), &
!                                         rcorr, rcorr_ijk, workarr_t(ka,j,i)
!                                    flush(9)
!                                 endif

                              ijk = ijk + 1
                           ENDDO
                        ENDIF
                     ENDDO
                  ENDIF
               ENDDO
!
!--            Velocity components tangential to the boundary need special 
!--            "finishing" because their anterpolation cells are flat in their 
!--            direction and hence do not cover all the child-grid boundary nodes. 
!--            These components are next linearly interpolated to those child-grid 
!--            boundary nodes which are not covered by the anterpolation cells. 
               IF  ( ( var == 'v' ) .AND. ( m > jcs ) )  THEN
                  DO  j = jfu(m-1)+1, jfl(m)-1
                     IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
                        DO  i = ifl(l), ifu(l)
                           IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
                              DO  ka = kfl(n), kfu(n)        ! This loop should be removed and fixed k-value be used instead
!AH                              f(ka,j,i) = r1y(j) * f(ka,jfu(m-1),i)             &
!AH                                   + r2y(j) * f(ka,jfl(m),i)
                                 workarr_t(ka,j,i) = r1y(j) * workarr_t(ka,jfu(m-1),i) &
                                      + r2y(j) * workarr_t(ka,jfl(m),i)
                              ENDDO
                           ENDIF
                        ENDDO
                     ENDIF
                  ENDDO
               ENDIF
               
               IF  ( ( var == 'u' ) .AND. ( l > icl ) )  THEN
                  DO  i = ifu(l-1)+1, ifl(l)-1
                     IF  ( ( i >= nxl ) .AND. ( i <= nxr+1 ) )  THEN
                        DO  j = jfl(m), jfu(m)
                           IF  ( ( j >= nys ) .AND. ( j <= nyn+1 ) )  THEN
                              DO  ka = kfl(n), kfu(n)        ! This loop should be removed and fixed k-value be used instead
!AH                              f(ka,j,i) = r1x(i) * f(ka,j,ifu(l-1)) +           &
!AH                                   r2x(i) * f(ka,j,ifl(l))
                                 workarr_t(ka,j,i) = r1x(i) * workarr_t(ka,j,ifu(l-1)) &
                                      + r2x(i) * workarr_t(ka,j,ifl(l))

!                                 if  ( i == 127 )  then
!                                    write(9,"('t4: ', 8(i4,2x),3(e12.5,2x))") nw, m, l, ka, j, ifu(l-1), i, ifl(l), &
!                                         workarr_t(ka,j,ifu(l-1)), workarr_t(ka,j,i), workarr_t(ka,j,ifl(l))
!                                    flush(9)
!                                 endif

                              ENDDO
                           ENDIF
                        ENDDO
                     ENDIF
                  ENDDO
               ENDIF
               
            ENDDO  ! m
         ENDDO  ! l
!
!--   Finally substitute the boundary values.
      f(k,nys:nyn,nxl:nxr) = workarr_t(k,nys:nyn,nxl:nxr)
      IF  ( var == 'w' )  THEN
         f(k+1,nys:nyn,nxl:nxr) = f(k,nys:nyn,nxl:nxr)
      ENDIF

    END SUBROUTINE pmci_interp_tril_t



    SUBROUTINE pmci_anterp_tophat( f, fc, kct, ifl, ifu, jfl, jfu, kfl, kfu,   &
                                   ijkfc, var )
!
!--    Anterpolation of internal-node values to be used as the parent-domain
!--    values. This subroutine is based on the first-order numerical
!--    integration of the fine-grid values contained within the coarse-grid
!--    cell.

       IMPLICIT NONE

       CHARACTER(LEN=*), INTENT(IN) ::  var   !< identifyer for treated variable

!AH       INTEGER(iwp), DIMENSION(0:kct,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH
!       INTEGER(iwp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
       INTEGER(iwp), DIMENSION(0:cg%nz+1,jcsa:jcna,icla:icra), INTENT(IN) :: ijkfc !< number of child grid points contributing to a parent grid box
!AH

       INTEGER(iwp) ::  i         !< Running index x-direction - fine-grid
       INTEGER(iwp) ::  iclant    !< Left boundary index for anterpolation along x
       INTEGER(iwp) ::  icrant    !< Right boundary index for anterpolation along x
       INTEGER(iwp) ::  ii        !< Running index x-direction - coarse grid
       INTEGER(iwp) ::  j         !< Running index y-direction - fine-grid
       INTEGER(iwp) ::  jcnant    !< North boundary index for anterpolation along y
       INTEGER(iwp) ::  jcsant    !< South boundary index for anterpolation along y
       INTEGER(iwp) ::  jj        !< Running index y-direction - coarse grid
       INTEGER(iwp) ::  k         !< Running index z-direction - fine-grid     
       INTEGER(iwp) ::  kcb = 0   !< Bottom boundary index for anterpolation along z
       INTEGER(iwp) ::  kctant    !< Top boundary index for anterpolation along z
       INTEGER(iwp) ::  kk        !< Running index z-direction - coarse grid
       INTEGER(iwp) ::  var_flag  !< bit number used to flag topography on respective grid

       INTEGER(iwp), INTENT(IN) ::  kct   !< Top boundary index for anterpolation along z

!AH
!       INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
!       INTEGER(iwp), DIMENSION(icl:icr), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
!       INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
!       INTEGER(iwp), DIMENSION(jcs:jcn), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
       INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifl !< Indicates start index of child cells belonging to certain parent cell - x direction
       INTEGER(iwp), DIMENSION(icla:icra), INTENT(IN) ::  ifu !< Indicates end index of child cells belonging to certain parent cell - x direction
       INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfl !< Indicates start index of child cells belonging to certain parent cell - y direction
       INTEGER(iwp), DIMENSION(jcsa:jcna), INTENT(IN) ::  jfu !< Indicates start index of child cells belonging to certain parent cell - y direction
!AH

!AH       INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
!AH       INTEGER(iwp), DIMENSION(0:kct), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction
       INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfl !< Indicates start index of child cells belonging to certain parent cell - z direction
       INTEGER(iwp), DIMENSION(0:cg%nz+1), INTENT(IN)   ::  kfu !< Indicates start index of child cells belonging to certain parent cell - z direction

       REAL(wp) ::  cellsum   !< sum of respective child cells belonging to parent cell 
       REAL(wp) ::  fra       !< relaxation faction

       REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg), INTENT(IN) ::  f   !< Treated variable - child domain
       REAL(wp), DIMENSION(0:cg%nz+1,jcs:jcn,icl:icr), INTENT(INOUT)  ::  fc  !< Treated variable - parent domain
  
!
!--    Define the index bounds iclant, icrant, jcsant and jcnant.
!--    Note that kcb is simply zero and kct enters here as a parameter and it is
!--    determined in pmci_init_anterp_tophat.
!--    Please note, grid points used also for interpolation (from parent to
!--    child) are excluded in anterpolation, e.g. anterpolation is only from 
!--    nzb:kct-1, as kct is used for interpolation.
       iclant = icl
       icrant = icr
       jcsant = jcs
       jcnant = jcn
       kctant = kct - 1

       kcb  = 0
       IF ( nesting_mode /= 'vertical' )  THEN
          IF ( bc_dirichlet_l )  THEN
             iclant = icl + 3
          ENDIF
          IF ( bc_dirichlet_r )  THEN
             icrant = icr - 3
          ENDIF

          IF ( bc_dirichlet_s )  THEN
             jcsant = jcs + 3
          ENDIF
          IF ( bc_dirichlet_n )  THEN
             jcnant = jcn - 3
          ENDIF
       ENDIF
!
!--    Set masking bit for topography flags 
       IF ( var == 'u' )  THEN 
          var_flag = 1 
       ELSEIF ( var == 'v' )  THEN
          var_flag = 2 
       ELSEIF ( var == 'w' )  THEN
          var_flag = 3
       ELSE
          var_flag = 0
       ENDIF  
!
!--    Note that ii, jj, and kk are coarse-grid indices and i,j, and k 
!--    are fine-grid indices.
       DO  ii = iclant, icrant
          DO  jj = jcsant, jcnant
!
!--          For simplicity anterpolate within buildings and under elevated
!--          terrain too
             DO  kk = kcb, kctant !kct - 1
                cellsum = 0.0_wp
                DO  i = ifl(ii), ifu(ii)
                   DO  j = jfl(jj), jfu(jj)
                      DO  k = kfl(kk), kfu(kk)
                         cellsum = cellsum + MERGE( f(k,j,i), 0.0_wp,          &
                                        BTEST( wall_flags_0(k,j,i), var_flag ) )
                      ENDDO
                   ENDDO
                ENDDO
!
!--             Spatial under-relaxation.
!--             The relaxation buffer zones are no longer needed with
!--             the new reversible interpolation algorithm. 23.10.2018.
!                fra  = frax(ii) * fray(jj) * fraz(kk)                
!
!--             In case all child grid points are inside topography, i.e. 
!--             ijkfc and cellsum are zero, also parent solution would have 
!--             zero values at that grid point, which may cause problems in 
!--             particular for the temperature. Therefore, in case cellsum is 
!--             zero, keep the parent solution at this point. 

                IF ( ijkfc(kk,jj,ii) /= 0 )  THEN
!AH                   fc(kk,jj,ii) = ( 1.0_wp - fra ) * fc(kk,jj,ii) +         &
!AH                        fra * cellsum  /                                    &
!AH                        REAL( ijkfc(kk,jj,ii), KIND=wp )
                   fc(kk,jj,ii) = cellsum / REAL( ijkfc(kk,jj,ii), KIND=wp )
                ENDIF

             ENDDO
          ENDDO
       ENDDO

    END SUBROUTINE pmci_anterp_tophat

#endif

 END SUBROUTINE pmci_child_datatrans

! Description:
! ------------
!> Set boundary conditions for the prognostic quantities after interpolation 
!> and anterpolation at upward- and downward facing surfaces.  
!> @todo: add Dirichlet boundary conditions for pot. temperature, humdidity and
!> passive scalar.
!------------------------------------------------------------------------------!
 SUBROUTINE pmci_boundary_conds

    USE chem_modules,                                                          &
        ONLY:  ibc_cs_b

    USE control_parameters,                                                    &
        ONLY:  ibc_pt_b, ibc_q_b, ibc_s_b, ibc_uv_b

    USE surface_mod,                                                           &
        ONLY:  bc_h

    IMPLICIT NONE

    INTEGER(iwp) ::  i  !< Index along x-direction
    INTEGER(iwp) ::  j  !< Index along y-direction
    INTEGER(iwp) ::  k  !< Index along z-direction
    INTEGER(iwp) ::  m  !< Running index for surface type
    INTEGER(iwp) ::  n  !< running index for number of chemical species
    
!
!-- Set Dirichlet boundary conditions for horizontal velocity components
    IF ( ibc_uv_b == 0 )  THEN
!
!--    Upward-facing surfaces
       DO  m = 1, bc_h(0)%ns
          i = bc_h(0)%i(m)            
          j = bc_h(0)%j(m)
          k = bc_h(0)%k(m)
          u(k-1,j,i) = 0.0_wp
          v(k-1,j,i) = 0.0_wp
       ENDDO
!
!--    Downward-facing surfaces
       DO  m = 1, bc_h(1)%ns
          i = bc_h(1)%i(m)            
          j = bc_h(1)%j(m)
          k = bc_h(1)%k(m)
          u(k+1,j,i) = 0.0_wp
          v(k+1,j,i) = 0.0_wp
       ENDDO
    ENDIF
!
!-- Set Dirichlet boundary conditions for vertical velocity component
!-- Upward-facing surfaces
    DO  m = 1, bc_h(0)%ns
       i = bc_h(0)%i(m)            
       j = bc_h(0)%j(m)
       k = bc_h(0)%k(m)
       w(k-1,j,i) = 0.0_wp
    ENDDO
!
!-- Downward-facing surfaces
    DO  m = 1, bc_h(1)%ns
       i = bc_h(1)%i(m)            
       j = bc_h(1)%j(m)
       k = bc_h(1)%k(m)
       w(k+1,j,i) = 0.0_wp
    ENDDO
!
!-- Set Neumann boundary conditions for potential temperature
    IF ( .NOT. neutral )  THEN
       IF ( ibc_pt_b == 1 )  THEN
          DO  m = 1, bc_h(0)%ns
             i = bc_h(0)%i(m)            
             j = bc_h(0)%j(m)
             k = bc_h(0)%k(m)
             pt(k-1,j,i) = pt(k,j,i)
          ENDDO
          DO  m = 1, bc_h(1)%ns
             i = bc_h(1)%i(m)            
             j = bc_h(1)%j(m)
             k = bc_h(1)%k(m)
             pt(k+1,j,i) = pt(k,j,i)
          ENDDO   
       ENDIF
    ENDIF
!
!-- Set Neumann boundary conditions for humidity and cloud-physical quantities
    IF ( humidity )  THEN
       IF ( ibc_q_b == 1 )  THEN
          DO  m = 1, bc_h(0)%ns
             i = bc_h(0)%i(m)            
             j = bc_h(0)%j(m)
             k = bc_h(0)%k(m)
             q(k-1,j,i) = q(k,j,i)
          ENDDO  
          DO  m = 1, bc_h(1)%ns
             i = bc_h(1)%i(m)            
             j = bc_h(1)%j(m)
             k = bc_h(1)%k(m)
             q(k+1,j,i) = q(k,j,i)
          ENDDO  
       ENDIF
       IF ( bulk_cloud_model  .AND.  microphysics_morrison )  THEN
          DO  m = 1, bc_h(0)%ns
             i = bc_h(0)%i(m)            
             j = bc_h(0)%j(m)
             k = bc_h(0)%k(m)
             nc(k-1,j,i) = 0.0_wp
             qc(k-1,j,i) = 0.0_wp
          ENDDO  
          DO  m = 1, bc_h(1)%ns
             i = bc_h(1)%i(m)            
             j = bc_h(1)%j(m)
             k = bc_h(1)%k(m)

             nc(k+1,j,i) = 0.0_wp
             qc(k+1,j,i) = 0.0_wp
          ENDDO  
       ENDIF

       IF ( bulk_cloud_model  .AND.  microphysics_seifert )  THEN
          DO  m = 1, bc_h(0)%ns
             i = bc_h(0)%i(m)            
             j = bc_h(0)%j(m)
             k = bc_h(0)%k(m)
             nr(k-1,j,i) = 0.0_wp
             qr(k-1,j,i) = 0.0_wp
          ENDDO  
          DO  m = 1, bc_h(1)%ns
             i = bc_h(1)%i(m)            
             j = bc_h(1)%j(m)
             k = bc_h(1)%k(m)
             nr(k+1,j,i) = 0.0_wp
             qr(k+1,j,i) = 0.0_wp
          ENDDO  
       ENDIF

    ENDIF
!
!-- Set Neumann boundary conditions for passive scalar
    IF ( passive_scalar )  THEN
       IF ( ibc_s_b == 1 )  THEN
          DO  m = 1, bc_h(0)%ns
             i = bc_h(0)%i(m)            
             j = bc_h(0)%j(m)
             k = bc_h(0)%k(m)
             s(k-1,j,i) = s(k,j,i)
          ENDDO 
          DO  m = 1, bc_h(1)%ns
             i = bc_h(1)%i(m)            
             j = bc_h(1)%j(m)
             k = bc_h(1)%k(m)
             s(k+1,j,i) = s(k,j,i)
          ENDDO  
       ENDIF
    ENDIF
!
!-- Set Neumann boundary conditions for chemical species
    IF ( air_chemistry  .AND.  nest_chemistry )  THEN
       IF ( ibc_cs_b == 1 )  THEN
          DO  n = 1, nspec
             DO  m = 1, bc_h(0)%ns
                i = bc_h(0)%i(m)            
                j = bc_h(0)%j(m)
                k = bc_h(0)%k(m)
                chem_species(n)%conc(k-1,j,i) = chem_species(n)%conc(k,j,i)
             ENDDO 
             DO  m = 1, bc_h(1)%ns
                i = bc_h(1)%i(m)            
                j = bc_h(1)%j(m)
                k = bc_h(1)%k(m)
                chem_species(n)%conc(k+1,j,i) = chem_species(n)%conc(k,j,i)
             ENDDO
          ENDDO
       ENDIF
    ENDIF

 END SUBROUTINE pmci_boundary_conds


END MODULE pmc_interface
